/*
 * This file is part of Generic Data Structures Library (GDSL).
 * Copyright (C) 1998-2001 Nicolas Darnis <ndarnis@free.fr>
 *
 * The GDSL library is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * The GDSL library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the GDSL library; see the file COPYING.
 * If not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 * File: _gdsl_bintree.c
 * $Revision: 1.5 $
 * $Date: 2002/02/21 09:55:51 $
 * $Author: ndarnis $
 * $Id: _gdsl_bintree.c,v 1.5 2002/02/21 09:55:51 ndarnis Exp $
 */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "_gdsl_bintree.h"
#include "_gdsl_mem.h"
#include "gdsl_types.h"
#include "gdsl_macros.h"

#define LEFT(t)       ( (t)->left )
#define RIGHT(t)      ( (t)->right )
#define PARENT(t)     ( (t)->parent )
#define CONTENT(t)    ( (t)->content )

#define IS_LEAF(t)    ( LEFT(t) == NULL && RIGHT(t) == NULL )
#define IS_EMPTY(t)   ( (t) == NULL )

struct gdsl_bintree_node
{
  struct gdsl_bintree_node *left;    /* Node's left sub-tree */
  struct gdsl_bintree_node *right;   /* Node's right sub-tree */
  struct gdsl_bintree_node *parent;  /* Node's parent */
  gdsl_element_t            content; /* Node's content */
};

static void
bintree_destroy (_gdsl_bintree_t t);

static void
bintree_destroy_with_func (_gdsl_bintree_t t, gdsl_free_func_t f);

/******************************************************************************/
/* Management                  functions                                      */
/******************************************************************************/

extern _gdsl_bintree_t
_gdsl_bintree_make (const gdsl_element_t e, const _gdsl_bintree_t l, 
		    const _gdsl_bintree_t r)
{
  _gdsl_bintree_t t;

  t = (_gdsl_bintree_t) _gdsl_mem_alloc (sizeof (struct gdsl_bintree_node));
  if (t == NULL)
    {
      return NULL;
    }
  
  t->left = l;
  t->right = r;

  if (l != NULL)
    {
      l->parent = t;
    }

  if (r != NULL)
    {
      r->parent = t;
    }

  t->parent = NULL;
  t->content = e;

  return t;
}

extern void 
_gdsl_bintree_destroy (_gdsl_bintree_t t, const gdsl_free_func_t free_f)
{
  free_f == NULL ? 
    bintree_destroy (t) : bintree_destroy_with_func (t, free_f);
}

extern _gdsl_bintree_t
_gdsl_bintree_copy (const _gdsl_bintree_t t, const gdsl_copy_func_t copy_f)
{
  _gdsl_bintree_t tmp;

  if (IS_EMPTY (t))
    {
      return NULL;
    }

  tmp = _gdsl_bintree_create (copy_f (CONTENT (t)));
  if (tmp == NULL)
    {
      return NULL;
    }

  tmp->left  = _gdsl_bintree_copy (LEFT (t), copy_f);
  if (LEFT (tmp) != NULL)
    {
      tmp->left->parent = tmp;
    }

  tmp->right = _gdsl_bintree_copy (RIGHT (t), copy_f);
  if (RIGHT (tmp) != NULL)
    {
      tmp->right->parent = tmp;
    }

  return tmp;
}

/******************************************************************************/
/* Information functions                                                      */
/******************************************************************************/

extern bool
_gdsl_bintree_is_empty (const _gdsl_bintree_t t)
{
  return (bool) (IS_EMPTY (t));
}

extern bool
_gdsl_bintree_is_leaf (const _gdsl_bintree_t t)
{
  assert (t != NULL);

  return (bool) (IS_LEAF (t));
}

extern gdsl_element_t
_gdsl_bintree_get_content (const _gdsl_bintree_t t)
{
  assert (t != NULL);

  return CONTENT (t);
}

extern bool
_gdsl_bintree_is_root (const _gdsl_bintree_t t)
{
  assert (t != NULL);

  return (bool) (PARENT (t) == NULL);
}

extern _gdsl_bintree_t
_gdsl_bintree_get_parent (const _gdsl_bintree_t t)
{
  assert (t != NULL);

  return PARENT (t);
}

extern _gdsl_bintree_t
_gdsl_bintree_get_left_subtree (const _gdsl_bintree_t t)
{
  assert (t != NULL);
  
  return LEFT (t);
}

extern _gdsl_bintree_t
_gdsl_bintree_get_right_subtree (const _gdsl_bintree_t t)
{
  assert (t != NULL);

  return RIGHT (t);
}

extern _gdsl_bintree_t *
_gdsl_bintree_get_left_ref (const _gdsl_bintree_t t)
{
  assert (t != NULL);

  return &LEFT (t);
}

extern _gdsl_bintree_t *
_gdsl_bintree_get_right_ref (const _gdsl_bintree_t t)
{
  assert (t != NULL);

  return &RIGHT (t);
}

extern ulong
_gdsl_bintree_height (const _gdsl_bintree_t t)
{
  if (IS_EMPTY (t)) 
    {
      return 0UL;
    }

  if (IS_LEAF (t))
    {
      return 0UL;
    }

  return (ulong) (1UL 
		  + MAX (_gdsl_bintree_height (LEFT (t)), 
			 _gdsl_bintree_height (RIGHT (t))));
}

extern ulong
_gdsl_bintree_size (const _gdsl_bintree_t t)
{
  if (IS_EMPTY (t))
    {
      return 0UL;
    }

  return (ulong) (1UL 
		  + _gdsl_bintree_size (LEFT (t)) 
		  + _gdsl_bintree_size (RIGHT (t)));
}

/******************************************************************************/
/* Modification functions                                                     */
/******************************************************************************/

extern void
_gdsl_bintree_set_content (_gdsl_bintree_t t, const gdsl_element_t e)
{
  assert (t != NULL);

  t->content = e;
}

extern void
_gdsl_bintree_set_parent (_gdsl_bintree_t t, const _gdsl_bintree_t p)
{
  assert (t != NULL);

  t->parent = p;
}

extern void
_gdsl_bintree_set_left_subtree (_gdsl_bintree_t t, const _gdsl_bintree_t l)
{
  assert (t != NULL);

  t->left = l;

  if (l != NULL)
    {
      l->parent = t;
    }
}

extern void 
_gdsl_bintree_set_right_subtree (_gdsl_bintree_t t, const _gdsl_bintree_t r)
{
  assert (t != NULL);

  t->right = r;

  if (r != NULL)
    {
      r->parent = t;
    }
}

/******************************************************************************/
/* Rotation functions                                                         */
/******************************************************************************/

extern _gdsl_bintree_t
_gdsl_bintree_left_rot (_gdsl_bintree_t *t)
{
  _gdsl_bintree_t rn;

  assert (*t != NULL);
  assert (RIGHT (*t) != NULL);

  rn = RIGHT (*t);
  (*t)->right = LEFT (rn);

  if (LEFT (rn) != NULL)
    {
      rn->left->parent = *t;
    }

  rn->parent = PARENT (*t);
  
  rn->left = *t;
  (*t)->parent = rn;
  *t = rn;

  return rn;
}

extern _gdsl_bintree_t
_gdsl_bintree_right_rot (_gdsl_bintree_t *t)
{
  _gdsl_bintree_t ln;

  assert (*t != NULL);
  assert (LEFT (*t) != NULL);

  ln = LEFT (*t);
  (*t)->left = RIGHT (ln);
  
  if (RIGHT (ln) != NULL)
    {
      ln->right->parent = *t;
    }
  
  ln->parent = PARENT (*t);
  
  ln->right = *t;
  (*t)->parent = ln;
  *t = ln;
  
  return ln;
}

extern _gdsl_bintree_t
_gdsl_bintree_left_right_rot (_gdsl_bintree_t *t)
{
  assert (*t != NULL);
  assert (LEFT (*t) != NULL);
  assert (RIGHT (LEFT (*t)) != NULL);

  _gdsl_bintree_left_rot (&LEFT (*t));
  return _gdsl_bintree_right_rot (t);
}

extern _gdsl_bintree_t
_gdsl_bintree_right_left_rot (_gdsl_bintree_t *t)
{
  assert (*t != NULL);
  assert (RIGHT (*t) != NULL);
  assert (LEFT (RIGHT (*t)) != NULL);

  _gdsl_bintree_right_rot (&RIGHT (*t));
  return _gdsl_bintree_left_rot (t);
}

/******************************************************************************/
/* Parse functions                                                            */
/******************************************************************************/

extern _gdsl_bintree_t
_gdsl_bintree_prefix_parse (const _gdsl_bintree_t t, 
			    const gdsl_bintree_map_func_t map_f, void *d)
{
  assert (map_f != NULL);

  if (!IS_EMPTY (t))
    {
      if (map_f (t, d) == 0)
	{
	  return t;
	}
      
      _gdsl_bintree_prefix_parse (LEFT (t), map_f, d);
      _gdsl_bintree_prefix_parse (RIGHT (t), map_f, d); 
    }
  
  return NULL;
}

extern _gdsl_bintree_t
_gdsl_bintree_infix_parse (const _gdsl_bintree_t t, 
			   const gdsl_bintree_map_func_t map_f, void *d)
{
  assert (map_f != NULL);

  if (!IS_EMPTY (t))
    {
      _gdsl_bintree_infix_parse (LEFT (t), map_f, d);
      
      if (map_f (t, d) == 0) 
	{
	  return t;
	}
      
      _gdsl_bintree_infix_parse (RIGHT (t), map_f, d); 
    }
  
  return NULL;
}

extern _gdsl_bintree_t
_gdsl_bintree_postfix_parse (const _gdsl_bintree_t t, 
			     const gdsl_bintree_map_func_t map_f, void *d)
{
  assert (map_f != NULL);

  if (!IS_EMPTY (t))
    {
      _gdsl_bintree_postfix_parse (LEFT (t), map_f, d);
      _gdsl_bintree_postfix_parse (RIGHT (t), map_f, d); 
      
      if (map_f (t, d) == 0) 
	{
	  return t;
	}
    }
  
  return NULL;
}

#if 0
extern gdsl_element_t
_gdsl_bintree_level_parse (const _gdsl_bintree_t t, 
			   const gdsl_map_func_t map_f, void *d)
{
  /* !!! TO DO ... */
  return NULL;
}
#endif
/******************************************************************************/
/* Input/output functions                                                     */
/******************************************************************************/

extern void
_gdsl_bintree_write (const _gdsl_bintree_t t, const gdsl_write_func_t write_f, 
		     FILE *file, void *d)
{
  assert (write_f != NULL);
  assert( file != NULL);

  if (!IS_EMPTY (t))
    { 
      write_f (CONTENT (t), file, d);
      _gdsl_bintree_write (LEFT (t), write_f, file, d);
      _gdsl_bintree_write (RIGHT (t), write_f , file, d);
    }
}

extern void
_gdsl_bintree_write_xml (const _gdsl_bintree_t t, 
			 const gdsl_write_func_t write_f, FILE *file, void *d)
{
  assert( file != NULL);

  if (!IS_EMPTY (t))
    { 
      if (IS_LEAF (t))
	{
	  fprintf (file, "<GDSL_BINTREE_LEAF REF=\"%p\"", (void *) t);
	}
      else
	{
	  fprintf (file, "<GDSL_BINTREE_NODE REF=\"%p\"", (void *) t);
	}

      if (!IS_LEAF (t))
	{
	  if (LEFT (t) != NULL)
	    {
	      fprintf (file, " LEFT=\"%p\"", (void *) LEFT (t));
	    }
	  else
	    {
	      fprintf (file, " LEFT=\"\"");
	    }
      
	  if (RIGHT (t) != NULL)
	    {
	      fprintf (file, " RIGHT=\"%p\"", (void *) RIGHT (t));
	    }
	  else
	    {
	      fprintf (file, " RIGHT=\"\"");
	    }
	}
      
      if (PARENT (t) != NULL)
	{
	  fprintf (file, " PARENT=\"%p\"", (void *) PARENT (t));
	}
      else
	{
	  fprintf (file, " PARENT=\"\"");
	}

      fprintf (file, ">");

      if (write_f != NULL)
	{
	  write_f (CONTENT (t), file, d);
	}

      if (IS_LEAF (t))
	{
	  fprintf (file, "</GDSL_BINTREE_LEAF>\n");
	}
      else
	{
	  fprintf (file, "</GDSL_BINTREE_NODE>\n");
	}

      _gdsl_bintree_write_xml (LEFT (t), write_f, file, d);
      _gdsl_bintree_write_xml (RIGHT (t), write_f , file, d);
    }
}

extern void
_gdsl_bintree_dump (const _gdsl_bintree_t t, const gdsl_write_func_t write_f, 
		    FILE *file, void *d)
{
  assert( file != NULL);

  if (!IS_EMPTY (t))
    { 
      if (IS_LEAF (t))
	{
	  fprintf (file, "<GDSL_BINTREE_LEAF REF=\"%p\"", (void *) t);
	}
      else
	{
	  fprintf (file, "<GDSL_BINTREE_NODE REF=\"%p\"", (void *) t);
	}

      if (CONTENT (t))
	{
	  fprintf (file, " CONTENT=\"%p\"",  (void *) CONTENT (t));
	}
      else
	{
	  fprintf (file, " CONTENT=\"\"");
	}
      
      if (!IS_LEAF (t))
	{
	  if (LEFT (t) != NULL)
	    {
	      fprintf (file, " LEFT=\"%p\"", (void *) LEFT (t));
	    }
	  else
	    {
	      fprintf (file, " LEFT=\"\"");
	    }
      
	  if (RIGHT (t) != NULL)
	    {
	      fprintf (file, " RIGHT=\"%p\"", (void *) RIGHT (t));
	    }
	  else
	    {
	      fprintf (file, " RIGHT=\"\"");
	    }
	}
      
      if (PARENT (t) != NULL)
	{
	  fprintf (file, " PARENT=\"%p\"", (void *) PARENT (t));
	}
      else
	{
	  fprintf (file, " PARENT=\"\"");
	}

      fprintf (file, ">");

      if (write_f != NULL)
	{
	  write_f (CONTENT (t), file, d);
	}

      if (IS_LEAF (t))
	{
	  fprintf (file, "</GDSL_BINTREE_LEAF>\n");
	}
      else
	{
	  fprintf (file, "</GDSL_BINTREE_NODE>\n");
	}

      _gdsl_bintree_dump (LEFT (t), write_f, file, d);
      _gdsl_bintree_dump (RIGHT (t), write_f , file, d);
    }
}

/******************************************************************************/
/* Display functions                                                          */
/******************************************************************************/

/******************************************************************************/
/* Private functions                                                          */
/******************************************************************************/

static void
bintree_destroy (_gdsl_bintree_t t)
{
  if (!IS_EMPTY (t))
    {
      bintree_destroy (LEFT (t));
      bintree_destroy (RIGHT (t));
      _gdsl_mem_free (t);
    }
}

static void
bintree_destroy_with_func (_gdsl_bintree_t t, gdsl_free_func_t free_f)
{
  if (!IS_EMPTY (t))
    {
      bintree_destroy_with_func (LEFT (t), free_f);
      bintree_destroy_with_func (RIGHT (t), free_f);
      free_f (CONTENT (t));
      _gdsl_mem_free (t);
    }
}

/*
 * WARNING: the code below is very dirty and is only here for debugging purpose.
 * In the future this code will change.
 */
#if defined (HAVE_GFX_DEBUG)

#include <string.h>
#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#define X_Win_pos 0
#define Y_Win_pos 0
#define WIN_NAME "BINARY TREE"

#define RAYON 6
#define BORD  3 * RAYON
#define DX  3 * RAYON
#define DY 6 * RAYON

#define COLOR1   "grey"
#define COLOR2   "blue"

/* declaration des identificateurs de ressources X */
Window Fen;                      /* la fenetre Globale    */
GC     ContGraph;                /* contexte graphique    */

/* declaration des variables utilisees par les fonctions X */
Display                *display;
XEvent                 evenement;            /* file des evenements */
XSizeHints             indic;
int                    ecran;
unsigned long          noir,blanc;
char                   *WinName = WIN_NAME;  /* nom de la Fen */
Colormap               Cmap;
XColor                 coul1;   /* meilleure approximation RVB rendue */
XColor                 coul2;
XColor                 exact_c; /* couleur RVB exacte correspondant au nom */
Font                   fonte;

/*Declaration des variabes propres a l'arbre */
int Xo,Yo,X,Y;
static bool InitContexte( int LFen, int HFen );
static void CloseFen( void );
static bool Touche( KeySym tch );
static void TraceABin( _gdsl_bintree_t A, int x, int y, char * (*key_display)(gdsl_element_t) );
#endif /* HAVE_GFX_DEBUG */

extern void 
_gdsl_bintree_gfx_display (const _gdsl_bintree_t A, const char * (*key_display)(gdsl_element_t) )
{
#ifdef HAVE_GFX_DEBUG
  int encore = TRUE;
  int Lfen, Hfen;      
  KeySym touche;
  
  if( !IS_EMPTY( A ) ) 
    {
      Lfen = (_gdsl_bintree_size( A ) - 1 ) * DX + 2 * BORD;  /* largeur de la Fen Globale     */
      Hfen = _gdsl_bintree_height( A ) * DY + 2 * BORD;     /* Hauteur de la Fen GLOBALE     */
      Xo = _gdsl_bintree_size( LEFT( A ) ) * DX + BORD;   /* abscisse de la racine de l'arbre  */
      X = Xo; Yo = BORD;Y = Yo;
      if ( InitContexte( Lfen, Hfen ) )
	{   
	  /* BOUCLE PRINCIPALE */
	  while( encore )
	    {
	      /*lecture de la file d'attente des evenements */
	      XNextEvent( display, &evenement );
	      switch( evenement.type )
		{
		case Expose: 
		  /* apres le dernier evenement du type expose(affichage complet de la Fen) */
		  if( evenement.xexpose.count == 0 )
		    TraceABin( A, X, Y, key_display );
		  break;

		case KeyPress:
		  touche = XLookupKeysym( (XKeyEvent *)&evenement, 0 );
		  if( ( encore = Touche( touche ) ) )
		    {
		      XClearArea(display, Fen, 0, 0, 0, 0, TRUE );
		      TraceABin( A, X, Y, key_display );
		    }
		}
	    }
	}
      CloseFen(); 
    }
#endif /* HAVE_GFX_DEBUG */
}

#ifdef HAVE_GFX_DEBUG
static bool 
InitContexte( int LFen, int HFen )
{
  /* demande dans la base de donnee          */
  /* connexion au serveur X                  */
  display = XOpenDisplay( NULL );

  if ( display == NULL ) 
    {
      printf( "Impossible d'ouvrir le display.\n" );
      return( FALSE );
    }
  
  /* consultation des caracteristiques du serveur X      */
  ecran = DefaultScreen( display );
  blanc = WhitePixel( display, ecran );
  noir = BlackPixel( display, ecran );
  
  /* definition de deux couleurs                         */
  Cmap = DefaultColormap( display, DefaultScreen( display ) );
  if( XAllocNamedColor( display, Cmap, COLOR1, &coul1, &exact_c ) == 0 )
    coul1.pixel = noir;
  
  if( XAllocNamedColor( display, Cmap, COLOR2, &coul2, &exact_c ) == 0 )
    coul2.pixel=blanc;
  
  /* Creation de la Fen top-niveau du nom de "GLOBALE" */
  indic.x = X_Win_pos;
  indic.y = Y_Win_pos;
  indic.width = LFen;
  indic.height = HFen;
  indic.flags = PPosition | PSize;
  
  Fen = XCreateSimpleWindow( display, DefaultRootWindow( display ),
			     indic.x, indic.y, indic.width, indic.height,
			     4, coul2.pixel, coul1.pixel );
  
  /* definition des proprietes standard de la Fen Globale pour gestionnaire X */
  XSetStandardProperties( display, Fen,WinName,WinName, None, NULL, 0, &indic );
  
  /* Choix des evenements qui nous interressent en ce qui concerne la Fen */
  XSelectInput( display, Fen, KeyPressMask | ExposureMask ); 
  
  /* projection de toutes la Fen pour la rendre visible            */
  /* l'action "visible" est dans le tampon, mais non encore effectuee  */
  /* les actions s'effectuent avec la fonction XNextEvent, par exemple */
  XMapRaised( display, Fen );

  /* Creation du contexte graphique necessaire pour les requetes d'affichage  */
  ContGraph = XCreateGC( display, Fen, 0, NULL );
  XSetBackground( display, ContGraph, coul1.pixel) ;
  XSetForeground( display, ContGraph, coul2.pixel );

  /* Selection de la police */
  fonte = XLoadFont( display, "6x10" );
  XSetFont( display, ContGraph, fonte );

  return TRUE;
}  

static void 
CloseFen( void )
{
  /* liberation des ressources de la Fen */
  XFreeGC( display, ContGraph );
  XDestroyWindow( display, Fen );
  XCloseDisplay( display );
}

static bool 
Touche( KeySym tch )
{
  switch( tch )
    {
    case XK_Up: 
      Y += 100;   
      break;
      
    case XK_Right:
      X -= 100;   
      break;

    case XK_Down:
      Y -= 100;   
      break;

    case XK_Left:
      X += 100;   
      break;
      
    case XK_Return:
      X = Xo;
      Y = Yo;
      break;

    case XK_space:
      return FALSE;
      break; 
    }

  return TRUE;
}

static void 
TraceABin( _gdsl_bintree_t A, int x, int y, char * (*key_display)(gdsl_element_t) )
{
  int Xg, Xd, Ygd;
  char chr[35];
  
  if( x >= 0 && y >=0 )
    { 
      /* condition lie a des problemes d'affichage hors fenetre... */
      XSetForeground(display, ContGraph, coul2.pixel); 
      sprintf( chr, "%s", key_display( CONTENT( A ) ) );
      XDrawString(display, Fen, ContGraph,x-RAYON ,y+RAYON ,chr ,strlen(chr));
    }
  
  XSetForeground( display, ContGraph, noir );
  Ygd = y + DY;
  
  if( !IS_EMPTY( LEFT( A ) ) )
    {
      Xg = x - ( 1 + _gdsl_bintree_size( RIGHT( LEFT( A ) ) ) ) * DX;
      XDrawLine(display, Fen, ContGraph, x-RAYON, y+RAYON, Xg+RAYON, Ygd-RAYON );
      TraceABin( LEFT( A ), Xg, Ygd, key_display );
    }
  if ( !IS_EMPTY( RIGHT( A ) ) )
    {
      Xd = x + ( 1 + _gdsl_bintree_size( LEFT( RIGHT( A ) ) ) ) * DX;
      if( Xd >= 0 && Ygd >= 0)
	XDrawLine(display, Fen, ContGraph, x+RAYON, y+RAYON, Xd-RAYON, Ygd-RAYON);
      
      TraceABin( RIGHT( A ), Xd, Ygd, key_display );
    }
}
#endif /* HAVE_GFX_DEBUG */
