/*
 * This file is part of Generic Data Structures Library (GDSL).
 * Copyright (C) 1998-2001 Nicolas Darnis <ndarnis@free.fr>
 *
 * The GDSL library is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * The GDSL library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the GDSL library; see the file COPYING.
 * If not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 * File: _gdsl_bstree.c
 * $Revision: 1.5 $
 * $Date: 2002/02/21 09:55:51 $
 * $Author: ndarnis $
 * $Id: _gdsl_bstree.c,v 1.5 2002/02/21 09:55:51 ndarnis Exp $
 */

#include <config.h>

#include <stdlib.h>
#include <assert.h>

#include "gdsl_types.h"
#include "_gdsl_bintree.h"
#include "_gdsl_bstree.h"
#include "gdsl_types.h"
#include "_gdsl_mem.h"

#define LEFT(t)       ( (_gdsl_bstree_t) _gdsl_bintree_get_left_subtree ((_gdsl_bintree_t) (t)) )
#define RIGHT(t)      ( (_gdsl_bstree_t) _gdsl_bintree_get_right_subtree ((_gdsl_bintree_t) (t)) )
#define PARENT(t)     ( (_gdsl_bstree_t) _gdsl_bintree_get_parent ((_gdsl_bintree_t) (t)) )
#define CONTENT(t)    ( _gdsl_bintree_get_content ((_gdsl_bintree_t) (t)) )

#define IS_LEAF(t)    ( _gdsl_bintree_is_leaf ((_gdsl_bintree_t) t) )
#define IS_EMPTY(t)   ( _gdsl_bintree_is_empty ((_gdsl_bintree_t) t) )

static gdsl_element_t
destroy_max (_gdsl_bstree_t *t);

static gdsl_element_t
destroy_min (_gdsl_bstree_t *t);

static void
cut (gdsl_element_t e, _gdsl_bstree_t t, _gdsl_bstree_t l, _gdsl_bstree_t r, 
     gdsl_compare_func_t f);

/******************************************************************************/
/* Allocation and deallocation functions                                      */
/******************************************************************************/

extern _gdsl_bstree_t
_gdsl_bstree_create (const gdsl_element_t e)
{
  return (_gdsl_bstree_t) _gdsl_bintree_create (e);
}

extern void 
_gdsl_bstree_destroy (_gdsl_bstree_t t, const gdsl_free_func_t free_f)
{
  _gdsl_bintree_destroy ((_gdsl_bintree_t) t, free_f);
}

extern _gdsl_bstree_t
_gdsl_bstree_copy (const _gdsl_bstree_t t, const gdsl_copy_func_t copy_f)
{
  assert (copy_f != NULL);

  return (_gdsl_bstree_t) _gdsl_bintree_copy ((_gdsl_bintree_t) t, copy_f);
}

/******************************************************************************/
/* Information functions                                                      */
/******************************************************************************/

extern bool
_gdsl_bstree_is_empty (const _gdsl_bstree_t t)
{
  return IS_EMPTY (t);
}

extern bool
_gdsl_bstree_is_leaf (const _gdsl_bstree_t t)
{
  assert (t != NULL);

  return IS_LEAF (t);
}

extern gdsl_element_t
_gdsl_bstree_get_content (const _gdsl_bstree_t t)
{
  assert (t != NULL);

  return CONTENT (t);
}

extern bool
_gdsl_bstree_is_root (const _gdsl_bstree_t t)
{
  assert (t != NULL);

  return _gdsl_bintree_is_root ((_gdsl_bintree_t) t);
}

extern _gdsl_bstree_t
_gdsl_bstree_get_parent (const _gdsl_bstree_t t)
{
  assert (t != NULL);

  return PARENT (t);
}

extern _gdsl_bstree_t
_gdsl_bstree_get_left_subtree (const _gdsl_bstree_t t)
{
  assert (t != NULL);

  return LEFT (t);
}

extern _gdsl_bstree_t
_gdsl_bstree_get_right_subtree (const _gdsl_bstree_t t)
{
  assert (t != NULL);

  return RIGHT (t);
}

extern ulong
_gdsl_bstree_height (const _gdsl_bstree_t t)
{
  return _gdsl_bintree_height ((_gdsl_bintree_t) t);
}

extern ulong
_gdsl_bstree_size (const _gdsl_bstree_t t)
{
  return _gdsl_bintree_size ((_gdsl_bintree_t) t);
}

/******************************************************************************/
/* Inserting/removing functions                                               */
/******************************************************************************/

extern gdsl_element_t
_gdsl_bstree_insert (_gdsl_bstree_t *t, const gdsl_compare_func_t f, 
		     const gdsl_element_t v)
{
  int comp = -1;

#if defined (_GDSL_OPTIMIZE_TIME_)

  _gdsl_bstree_t n;
  _gdsl_bintree_t root;
  _gdsl_bintree_t parent = NULL;

  assert (f != NULL);

  root = (_gdsl_bintree_t) *t;

  while (!IS_EMPTY (root))
    {
      parent = root;
      comp = f (CONTENT (root), v);

      /* Duplicates keys are not allowed (must be changed) */
      if (comp == 0)
	{
	  return NULL;
	}

      root = (comp > 0) ? LEFT (root) : RIGHT (root);
    }
  
  n = (_gdsl_bstree_t) _gdsl_bintree_create (v);
  if (IS_EMPTY (n))
    {
      return NULL;
    }

  if (IS_EMPTY (parent))
    {
      *t = n;
    }
  else
    {
      if (comp > 0)
	{
	  _gdsl_bintree_set_left_subtree (parent, n);
	}
      else
	{
	  _gdsl_bintree_set_right_subtree (parent, n);
	}
    }

  _gdsl_bintree_set_parent (n, parent);

  return _gdsl_bintree_get_content ((_gdsl_bintree_t) n);

#else

  gdsl_element_t e = NULL;

  assert (f != NULL);

  if (IS_EMPTY (*t))
    {
      *t = (_gdsl_bstree_t) _gdsl_bintree_create (v);
      if (IS_EMPTY (*t))
	{
	  return NULL;
	}

      return _gdsl_bintree_get_content ((_gdsl_bintree_t) *t);
    }

  {
    _gdsl_bstree_t l, r;

    comp = f (CONTENT (*t), v);

    /* Duplicates keys are not allowed (must be changed) */
    if (comp == 0)
      {
	return NULL;
      }

    if (comp > 0) /* v < CONTENT( *t ) */
      {
	l = (_gdsl_bstree_t) _gdsl_bintree_get_left_subtree ((_gdsl_bintree_t) (*t));
	e = _gdsl_bstree_insert (&l, f, v);
	_gdsl_bintree_set_left_subtree ((_gdsl_bintree_t) (*t), l);
      }
    else /* v > CONTENT( *t ) */
      {
	r = (_gdsl_bstree_t) _gdsl_bintree_get_right_subtree ((_gdsl_bintree_t) (*t));
	e = _gdsl_bstree_insert (&r, f, v);
	_gdsl_bintree_set_right_subtree ((_gdsl_bintree_t) (*t), r);
      }
  }

  return e;

#endif /* not _GDSL_OPTIMIZE_TIME_ */
}

extern gdsl_element_t
_gdsl_bstree_remove_by_value (_gdsl_bstree_t *t, 
			      const gdsl_compare_func_t comp_f, void *v)
{
  gdsl_element_t e;
  _gdsl_bstree_t l;
  _gdsl_bstree_t r;

  assert (comp_f != NULL);

  if (IS_EMPTY (*t))
    {
      return NULL;
    }

  e = CONTENT (*t);

  {
    int comp = comp_f (v, e);
    
    if (comp < 0)
      {
	return _gdsl_bstree_remove_by_value (_gdsl_bintree_get_left_ref (*t), 
					     comp_f, v);
      }

    if (comp > 0)
      {
	return _gdsl_bstree_remove_by_value (_gdsl_bintree_get_right_ref (*t), 
					     comp_f, v);
      }
  }
  
  /* comp == 0 */
  l = LEFT (*t);
  r = RIGHT (*t);

  if (IS_EMPTY (l))
    {
      e = CONTENT (*t);
      _gdsl_mem_free (*t);

      if (!IS_EMPTY (r))
	{
	  _gdsl_bintree_set_parent (r, NULL);
	}

      *t = r;
      return e;
    }
 
  if (IS_EMPTY (r))
    {
      e = CONTENT (*t);
      _gdsl_mem_free (*t);

      if (!IS_EMPTY (l))
	{
	  _gdsl_bintree_set_parent (l, NULL);
	}

      *t = l;
      return e;
    }

  /*
   * NOTE: here, we choose to remove the max element from t's left sub-tree. But
   * there exists another alternative that consist of removing the min element
   * from t's right sub-tree. It's an arbitrary choice, but R. Sedgewick tells
   * that statistically, the first method is more efficient.
   *
   * To use the alternative:
   * _gdsl_bintree_set_content ((_gdsl_bintree_t) (*t), 
   *                           destroy_min (_gdsl_bintree_get_right_ref (*t)));
   */
  _gdsl_bintree_set_content ((_gdsl_bintree_t) (*t), 
			     destroy_max (_gdsl_bintree_get_left_ref (*t)));

  return e;
}

/******************************************************************************/
/* Research functions                                                         */
/******************************************************************************/

extern _gdsl_bstree_t
_gdsl_bstree_search (const _gdsl_bstree_t t, const gdsl_compare_func_t f, 
		     void *v)
{
  int comp;

  assert (f != NULL);

#if defined (_GDSL_OPTIMIZE_TIME_)

  while (!IS_EMPTY (t))
    {
      comp = f (CONTENT (t), v);

      if (comp == 0)
	{
	  return t;
	}

      if (comp > 0)
	{
	  t = LEFT (t);
	}
      else
	{
	  t = RIGHT (t);
	}
    }

  return NULL;

#else
  
  if (IS_EMPTY (t))
    {
      return NULL;
    }

  comp = f (CONTENT (t), v);

  if (comp == 0)
    {
      return t;
    }

  if (comp > 0)
    {
      return _gdsl_bstree_search (LEFT (t), f, v);
    }

  return _gdsl_bstree_search (RIGHT (t), f, v);

#endif /* not _GDSL_OPTIMIZE_TIME_ */
}

#if 0
extern _gdsl_bstree_t
_gdsl_bstree_get_next (const _gdsl_bstree_t t, const gdsl_compare_func_t f,
		       void *key )
{
#ifdef _GDSL_OPTIMIZE_TIME_
#else

  _gdsl_bstree_t b;
  _gdsl_bstree_t c;

  b = _gdsl_bstree_search( t, f, key );
  if( !b )
    return NULL;

  c = (_gdsl_bstree_t)_gdsl_bintree_get_right_subtree( (_gdsl_bintree_t)b );
  if( c )
    {
      while( _gdsl_bintree_get_left_subtree( (_gdsl_bintree_t)c ) )
	c = (_gdsl_bstree_t)_gdsl_bintree_get_left_subtree( (_gdsl_bintree_t)c );
      return c;
    }
  else
    {
      /* !! TO DO... !! */;
    }

  return NULL;

#endif
}
#endif

/******************************************************************************/
/* Parse functions                                                            */
/******************************************************************************/

extern _gdsl_bstree_t
_gdsl_bstree_prefix_parse (const _gdsl_bstree_t t, 
			   const gdsl_bstree_map_func_t map_f, void *d)
{
  assert (map_f != NULL);

  if (!IS_EMPTY (t))
    {
      if (map_f ( t, d) == 0) 
	{
	  return t;
	}

      _gdsl_bstree_prefix_parse ((_gdsl_bstree_t) 
				 _gdsl_bintree_get_left_subtree ((_gdsl_bintree_t)
								 t), map_f, d);
      _gdsl_bstree_prefix_parse ((_gdsl_bstree_t) 
				 _gdsl_bintree_get_right_subtree ((_gdsl_bintree_t)
								  t), map_f, d); 
    }

  return NULL;
}

extern _gdsl_bstree_t
_gdsl_bstree_infix_parse (const _gdsl_bstree_t t, 
			  const gdsl_bstree_map_func_t map_f, void *d)
{
  assert (map_f != NULL);

  if (!IS_EMPTY (t))
    {
      _gdsl_bstree_infix_parse ((_gdsl_bstree_t) 
				_gdsl_bintree_get_left_subtree ((_gdsl_bintree_t)
								t), map_f, d);

      if (map_f (t, d) == 0) 
	{
	  return t;
	}

      _gdsl_bstree_infix_parse ((_gdsl_bstree_t) 
				_gdsl_bintree_get_right_subtree ((_gdsl_bintree_t)
								 t), map_f, d); 
    }

  return NULL;
}

extern _gdsl_bstree_t
_gdsl_bstree_postfix_parse (const _gdsl_bstree_t t, 
			    const gdsl_bstree_map_func_t map_f, void *d)
{
  assert (map_f != NULL);

  if (!IS_EMPTY (t))
    {
      _gdsl_bstree_postfix_parse ((_gdsl_bstree_t) 
				  _gdsl_bintree_get_left_subtree ((_gdsl_bintree_t)
								  t), map_f, d);
      _gdsl_bstree_postfix_parse ((_gdsl_bstree_t) 
				  _gdsl_bintree_get_right_subtree ((_gdsl_bintree_t)
								   t), map_f, d); 

      if (map_f (t, d) == 0) 
	{
	  return t;
	}
    }

  return NULL;
}

#if 0
extern _gdsl_bstree_t
_gdsl_bstree_level_parse (const _gdsl_bstree_t t, const gdsl_map_func_t map_f,
			  void *d)
{
  /* !! TO DO... !! */
  return NULL;
}
#endif

/******************************************************************************/
/* Input/output functions                                                     */
/******************************************************************************/

extern void
_gdsl_bstree_write (const _gdsl_bstree_t t, const gdsl_write_func_t write_f, 
		    FILE *file, void *d)
{
  assert (write_f != NULL);
  assert (file != NULL);

  _gdsl_bintree_write ((_gdsl_bintree_t) t, write_f, file, d);
}

extern void
_gdsl_bstree_write_xml (const _gdsl_bstree_t t, const gdsl_write_func_t write_f, 
			FILE *file, void *d)
{
  assert (file != NULL);

  if (!IS_EMPTY (t))
    { 
      if (IS_LEAF (t))
	{
	  fprintf (file, "<GDSL_BSTREE_LEAF REF=\"%p\"", (void *) t);
	}
      else
	{
	  fprintf (file, "<GDSL_BSTREE_NODE REF=\"%p\"", (void *) t);
	}

      if (!IS_LEAF (t))
	{
	  if (LEFT (t))
	    {
	      fprintf (file, " LEFT=\"%p\"", (void *) LEFT (t));
	    }
	  else
	    {
	      fprintf (file, " LEFT=\"\"");
	    }
      
	  if (RIGHT (t))
	    {
	      fprintf (file, " RIGHT=\"%p\"", (void *) RIGHT (t));
	    }
	  else
	    {
	      fprintf (file, " RIGHT=\"\"");
	    }
	}
      
      if (PARENT (t))
	{
	  fprintf (file, " PARENT=\"%p\"", (void *) PARENT (t));
	}
      else
	{
	  fprintf (file, " PARENT=\"\"");
	}

      fprintf (file, ">");

      if (write_f != NULL)
	{
	  write_f (CONTENT (t), file, d);
	}

      if (IS_LEAF (t))
	{
	  fprintf (file, "</GDSL_BSTREE_LEAF>\n");
	}
      else
	{
	  fprintf (file, "</GDSL_BINTREE_NODE>\n");
	}

      _gdsl_bstree_write_xml (LEFT (t), write_f, file, d);
      _gdsl_bstree_write_xml (RIGHT (t), write_f , file, d);
    }
}

extern void
_gdsl_bstree_dump (const _gdsl_bstree_t t, const gdsl_write_func_t write_f, 
		   FILE *file, void *d)
{
  assert (file != NULL);

  if (!IS_EMPTY (t))
    { 
      if (IS_LEAF (t))
	{
	  fprintf (file, "<GDSL_BSTREE_LEAF REF=\"%p\"", (void *) t);
	}
      else
	{
	  fprintf (file, "<GDSL_BSTREE_NODE REF=\"%p\"", (void *) t);
	}

      if (CONTENT (t) != NULL)
	{
	  fprintf (file, " CONTENT=\"%p\"",  (void *) CONTENT (t));
	}
      else
	{
	  fprintf (file, " CONTENT=\"\"");
	}
      
      if (!IS_LEAF (t))
	{
	  if (LEFT (t))
	    {
	      fprintf (file, " LEFT=\"%p\"", (void *) LEFT (t));
	    }
	  else
	    {
	      fprintf (file, " LEFT=\"\"");
	    }
      
	  if (RIGHT (t))
	    {
	      fprintf (file, " RIGHT=\"%p\"", (void *) RIGHT (t));
	    }
	  else
	    {
	      fprintf (file, " RIGHT=\"\"");
	    }
	}
      
      if (PARENT (t))
	{
	  fprintf (file, " PARENT=\"%p\"", (void *) PARENT (t));
	}
      else
	{
	  fprintf (file, " PARENT=\"\"");
	}

      fprintf (file, ">");

      if (write_f != NULL)
	{
	  write_f (CONTENT (t), file, d);
	}

      if (IS_LEAF (t))
	{
	  fprintf (file, "</GDSL_BSTREE_LEAF>\n");
	}
      else
	{
	  fprintf (file, "</GDSL_BSTREE_NODE>\n");
	}

      _gdsl_bstree_dump (LEFT (t), write_f, file, d);
      _gdsl_bstree_dump (RIGHT (t), write_f , file, d);
    }
}

/******************************************************************************/
/* Display functions                                                          */
/******************************************************************************/

/*
 *
 */
extern void 
_gdsl_bstree_gfx_display (const _gdsl_bstree_t t, 
			  const char * (*key_display)(gdsl_element_t))
{
  _gdsl_bintree_gfx_display ((_gdsl_bintree_t) t, key_display);
}

/*******************************************************************/
/* Private functions                                               */
/*******************************************************************/

static gdsl_element_t
destroy_max (_gdsl_bstree_t *tree)
{
  if (IS_EMPTY (RIGHT (*tree)))
    {
      gdsl_element_t max = CONTENT (*tree);
      _gdsl_bstree_t t = LEFT (*tree);
      _gdsl_mem_free (*tree);
      *tree = t;
      return max;
    }

  return destroy_max (_gdsl_bintree_get_right_ref (*tree));
}

static gdsl_element_t
destroy_min (_gdsl_bstree_t *tree)
{
    if (IS_EMPTY (LEFT (*tree)))
    {
      gdsl_element_t min = CONTENT (*tree);
      _gdsl_bstree_t t = RIGHT (*tree);
      _gdsl_mem_free (*tree);
      *tree = t;
      return min;
    }

  return destroy_min (_gdsl_bintree_get_left_ref (*tree));
}

static void
cut (gdsl_element_t e, _gdsl_bstree_t t, _gdsl_bstree_t l, _gdsl_bstree_t r, 
     gdsl_compare_func_t f)
{
  if (IS_EMPTY (t))
    {
      l = NULL;
      r = NULL;
      return;
    }

  if (f (e, CONTENT (t)) < 0)
    {
      r = t;
      cut (e, LEFT (t), l, LEFT (r), f);
    }

  l = t;
  cut (e, RIGHT (t), RIGHT (l), r, f);
}

