/*
 * This file is part of Generic Data Structures Library (GDSL).
 * Copyright (C) 1998-2001 Nicolas Darnis <ndarnis@free.fr>
 *
 * The GDSL library is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * The GDSL library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the GDSL library; see the file COPYING.
 * If not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 * File: _gdsl_list.c
 * $Revision: 1.4 $
 * $Date: 2002/02/21 09:55:51 $
 * $Author: ndarnis $
 * $Id: _gdsl_list.c,v 1.4 2002/02/21 09:55:51 ndarnis Exp $
 */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "_gdsl_list_node.h"
#include "gdsl_types.h"
#include "_gdsl_list.h"

/******************************************************************************/
/* Management functions                                                       */
/******************************************************************************/

extern _gdsl_list_t
_gdsl_list_create (const gdsl_element_t e)
{
  _gdsl_list_node_t n;

  n = _gdsl_list_node_alloc ();

  if (n == NULL)
    {
      return NULL;
    }

  _gdsl_list_node_set_content (n, e);

  return n;
}

extern void 
_gdsl_list_destroy (_gdsl_list_node_t n, const gdsl_free_func_t f)
{
  _gdsl_list_node_t save;

  if (f != NULL)
    {
      while (n != NULL)
	{
	  save = _gdsl_list_node_get_succ (n);
	  f( _gdsl_list_node_get_content (n));
	  _gdsl_list_node_free (n);
	  n = save;
	}

      return;
    }

  while (n != NULL)
    {
      save = _gdsl_list_node_get_succ (n);
      _gdsl_list_node_free (n);
      n = save;
    }
}

/******************************************************************************/
/* Information functions                                                      */
/******************************************************************************/

extern bool
_gdsl_list_is_empty (const _gdsl_list_node_t n)
{
  return (bool) (n == NULL);
}

/******************************************************************************/
/* Link/unlink functions                                                      */
/******************************************************************************/

extern void
_gdsl_list_link (_gdsl_list_node_t node1, _gdsl_list_node_t node2)
{
  assert (node1 != NULL);
  assert (node2 != NULL);

  _gdsl_list_node_set_succ (node1, node2);
  _gdsl_list_node_set_pred (node2, node1);
}

extern void
_gdsl_list_unlink (_gdsl_list_node_t node1, _gdsl_list_node_t node2)
{
  assert (node1 != NULL);
  assert (node2 != NULL);

  _gdsl_list_node_set_succ (node1, NULL);
  _gdsl_list_node_set_pred (node2, NULL);
}

/******************************************************************************/
/* Research functions                                                         */
/******************************************************************************/

extern _gdsl_list_node_t
_gdsl_list_search_by_value (_gdsl_list_node_t n, const gdsl_compare_func_t f, 
			    void *v)
{
  _gdsl_list_node_t tmp;

  assert (f != NULL);

  tmp = n;
  while (tmp != NULL)
    {
      if (f (_gdsl_list_node_get_content (tmp), v) == 0)
	{
	  return tmp;
	}

      tmp = _gdsl_list_node_get_succ (tmp);
    }

  return NULL;
}

/******************************************************************************/
/* Insert functions                                                           */
/******************************************************************************/

extern void
_gdsl_list_insert_after (_gdsl_list_node_t n, _gdsl_list_node_t prev)
{
  _gdsl_list_node_t tmp;

  assert (n != NULL);
  assert (prev != NULL);

  tmp = _gdsl_list_node_get_succ (prev);
  _gdsl_list_link (prev, n);

  if (tmp != NULL)
    {
      _gdsl_list_link (n, tmp);
    }
}

extern void
_gdsl_list_insert_before (_gdsl_list_node_t n, _gdsl_list_node_t succ)
{
  _gdsl_list_node_t tmp;

  assert (n != NULL);
  assert (succ != NULL);

  tmp = _gdsl_list_node_get_pred (succ);
  if (tmp != NULL)
    {
      _gdsl_list_link (tmp, n);
    }

  _gdsl_list_link (n, succ);
}

/******************************************************************************/
/* Remove functions                                                           */
/******************************************************************************/

extern gdsl_element_t
_gdsl_list_remove (_gdsl_list_node_t n)
{
  _gdsl_list_node_t succ;
  _gdsl_list_node_t pred;

  assert (n != NULL);

  succ = _gdsl_list_node_get_succ (n);
  pred = _gdsl_list_node_get_pred (n);

  if (succ != NULL)
    {
      _gdsl_list_node_set_pred (succ, pred);
    }

  if (pred != NULL)
    {
      _gdsl_list_node_set_succ (pred, succ);
    }

  _gdsl_list_node_set_pred (n, NULL);
  _gdsl_list_node_set_succ (n, NULL);

  return _gdsl_list_node_get_content (n);
}

/******************************************************************************/
/* Map functions                                                              */
/******************************************************************************/

extern _gdsl_list_node_t
_gdsl_list_map_forward (const _gdsl_list_node_t n, 
			const gdsl_list_map_func_t map_f, void *d)
{
  _gdsl_list_node_t tmp;

  assert (map_f != NULL);

  tmp = n;
  while (tmp != NULL)
    {
      if (map_f (tmp, d) == 0)
	{
	  return tmp;
	}

      tmp = _gdsl_list_node_get_succ (tmp);
    }

  return NULL;
}
		
extern _gdsl_list_node_t
_gdsl_list_map_backward (const _gdsl_list_node_t n, 
			 const gdsl_list_map_func_t map_f, void *d)
{
  _gdsl_list_node_t tmp;

  assert (map_f != NULL);

  tmp  = n;
  while (tmp != NULL)
    {
      if (map_f (tmp, d) == 0)
	{
	  return tmp;
	}

      tmp = _gdsl_list_node_get_pred (tmp);
    }

  return NULL;
}

/******************************************************************************/
/* Input/output functions                                                     */
/******************************************************************************/

extern void
_gdsl_list_write (const _gdsl_list_node_t n, 
		  const gdsl_write_func_t write_f, FILE *file, void *user_data)
{
  _gdsl_list_node_t tmp;

  assert (write_f != NULL);
  assert (file != NULL);

  tmp = n;
  while (tmp != NULL)
    {
      write_f (_gdsl_list_node_get_content (tmp), file, user_data);
      tmp = _gdsl_list_node_get_succ (tmp);
    }
}

extern void
_gdsl_list_write_xml (const _gdsl_list_node_t n, 
		      const gdsl_write_func_t write_f, FILE *file, void *user_data)
{
  _gdsl_list_node_t tmp;

  assert (file != NULL);

  tmp = n;
  while (tmp != NULL)
    {
      _gdsl_list_node_write_xml (tmp, write_f, file, user_data);
      tmp = _gdsl_list_node_get_succ (tmp);
    }
}

extern void
_gdsl_list_dump (const _gdsl_list_node_t n, 
		 const gdsl_write_func_t write_f, FILE *file, void *user_data)
{
  _gdsl_list_node_t tmp;

  assert (file != NULL);

  tmp = n;
  while (tmp != NULL)
    {
      _gdsl_list_node_dump (tmp, write_f, file, user_data);
      tmp = _gdsl_list_node_get_succ (tmp);
    }
}
