/*
 * This file is part of Generic Data Structures Library (GDSL).
 * Copyright (C) 1998-2001 Nicolas Darnis <ndarnis@free.fr>
 *
 * The GDSL library is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * The GDSL library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the GDSL library; see the file COPYING.
 * If not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 * File: _gdsl_list_node.c
 * $Revision: 1.5 $
 * $Date: 2002/02/21 09:55:51 $
 * $Author: ndarnis $
 * $Id: _gdsl_list_node.c,v 1.5 2002/02/21 09:55:51 ndarnis Exp $
 */

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "_gdsl_list_node.h"
#include "gdsl_types.h"
#include "_gdsl_mem.h"

struct _gdsl_list_node
{
  struct _gdsl_list_node *succ;    /* Node's successor */
  struct _gdsl_list_node *pred;    /* Node's predecessor */
  gdsl_element_t          content; /* Node's content */
};

/******************************************************************************/
/* Management functions                                                       */
/******************************************************************************/

extern _gdsl_list_node_t
_gdsl_list_node_alloc (void)
{
  _gdsl_list_node_t n;

  n = (_gdsl_list_node_t) _gdsl_mem_alloc (sizeof (struct _gdsl_list_node));

  if (n == NULL)
    {
      return NULL;
    }

  n->content = NULL;
  n->succ    = NULL;
  n->pred    = NULL;

  return n;
}

extern gdsl_element_t
_gdsl_list_node_free (_gdsl_list_node_t n)
{
  gdsl_element_t e;

  assert (n != NULL);

  e = n->content;
  _gdsl_mem_free (n);

  return e;
}

/******************************************************************************/
/* Information functions                                                      */
/******************************************************************************/

extern _gdsl_list_node_t
_gdsl_list_node_get_succ (const _gdsl_list_node_t n)
{
  assert (n != NULL);

  return n->succ;
}

extern _gdsl_list_node_t
_gdsl_list_node_get_pred (const _gdsl_list_node_t n)
{
  assert (n != NULL);

  return n->pred;
}

extern gdsl_element_t
_gdsl_list_node_get_content (const _gdsl_list_node_t n)
{
  assert (n != NULL);

  return n->content;
}

/******************************************************************************/
/* Modification functions                                                     */
/******************************************************************************/

extern void
_gdsl_list_node_set_succ (_gdsl_list_node_t n, const _gdsl_list_node_t succ)
{
  assert (n != NULL);

  n->succ = succ;
}

extern void
_gdsl_list_node_set_pred (_gdsl_list_node_t n, const _gdsl_list_node_t pred)
{
  assert (n != NULL);

  n->pred = pred;
}

extern void
_gdsl_list_node_set_content (_gdsl_list_node_t n, const gdsl_element_t e)
{
  assert (n != NULL);

  n->content = e;
}

/******************************************************************************/
/* Input/output functions                                                     */
/******************************************************************************/

extern void
_gdsl_list_node_write (const _gdsl_list_node_t n, 
		       const gdsl_write_func_t write_f, FILE *file, void *d)
{
  assert (n != NULL);
  assert (file != NULL);
  assert (write_f != NULL);

  write_f (n->content, file, d);
}

extern void
_gdsl_list_node_write_xml (const _gdsl_list_node_t n, 
			   const gdsl_write_func_t write_f, FILE *file, void *d)
{
  assert (n != NULL);
  assert (file != NULL);

  fprintf (file, "<GDSL_LIST_NODE REF=\"%p\"", (void *) n);

  if (n->succ != NULL)
    {
      fprintf (file, " SUCC=\"%p\"", (void *) n->succ);
    }
  else
    {
      fprintf (file, " SUCC=\"\"");
    }
  
  if (n->pred != NULL)
    {
      fprintf (file, " PRED=\"%p\">", (void *) n->pred);
    }
  else
    {
      fprintf (file, " PRED=\"\">");
    }
  
  if (write_f != NULL)
    {
      write_f (n->content, file, d);
    }
  
  fprintf (file, "</GDSL_LIST_NODE>\n");
}

extern void
_gdsl_list_node_dump (const _gdsl_list_node_t n, 
		      const gdsl_write_func_t write_f, FILE *file, void *d)
{
  assert (n != NULL);
  assert (file != NULL);

  fprintf (file, "<GDSL_LIST_NODE REF=\"%p\"", (void *) n);
  
  if (n->content != NULL)
    {
      fprintf (file, " CONTENT=\"%p\"", (void *) n->content);
    }
  else
    {
      fprintf (file, " CONTENT=\"\"");
    }

  if (n->succ != NULL)
    {
      fprintf (file, " SUCC=\"%p\"", (void *) n->succ);
    }
  else
    {
      fprintf (file, " SUCC=\"\"");
    }

  if (n->pred != NULL)
    {
      fprintf (file, " PRED=\"%p\">", (void *) n->pred);
    }
  else
    {
      fprintf (file, " PRED=\"\">");
    }

  if (write_f != NULL)
    {
      write_f (n->content, file, d);
    }

  fprintf (file, "</GDSL_LIST_NODE>\n");
}

