/*
 * This file is part of Generic Data Structures Library (GDSL).
 * Copyright (C) 1998-2001 Nicolas Darnis <ndarnis@free.fr>
 *
 * The GDSL library is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * The GDSL library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the GDSL library; see the file COPYING.
 * If not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 * File: _gdsl_mem.c
 * $Revision: 1.3 $
 * $Date: 2002/02/21 09:55:51 $
 * $Author: ndarnis $
 * $Id: _gdsl_mem.c,v 1.3 2002/02/21 09:55:51 ndarnis Exp $
 */

#include <config.h>
#include <stdlib.h>

#ifdef HAVE_MEM_DEBUG
#include <pthread.h>
#endif /* HAVE_MEM_DEBUG */

#include "_gdsl_mem.h"
#include "mem.h"

#ifdef HAVE_MEM_DEBUG
pthread_mutex_t allocated_memory_m = PTHREAD_MUTEX_INITIALIZER;
static long int allocated_memory = 0;
#endif /* HAVE_MEM_DEBUG */

/******************************************************************************/
/* Public functions                                                           */
/******************************************************************************/

extern void *
_gdsl_mem_alloc (size_t size)
{
#if defined (HAVE_MEM_DEBUG)
  void *p = mem_alloc (size);

  if (p != NULL)
    {
      pthread_mutex_lock (&allocated_memory_m);
      allocated_memory += sizeof (p);
      pthread_mutex_unlock (&allocated_memory_m);
    }

  return p;
#else
  return mem_alloc (size);
#endif
}

extern void *
_gdsl_mem_calloc (size_t number, size_t size)
{
#if defined (HAVE_MEM_DEBUG)
  void *p = mem_calloc (number, size);
  
  if (p != NULL)
    {
      pthread_mutex_lock (&allocated_memory_m);
      allocated_memory += sizeof (p);
      pthread_mutex_unlock (&allocated_memory_m);
    }

  return p;
#else
  return mem_calloc (number, size);
#endif
}

extern void *
_gdsl_mem_realloc (void *ptr, size_t size)
{
#if defined (HAVE_MEM_DEBUG)
  void *p;

  pthread_mutex_lock (&allocated_memory_m);
  allocated_memory -= sizeof (ptr);
  pthread_mutex_unlock (&allocated_memory_m);

  p = mem_realloc (ptr, size);
  if (p != NULL)
    {
      pthread_mutex_lock (&allocated_memory_m);
      allocated_memory += sizeof (p);
      pthread_mutex_unlock (&allocated_memory_m);
    }
  
  return p;
#else
  return mem_realloc (ptr, size);
#endif
}

extern void 
_gdsl_mem_free (void *p)
{
#if defined (HAVE_MEM_DEBUG)
  pthread_mutex_lock (&allocated_memory_m);
  allocated_memory -= sizeof (p);
  pthread_mutex_unlock (&allocated_memory_m);
#endif
  mem_free (p);
}

extern void 
_gdsl_mem_set_functions (void * (*malloc_user_fun) (size_t),
			void (*free_user_fun) (void *),
			void * (*calloc_user_fun) (size_t, size_t),
			void * (*realloc_user_fun) (void *, size_t))
{
  mem_set_functions (malloc_user_fun, free_user_fun, calloc_user_fun, realloc_user_fun);
}

extern void 
_gdsl_mem_set_default_functions (void)
{
  mem_set_default_functions ();
}

extern long int
_gdsl_mem_get_allocated_memory (void)
{
#ifdef HAVE_MEM_DEBUG
  long int allocated_mem;

  pthread_mutex_lock (&allocated_memory_m);
  allocated_mem = allocated_memory;
  pthread_mutex_unlock (&allocated_memory_m);

  return allocated_mem;
#else
  return 0;
#endif /* not HAVE_MEM_DEBUG */
}

