/*
 * This file is part of Generic Data Structures Library (GDSL).
 * Copyright (C) 1998-2001 Nicolas Darnis <ndarnis@free.fr>
 *
 * The GDSL library is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * The GDSL library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the GDSL library; see the file COPYING.
 * If not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 * File: main_perm.c
 * $Revision: 1.8 $
 * $Date: 2002/02/21 09:55:51 $
 * $Author: ndarnis $
 * $Id: main_perm.c,v 1.8 2002/02/21 09:55:51 ndarnis Exp $
 */

#include <config.h>
#include <stdio.h>
#include <stdlib.h>

#include "gdsl.h"
#include "_integers.h"

#ifdef HAVE_MEM_DEBUG
#include "_gdsl_mem.h"
#endif

static void usage (void);
static void write_integer (gdsl_element_t e, FILE *file, void *user_data);

int main (int argc, char *argv[])
{
  ulong i, j, n;
  gdsl_perm_t alpha;

  if (argc < 2)
    {
      usage ();
      return EXIT_FAILURE;
    }

  n = atoi (argv[1]);

  alpha = gdsl_perm_create ("ALPHA", n, NULL, NULL, NULL, NULL);
  gdsl_perm_randomize (alpha);

  printf ("alpha         = ");
  for (i = 0; i < n - 1; i++)
    printf ("%ld, ", gdsl_perm_get (alpha, i));

  printf ("%ld (%ld cycles, %ld inversions)\n", gdsl_perm_get (alpha, i), 
	  gdsl_perm_cycles_count (alpha),
	  gdsl_perm_inversions_count (alpha));

  gdsl_perm_reverse (alpha);

  printf ("~alpha        = ");
  for (i = 0; i < n - 1; i++)
    printf ("%ld, ", gdsl_perm_get (alpha, i));

  printf ("%ld (%ld cycles, %ld inversions)\n", gdsl_perm_get (alpha, i), 
	  gdsl_perm_cycles_count (alpha),
	  gdsl_perm_inversions_count (alpha));

  gdsl_perm_reverse (alpha);
  gdsl_perm_inverse (alpha);

  printf ("alpha^-1      = ");
  for (i = 0; i < n - 1; i++)
    printf ("%ld, ", gdsl_perm_get (alpha, i));

  printf ("%ld (%ld cycles, %ld inversions)\n", gdsl_perm_get (alpha, i), 
	  gdsl_perm_cycles_count (alpha),
	  gdsl_perm_inversions_count (alpha));

  gdsl_perm_inverse (alpha);

  printf ("alpha         = ");
  for (i = 0; i < n - 1; i++)
    printf ("%ld, ", gdsl_perm_get (alpha, i));

  printf ("%ld (%ld cycles, %ld inversions)\n", gdsl_perm_get (alpha, i), 
	  gdsl_perm_cycles_count (alpha),
	  gdsl_perm_inversions_count (alpha));

  gdsl_perm_to_canonical (alpha);
  printf ("cycles(alpha) = ");
  for (i = 0; i < n - 1; i++)
    printf ("%ld, ", gdsl_perm_get (alpha, i));
  printf ("%ld (%d cycles)\n", gdsl_perm_get (alpha, i), gdsl_perm_canonical_cycles_count (alpha));

  gdsl_perm_canonical_to_perm (alpha);
  printf ("alpha         = ");
  for (i = 0; i < n - 1; i++)
    printf ("%ld, ", gdsl_perm_get (alpha, i));
  printf ("%ld\n", gdsl_perm_get (alpha, i));

  gdsl_perm_destroy (alpha);

  n = 3;
  alpha = gdsl_perm_create (NULL, n, NULL, NULL, NULL, NULL);
  printf ("alpha         = ");
  for (i = 0; i < n - 1; i++)
    printf ("%ld, ", gdsl_perm_get (alpha, i));
  printf ("%ld\n", gdsl_perm_get (alpha, i));
  for (j = 0; j < 6; j++)
    {
      if (gdsl_perm_get_next (alpha) != NULL)
	{
	  printf ("next(alpha)   = ");
	  for (i = 0; i < n - 1; i++)
	    printf ("%ld, ", gdsl_perm_get (alpha, i));
	  printf ("%ld\n", gdsl_perm_get (alpha, i));
	}
    }

  gdsl_perm_destroy (alpha);

  {
    const ulong va[] = {6,1,0,2,8,7,5,4,3};
    int ca[]         = {0,1,2,3,4,5,6,7,8};
    const ulong vb[] = {2,4,0,3,6,8,7,5,1};
    int cb[]         = {0,1,2,3,4,5,6,7,8};
    gdsl_perm_t a, b, ab;

    a = gdsl_perm_create (NULL,9,(const ulong *) &va,NULL,alloc_integer,free_integer);
    for (i = 0; i < gdsl_perm_get_card (a); i++)
      {
	gdsl_perm_set_content (a, i, &ca[i]);
      }
    printf ("a   = ");
    gdsl_perm_write(a,write_integer,stdout,NULL);
    
    printf ("\nb   = ");
    b = gdsl_perm_create (NULL,9,(const ulong *) &vb,NULL,alloc_integer,free_integer);
    for (i = 0; i < gdsl_perm_get_card (b); i++)
      {
	gdsl_perm_set_content (b, i, &cb[i]);
      }
    gdsl_perm_write(b,write_integer,stdout,NULL);

    ab = gdsl_perm_create (NULL,9,NULL,NULL,NULL,NULL);

    gdsl_perm_mul (ab, a, b);
    printf ("\na.b = ");
    gdsl_perm_write(ab,write_integer,stdout,NULL);
    printf ("\n");

    gdsl_perm_inverse (a);
    printf ("a^-1= ");
    gdsl_perm_write(a,write_integer,stdout,NULL);
    printf ("\n");
    gdsl_perm_inverse (a);

    gdsl_perm_reverse (a);
    printf ("~a  = ");
    gdsl_perm_write(a,write_integer,stdout,NULL);
    printf ("\n");
    gdsl_perm_reverse (a);

    gdsl_perm_to_canonical (a);
    printf ("c(a)= ");
    gdsl_perm_write(a,write_integer,stdout,NULL);
    printf ("\n");
    gdsl_perm_canonical_to_perm (a);
    printf ("a   = ");
    gdsl_perm_write(a,write_integer,stdout,NULL);
    printf ("\n");

    gdsl_perm_destroy (a);
    gdsl_perm_destroy (b);
    gdsl_perm_destroy (ab);
  }

#ifdef HAVE_MEM_DEBUG
  printf ("<MEMORY_STATUS: %ld>\n", _gdsl_mem_get_allocated_memory ());
#endif

  return EXIT_SUCCESS;
}

static void usage (void)
{
  printf ("Usage: perm <n>\n");
}

static void write_integer (gdsl_element_t e, FILE *file, void *user_data)
{
  int **n = (int **) e;
  fprintf (file, "[%d] ", *n);
}
