/*
 * This file is part of Generic Data Structures Library (GDSL).
 * Copyright (C) 1998-2001 Nicolas Darnis <ndarnis@free.fr>
 *
 * The GDSL library is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * The GDSL library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the GDSL library; see the file COPYING.
 * If not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 * File: gdsl_2darray.c
 * $Revision: 1.4 $
 * $Date: 2002/02/21 09:55:51 $
 * $Author: ndarnis $
 * $Id: gdsl_2darray.c,v 1.4 2002/02/21 09:55:51 ndarnis Exp $
 */

#include <config.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "gdsl_2darray.h"
#include "_gdsl_mem.h"
#include "gdsl_types.h"

struct gdsl_2darray
{
  char *           name;
  ulong            rows;
  ulong            cols;
  gdsl_element_t ** elements;
  gdsl_alloc_func_t alloc_f;
  gdsl_free_func_t  free_f;
};

static gdsl_element_t default_alloue (void *e);
static void default_detruis (gdsl_element_t e);

/******************************************************************************/
/* Management functions                                                       */
/******************************************************************************/

extern gdsl_2darray_t
gdsl_2darray_create (const char *name, const ulong l, const ulong c,
		     const gdsl_alloc_func_t alloc_f, 
		     const gdsl_free_func_t free_f)
{
  register ulong i;
  gdsl_2darray_t m;

  m = (gdsl_2darray_t) _gdsl_mem_alloc (sizeof (struct gdsl_2darray));
  if (m == NULL)
    {
      return NULL;
    }

  m->name = NULL;

  if (name != NULL)
    {
      m->name = (char *) _gdsl_mem_alloc ((1 + strlen (name)) * sizeof (char));

      if (m->name == NULL)
	{
	  _gdsl_mem_free (m);
	  return NULL;
	}

      strcpy (m->name, name);
    }

  m->rows = l;
  m->cols = c;

  m->elements = (gdsl_element_t **) _gdsl_mem_alloc (sizeof (gdsl_element_t *) * l);
  if (m->elements == NULL)
    {
      _gdsl_mem_free (m->name);
      _gdsl_mem_free (m);
      return NULL;
    }

  (m->elements)[0] = (gdsl_element_t *) _gdsl_mem_alloc (sizeof (gdsl_element_t) * l * c);
  if ((m->elements)[0] == NULL)
    {
      _gdsl_mem_free (m->name);
      _gdsl_mem_free (m->elements);
      _gdsl_mem_free (m);
      return NULL;
    }

  for (i = 1; i < l; i++)
    {
      (m->elements)[i] = (m->elements)[i-1] + c;
    }
  
  for (i = 0; i < l * c; i++)
    {
      (m->elements)[0][i] = NULL;
    }
  
  m->alloc_f = alloc_f ? alloc_f : default_alloue;
  m->free_f  = free_f  ? free_f  : default_detruis;

  return m;
}

extern void 
gdsl_2darray_destroy (gdsl_2darray_t m)
{
  register ulong i;
  
  assert (m != NULL);

  for (i = 0; i < m->rows * m->cols; i++)
    {
      (m->free_f) ((m->elements)[0][i]);
    }

  _gdsl_mem_free ((m->elements)[0]);
  _gdsl_mem_free (m->elements);
  _gdsl_mem_free (m->name);
  _gdsl_mem_free (m);
}

extern const char *
gdsl_2darray_get_name (const gdsl_2darray_t m)
{
  assert (m != NULL);

  return m->name;
}

extern gdsl_2darray_t
gdsl_2darray_set_name (gdsl_2darray_t m, const char *name)
{
  assert (m != NULL);

  if (m->name != NULL)
    {
      _gdsl_mem_free (m->name);
      m->name = NULL;
    }

  if (name != NULL)
    {
      m->name = (char *) _gdsl_mem_alloc ((1 + strlen (name)) * sizeof (char));

      if (m->name == NULL)
	{
	  return NULL;
	}

      strcpy (m->name, name);
    }

  return m;
}

extern ulong
gdsl_2darray_get_rows_count (const gdsl_2darray_t a)
{
  assert (a != NULL);

  return a->rows;
}

extern ulong
gdsl_2darray_get_columns_count (const gdsl_2darray_t a)
{
  assert (a != NULL);

  return a->cols;
}

extern gdsl_element_t
gdsl_2darray_get_content (const gdsl_2darray_t a, const ulong l, const ulong c)
{
  assert (a != NULL);
  assert (l <= a->rows);
  assert (c <= a->cols);

  return (a->elements)[l][c];
}

extern gdsl_element_t
gdsl_2darray_set_content (gdsl_2darray_t a, const ulong l, const ulong c, 
			  void *e)
{
  assert (a != NULL);
  assert (l <= a->rows);
  assert (c <= a->cols);

  (a->elements)[l][c] = (a->alloc_f) (e);
  return (a->elements)[l][c];
}

/******************************************************************************/
/* Input/output functions                                                     */
/******************************************************************************/

extern void
gdsl_2darray_write (const gdsl_2darray_t a, 
		    const gdsl_2darray_write_func_t write_f, FILE *file, 
		    void *user_data)
{
  ulong i;
  ulong j;

  assert (a != NULL);
  assert (write_f != NULL);
  assert (file != NULL);

  for (i = 0; i < a->rows; i++)
    {
      write_f ((a->elements)[i][0], file, TRUE, FALSE, user_data);

      for (j = 1; j < a->cols - 1; j++)
	{
	  write_f ((a->elements)[i][j], file, FALSE, FALSE, user_data);
	}

      write_f ((a->elements)[i][j], file, FALSE, TRUE, user_data);
    }
}

extern void
gdsl_2darray_write_xml (const gdsl_2darray_t a, 
			const gdsl_write_func_t write_f, FILE *file, 
			void *user_data)
{
  ulong i;
  ulong j;

  assert (a != NULL);
  assert (file != NULL);

  fprintf (file, "<GDSL_2DARRAY REF=\"%p\" NAME=", a);

  if (a->name == NULL)
    {
      fprintf (file, "\"\"");
    }
  else
    {
      fprintf (file, "\"%s\"", a->name);
    }

  fprintf (file, " ROWS=\"%ld\" COLS=\"%ld\">\n", a->rows, a->cols);

  for (i = 0; i < a->rows; i++)
    {
      fprintf (file, "<GDSL_2DARRAY_ROW VALUE=\"%ld\">\n", i);

      for (j = 0; j < a->cols; j++)
	{
	  fprintf (file, "<GDSL_2DARRAY_COL VALUE=\"%ld\">", j);

	  if (write_f != NULL)
	    {
	      write_f ((a->elements)[i][j], file, user_data);
	    }

	  fprintf (file, "</GDSL_2DARRAY_COL>\n");
	}
      fprintf (file, "</GDSL_2DARRAY_ROW>\n");
    }

  fprintf (file, "</GDSL_2DARRAY>\n");
}

extern void
gdsl_2darray_dump (const gdsl_2darray_t a, const gdsl_write_func_t write_f, 
		   FILE *file, void *user_data)
{
  ulong i;
  ulong j;

  assert (a != NULL);
  assert (file != NULL);

  fprintf (file, "<GDSL_2DARRAY REF=\"%p\" NAME=", a);

  if (a->name == NULL)
    {
      fprintf (file, "\"\"");
    }
  else
    {
      fprintf (file, "\"%s\"", a->name);
    }

  fprintf (file, " ROWS=\"%ld\" COLS=\"%ld\">\n", a->rows, a->cols);

  for (i = 0; i < a->rows; i++)
    {
      fprintf (file, "<GDSL_2DARRAY_ROW VALUE=\"%ld\">\n", i);

      for (j = 0; j < a->cols; j++)
	{
	  if ((a->elements)[i][j] == NULL)
	    {
	      fprintf (file, "<GDSL_2DARRAY_COL VALUE=\"%ld\" CONTENT=\"\">", j);
	    }
	  else
	    {
	      fprintf (file, "<GDSL_2DARRAY_COL VALUE=\"%ld\" CONTENT=\"%p\">", 
		       j, (a->elements)[i][j]);
	    }

	  if (write_f != NULL)
	    {
	      write_f ((a->elements)[i][j], file, user_data);
	    }

	  fprintf (file, "</GDSL_2DARRAY_COL>\n");
	}
      fprintf (file, "</GDSL_2DARRAY_ROW>\n");
    }

  fprintf (file, "</GDSL_2DARRAY>\n");
}

/******************************************************************************/
/* Private functions                                                          */
/******************************************************************************/

static gdsl_element_t default_alloue (void *e)
{
  return e;
}

static void default_detruis (gdsl_element_t e)
{
  ;
}
