/*
 * This file is part of Generic Data Structures Library (GDSL).
 * Copyright (C) 1998-2001 Nicolas Darnis <ndarnis@free.fr>
 *
 * The GDSL library is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * The GDSL library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the GDSL library; see the file COPYING.
 * If not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 * File: gdsl_bstree.c
 * $Revision: 1.5 $
 * $Date: 2002/02/21 09:55:51 $
 * $Author: ndarnis $
 * $Id: gdsl_bstree.c,v 1.5 2002/02/21 09:55:51 ndarnis Exp $
 */

#include <config.h>

#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "_gdsl_bintree.h"
#include "gdsl_bstree.h"
#include "_gdsl_mem.h"

struct gdsl_bstree
{
  char *              name;    /* Tree's name */
  ulong               card;    /* Tree's cardinality */
  _gdsl_bintree_t     sent;    /* Tree's sentinel */
  gdsl_alloc_func_t   alloc_f; /* */
  gdsl_free_func_t    free_f;  /* */
  gdsl_compare_func_t comp_f;  /* */
};

#define LEFT(t)       ( _gdsl_bintree_get_left_subtree( (t) ) )
#define RIGHT(t)      ( _gdsl_bintree_get_right_subtree( (t) ) )
#define CONTENT(t)    ( _gdsl_bintree_get_content( (t) ) )
#define PARENT(t)     ( _gdsl_bintree_get_parent( (t) ) )

#define SENT(t)       ( (t)->sent )
#define ROOT(t)       ( _gdsl_bintree_get_right_subtree( SENT(t) ) )

static gdsl_element_t 
default_alloc (void *e);

static void 
default_free (gdsl_element_t e);

static int 
default_compare (gdsl_element_t e, void *key);

static void 
bstree_destroy (_gdsl_bintree_t n, _gdsl_bintree_t sent, gdsl_free_func_t f);

static ulong
bstree_size (_gdsl_bintree_t n, _gdsl_bintree_t sent);

static ulong
bstree_height (_gdsl_bintree_t n, _gdsl_bintree_t sent);

static _gdsl_bintree_t
bstree_search (_gdsl_bintree_t n, _gdsl_bintree_t sent, gdsl_compare_func_t f,
	       void *v );

static _gdsl_bintree_t
bstree_next (gdsl_bstree_t t, _gdsl_bintree_t n);

static gdsl_element_t
bstree_prefix_parse (_gdsl_bintree_t t, _gdsl_bintree_t sent, gdsl_map_func_t f, 
		     void *d);

static gdsl_element_t
bstree_infix_parse (_gdsl_bintree_t t, _gdsl_bintree_t sent, gdsl_map_func_t f, 
		    void *d);

static gdsl_element_t
bstree_postfix_parse (_gdsl_bintree_t t, _gdsl_bintree_t sent, gdsl_map_func_t f, 
		      void *d);

static void
bstree_write (_gdsl_bintree_t n, _gdsl_bintree_t sent, gdsl_write_func_t f,
	      FILE *file, void *d);

static void
bstree_write_xml (_gdsl_bintree_t n, _gdsl_bintree_t sent, 
		  gdsl_write_func_t f, FILE *file, void *d );
static void
bstree_dump (_gdsl_bintree_t n, _gdsl_bintree_t sent, gdsl_write_func_t f, 
	     FILE *file, void *d);

/******************************************************************************/
/* Allocation and deallocation functions                                      */
/******************************************************************************/

extern gdsl_bstree_t
gdsl_bstree_create (const char *name, const gdsl_alloc_func_t alloc_f, 
		    const gdsl_free_func_t free_f, 
		    const gdsl_compare_func_t comp_f)
{
  gdsl_bstree_t t;

  t = (gdsl_bstree_t) _gdsl_mem_alloc (sizeof (struct gdsl_bstree));
  if (t == NULL)
    {
      return NULL;
    }

  t->sent = _gdsl_bintree_create (NULL);
  if (t->sent == NULL)
    {
      _gdsl_mem_free (t);
      return NULL;
    }

  _gdsl_bintree_set_parent ((_gdsl_bintree_t) (t->sent), (_gdsl_bintree_t) (t->sent));
  _gdsl_bintree_set_left_subtree ((_gdsl_bintree_t) (t->sent), (_gdsl_bintree_t) (t->sent));
  _gdsl_bintree_set_right_subtree ((_gdsl_bintree_t) (t->sent), (_gdsl_bintree_t) (t->sent));

  t->name = NULL;
  if (name != NULL)
    {
      t->name = (char *) _gdsl_mem_alloc ((1 + strlen (name)) * sizeof (char));
      if (t->name == NULL)
	{
	  _gdsl_mem_free (t);
	  return NULL;
	}
      strcpy (t->name, name);
    }

  t->comp_f  = comp_f  ? comp_f  : default_compare;
  t->alloc_f = alloc_f ? alloc_f : default_alloc;
  t->free_f  = free_f  ? free_f  : default_free;

  t->card = 0UL;

  return t;
}

extern void
gdsl_bstree_destroy (gdsl_bstree_t t)
{
  assert (t != NULL);

  bstree_destroy (ROOT (t), SENT (t), t->free_f);

  /*
   * REMARK: 
   * As SENT(t) is allocated, we must deallocate it, so we set
   * its left and right sons to NULL to avoid infinite recursion:
   */
  _gdsl_bintree_set_left_subtree ((_gdsl_bintree_t) SENT (t), NULL);
  _gdsl_bintree_set_right_subtree ((_gdsl_bintree_t) SENT (t), NULL);
  _gdsl_bintree_destroy (SENT (t), NULL);

  if (t->name != NULL)
    {
      _gdsl_mem_free (t->name);
    }

  _gdsl_mem_free (t);
}

extern void
gdsl_bstree_flush (gdsl_bstree_t t)
{
  assert (t != NULL);

  bstree_destroy (ROOT (t), SENT (t), t->free_f);
  _gdsl_bintree_set_left_subtree ((_gdsl_bintree_t) SENT (t), (_gdsl_bintree_t) SENT (t));
  _gdsl_bintree_set_right_subtree ((_gdsl_bintree_t) SENT (t), (_gdsl_bintree_t) SENT (t));
  t->card = 0UL;
}

#if 0
extern gdsl_bstree_t
gdsl_bstree_copy (const gdsl_bstree_t t)
{
  /* !! TO DO... !! */
  return NULL;
}
#endif

extern gdsl_bstree_t
gdsl_bstree_set_name (gdsl_bstree_t t, const char *name)
{
  assert (t != NULL);

  if (t->name != NULL)
    {
      _gdsl_mem_free (t->name);
      t->name = NULL;
    }

  if (name != NULL)
    {
      t->name = (char *) _gdsl_mem_alloc ((1 + strlen (name)) * sizeof (char));
      if (t->name == NULL)
	return NULL;
      strcpy (t->name, name);
    }

  return t;
}

/******************************************************************************/
/* Information functions                                                      */
/******************************************************************************/

extern char *
gdsl_bstree_get_name (const gdsl_bstree_t t)
{
  assert (t != NULL);

  return t->name;
}

extern bool
gdsl_bstree_is_empty (const gdsl_bstree_t t)
{
  assert (t != NULL);

  return (bool) (t->card == 0); /* alt. ROOT( t ) == SENT( t ) */
}

extern gdsl_element_t
gdsl_bstree_get_root (const gdsl_bstree_t t)
{
  assert (t != NULL);

  return CONTENT (ROOT (t));
}

extern ulong
gdsl_bstree_get_size (const gdsl_bstree_t t)
{
  assert (t != NULL);

  return t->card;
}

extern ulong
gdsl_bstree_height (const gdsl_bstree_t t)
{
  assert (t != NULL);

  return bstree_height (ROOT (t), SENT (t));
}

/******************************************************************************/
/* Inserting/removing functions                                               */
/******************************************************************************/

extern gdsl_element_t
gdsl_bstree_insert (gdsl_bstree_t t, void *v)
{
  int comp = -1;
  gdsl_element_t e;
  _gdsl_bintree_t root;
  _gdsl_bintree_t parent;
  _gdsl_bintree_t n;

  assert (t != NULL);

  /* Classic binary search tree insertion: */
  root = ROOT (t);
  parent = SENT (t);
  while (root != SENT (t))
    {
      parent = root;
      comp = t->comp_f (CONTENT (root), v);

      /* Duplicates keys are not allowed (must be changed) */
      if (comp == 0)
	{
	  return NULL;
	}

      root = (comp > 0) ? LEFT (root) : RIGHT (root); 
    }

  /* Then, we create the new node n and we insert it into t: */
  e = (t->alloc_f) (v);
  if (e == NULL)
    {
      return NULL;
    }

  n = _gdsl_bintree_create (e);
  if (n == NULL)
    {
      t->free_f (e);
      return NULL;
    }

  /* Insertion of n into t: */
  _gdsl_bintree_set_parent ((_gdsl_bintree_t) n, (_gdsl_bintree_t) parent);
  _gdsl_bintree_set_left_subtree ((_gdsl_bintree_t) n, (_gdsl_bintree_t) SENT (t));
  _gdsl_bintree_set_right_subtree ((_gdsl_bintree_t) n, (_gdsl_bintree_t) SENT (t));

  if (comp > 0)
    {
      _gdsl_bintree_set_left_subtree ((_gdsl_bintree_t) parent, (_gdsl_bintree_t) n);
    }
  else
    {
      _gdsl_bintree_set_right_subtree ((_gdsl_bintree_t) parent, (_gdsl_bintree_t) n);
    }

  t->card++;

  return e;
}

extern gdsl_element_t
gdsl_bstree_remove_by_value (gdsl_bstree_t t, void *v)
{
  gdsl_element_t e;
  _gdsl_bintree_t n;
  _gdsl_bintree_t child;

  assert (t != NULL);

  /* First we search into t for a node n containing v: */
  n = bstree_search (ROOT (t), SENT (t), t->comp_f, v);
  if (n == NULL)
    {
      return NULL;
    }

  /* Then, we do a classical removing of node n from a binary search tree t: */
  if (LEFT (n) != SENT (t) && RIGHT (n) != SENT (t))
    {
      _gdsl_bintree_t next = bstree_next (t, n);
      _gdsl_bintree_t nextparent = PARENT (next);

      child = RIGHT (next);
      _gdsl_bintree_set_parent (child, nextparent);

      if (LEFT (nextparent) == next)
	{
	  _gdsl_bintree_set_left_subtree (nextparent, child);
	}
      else
	{
	  _gdsl_bintree_set_right_subtree (nextparent, child);
	}

      _gdsl_bintree_set_parent (next, PARENT (n));
      _gdsl_bintree_set_left_subtree (next, LEFT (n));
      _gdsl_bintree_set_right_subtree (next, RIGHT (n));
      _gdsl_bintree_set_parent (LEFT (next), next);
      _gdsl_bintree_set_parent (RIGHT (next), next);
      
      if (LEFT (PARENT (n)) == n)
	{
	  _gdsl_bintree_set_left_subtree (PARENT (n), next);
	}
      else
	{
	  _gdsl_bintree_set_right_subtree (PARENT (n), next);
	}
    }
  else
    {
      child = LEFT (n) != SENT (t) ? LEFT (n) : RIGHT (n);

      _gdsl_bintree_set_parent (child, PARENT (n));

      if (n == LEFT (PARENT (n)))
	{
	  _gdsl_bintree_set_left_subtree (PARENT (n), child);
	}
      else
	{
	  _gdsl_bintree_set_right_subtree (PARENT (n), child);
	}
    }

  t->card--;

  e = CONTENT (n);
  _gdsl_bintree_set_left_subtree ((_gdsl_bintree_t) n, NULL);
  _gdsl_bintree_set_right_subtree ((_gdsl_bintree_t) n, NULL);
  _gdsl_bintree_destroy (n, NULL);

  return e;
}

extern gdsl_bstree_t
gdsl_bstree_destroy_by_value (gdsl_bstree_t t, void *v)
{
  gdsl_element_t e;

  assert (t != NULL);

  e = gdsl_bstree_remove_by_value (t, v);
  if (e == NULL)
    {
      return NULL;
    }

  t->free_f (e);

  return t;
}

/******************************************************************************/
/* Research functions                                                         */
/******************************************************************************/

extern gdsl_element_t
gdsl_bstree_search (const gdsl_bstree_t t, const gdsl_compare_func_t f, void *v)
{
  _gdsl_bintree_t n;

  assert (t != NULL);
 
  n = bstree_search (ROOT (t), SENT (t), f ? f : t->comp_f, v);

  return (n == NULL) ? NULL : CONTENT (n);
}

/******************************************************************************/
/* Parse functions                                                            */
/******************************************************************************/

extern gdsl_element_t
gdsl_bstree_prefix_parse (const gdsl_bstree_t t, const gdsl_map_func_t map_f, 
			  void *d)
{
  assert (t != NULL);
  assert (map_f != NULL);

  return bstree_prefix_parse (ROOT (t), SENT (t), map_f, d);
}

extern gdsl_element_t
gdsl_bstree_infix_parse (const gdsl_bstree_t t, const gdsl_map_func_t map_f, 
			 void *d)
{
  assert (t != NULL);
  assert (map_f != NULL);

  return bstree_infix_parse (ROOT (t), SENT (t), map_f, d);
}

extern gdsl_element_t
gdsl_bstree_postfix_parse (const gdsl_bstree_t t, const gdsl_map_func_t map_f, 
			   void *d)
{
  assert (t != NULL);
  assert (map_f != NULL);

  return bstree_postfix_parse (ROOT (t), SENT (t), map_f, d);
}

#if 0
extern gdsl_element_t
gdsl_bstree_level_parse (const gdsl_bstree_t t, const gdsl_map_func_t map_f, 
			 void *d)
{
  assert (t != NULL);
  assert (map_f != NULL);

  /* !! TO DO... !! */
  return NULL;
}
#endif

/******************************************************************************/
/* Output functions                                                           */
/******************************************************************************/

extern void
gdsl_bstree_write (const gdsl_bstree_t t, const gdsl_write_func_t write_f, 
		   FILE *file, void *d)
{
  assert (t != NULL);
  assert (write_f != NULL);
  assert (file != NULL);

  bstree_write (ROOT (t), SENT (t), write_f, file, d);
}

extern void
gdsl_bstree_write_xml (const gdsl_bstree_t t, const gdsl_write_func_t write_f, 
		       FILE *file, void *d)
{
  assert (t != NULL);
  assert (file != NULL);

  fprintf (file, "<GDSL_BSTREE REF=\"%p\" NAME=\"%s\" CARD=\"%ld\">\n",
	   (void *) t, t->name, t->card);

  bstree_write_xml (ROOT (t), SENT (t), write_f, file, d);

  fprintf (file, "</GDSL_BSTREE>\n");
}

extern void
gdsl_bstree_dump (const gdsl_bstree_t t, const gdsl_write_func_t write_f, 
		  FILE *file, void *d)
{
  assert (t != NULL);
  assert (file != NULL);

  fprintf (file, "<GDSL_BSTREE REF=\"%p\" NAME=\"%s\" CARD=\"%ld\">\n",
	   (void *) t, t->name, t->card);
  fprintf (file, "<GDSL_BSTREE_SENT REF=\"%p\" LEFT=\"%p\" RIGHT=\"%p\" PARENT=\"%p\"/>\n", 
	   (void *) SENT (t), (void *) LEFT (SENT (t)), 
	   (void *) RIGHT (SENT (t)), (void *) PARENT (SENT (t)));

  bstree_dump (ROOT (t), SENT (t), write_f, file, d);

  fprintf (file, "</GDSL_BSTREE>\n");
}

/******************************************************************************/
/* Display functions                                                          */
/******************************************************************************/

/******************************************************************************/
/* Private functions                                                          */
/******************************************************************************/

static gdsl_element_t 
default_alloc (void *e)
{
  return e;
}

static void 
default_free (gdsl_element_t e)
{
  ;
}

static int 
default_compare (gdsl_element_t e, void *key)
{
  return 0;
}

static void 
bstree_destroy (_gdsl_bintree_t n, _gdsl_bintree_t sent, gdsl_free_func_t f)
{
  if (n != sent)
    {
      bstree_destroy (LEFT (n), sent, f);
      bstree_destroy (RIGHT (n), sent, f);
      f (CONTENT (n));
      _gdsl_mem_free (n);
    }
}

static ulong
bstree_size (_gdsl_bintree_t n, _gdsl_bintree_t sent)
{
  if (n == sent)
    return 0UL;

  return (ulong) (1UL 
		  + bstree_size (LEFT (n), sent)
		  + bstree_size (RIGHT (n), sent));
}

static ulong
bstree_height (_gdsl_bintree_t n, _gdsl_bintree_t sent)
{
  if (n == sent)
    return 0UL;

  if (LEFT (n) == sent && RIGHT (n) == sent)
    return 0UL;

  return (ulong) (1UL + 
		  MAX (bstree_height (LEFT (n), sent),
		       bstree_height (RIGHT (n), sent)));
}

static _gdsl_bintree_t
bstree_search (_gdsl_bintree_t root, _gdsl_bintree_t sent, gdsl_compare_func_t f, void *v)
{
  int comp;

#if defined (_GDSL_OPTIMIZE_TIME_)
  while (root != sent)
    {
      comp = f (CONTENT (root), v);

      if (comp == 0)
	return root;

      if (comp > 0)
	root = LEFT (root);
      else
	root = RIGHT (root);
    }

  return NULL;
#else
  if (root == sent)
    return NULL;

  comp = f (CONTENT (root), v);

  if (comp == 0)
    return root;

  if (comp > 0)
    return bstree_search (LEFT (root), sent, f, v);
  
  return bstree_search (RIGHT (root), sent, f, v);
#endif
}

static _gdsl_bintree_t
bstree_next (gdsl_bstree_t t, _gdsl_bintree_t n)
{
  n = RIGHT (n);

  while (LEFT (n) != SENT (t))
    {
      n = LEFT (n);
    }

  return n;
}

static gdsl_element_t
bstree_prefix_parse (_gdsl_bintree_t root, _gdsl_bintree_t sent, gdsl_map_func_t f, void *d)
{
  if (root != sent)
    {
      gdsl_element_t e = CONTENT (root);
      if (f (e, d) == 0) 
	{
	  return e;
	}

      bstree_prefix_parse (LEFT (root), sent, f, d);
      bstree_prefix_parse (RIGHT (root), sent, f, d); 
    }

  return NULL;
}

static gdsl_element_t
bstree_infix_parse (_gdsl_bintree_t root, _gdsl_bintree_t sent, gdsl_map_func_t f, void *d)
{
  if (root != sent)
    {
      gdsl_element_t e;

      bstree_infix_parse (LEFT (root), sent, f, d);

      e = CONTENT (root);
      if (f (e, d) == 0) 
	{
	  return e;
	}

      bstree_infix_parse (RIGHT (root), sent, f, d); 
    }

  return NULL;
}

static gdsl_element_t
bstree_postfix_parse (_gdsl_bintree_t root, _gdsl_bintree_t sent, gdsl_map_func_t f, void *d)
{
  if (root != sent)
    {
      gdsl_element_t e;

      bstree_postfix_parse (LEFT (root), sent, f, d);
      bstree_postfix_parse (RIGHT (root), sent, f, d); 

      e = CONTENT (root);
      if (f (e, d) == 0) 
	{
	  return e;
	}
    }

  return NULL;
}

static void
bstree_write (_gdsl_bintree_t n, _gdsl_bintree_t sent, gdsl_write_func_t f, FILE *file, void *d)
{
  if (n != sent)
    {
      f (CONTENT (n), file, d);
      bstree_write (LEFT (n), sent, f, file, d);
      bstree_write (RIGHT (n), sent, f, file, d);
    }
}

static void
bstree_write_xml (_gdsl_bintree_t n, _gdsl_bintree_t sent, gdsl_write_func_t f, FILE *file, void *d)
{
  if (n != sent)
    {
      if (LEFT (n) == sent && RIGHT (n) == sent)
	{
	  fprintf (file, "<GDSL_BSTREE_LEAF REF=\"%p\"", (void *) n);
	}
      else
	{
	  fprintf (file, "<GDSL_BSTREE_NODE REF=\"%p\"", (void *) n);
	}

      if (LEFT (n) != sent || RIGHT (n) != sent)
	{
	  if (LEFT (n) != sent)
	    {
	      fprintf (file, " LEFT=\"%p\"", (void *) LEFT (n));
	    }
	  else
	    {
	      fprintf (file, " LEFT=\"\"");
	    }
	  
	  if (RIGHT (n) != sent)
	    {
	      fprintf (file, " RIGHT=\"%p\"", (void *) RIGHT (n));
	    }
	  else
	    {
	      fprintf (file, " RIGHT=\"\"");
	    }
	}

      if (PARENT (n) != sent)
	{
	  fprintf (file, " PARENT=\"%p\">", (void *) PARENT (n));
	}
      else
	{
	  fprintf (file, " PARENT=\"\">");
	}

      if (f != NULL)
	{
	  f (CONTENT (n), file, d);
	}

      if (LEFT (n) == sent && RIGHT (n) == sent)
	{
	  fprintf (file, "</GDSL_BSTREE_LEAF>\n");
	}
      else
	{
	  fprintf (file, "</GDSL_BSTREE_NODE>\n");
	}

      bstree_write_xml (LEFT (n), sent, f, file, d);
      bstree_write_xml (RIGHT (n), sent, f, file, d);
    }
}

static void
bstree_dump (_gdsl_bintree_t n, _gdsl_bintree_t sent, gdsl_write_func_t f, FILE *file, void *d)
{
  if (n != sent)
    {
      fprintf (file, "<GDSL_BSTREE_NODE REF=\"%p\"", (void *) n);

      if (CONTENT (n))
	{
	  fprintf (file, " CONTENT=\"%p\"", (void *) CONTENT (n));
	}
      else
	{
	  fprintf (file, " CONTENT=\"\"");
	}

      fprintf (file, " LEFT=\"%p\" RIGHT=\"%p\"", (void *) LEFT (n), (void *) RIGHT (n));

      if (PARENT (n))
	{
	  fprintf (file, " PARENT=\"%p\">", (void *) PARENT (n));
	}
      else
	{
	  fprintf (file, " PARENT=\"\">");
	}

      if (f != NULL)
	{
	  f (CONTENT (n), file, d);
	}

      fprintf (file, "</GDSL_BSTREE_NODE>\n");

      bstree_dump (LEFT (n), sent, f, file, d);
      bstree_dump (RIGHT (n), sent, f, file, d);
    }
}

#ifdef HAVE_GFX_DEBUG
#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#define X_Win_pos 0
#define Y_Win_pos 0

#define RAYON 6
#define BORD  3 * RAYON
#define DX  3 * RAYON
#define DY 6 * RAYON

#define COLOR1   "grey"
#define COLOR2   "blue"

/* declaration des identificateurs de ressources X */
Window Fen;                      /* la fenetre Globale    */
GC     ContGraph;                /* contexte graphique    */

/* declaration des variables utilisees par les fonctions X */
Display                *display;
XEvent                 evenement;            /* file des evenements */
XSizeHints             indic;
int                    ecran;
unsigned long          noir,blanc;
Colormap               Cmap;
XColor                 coul1;   /* meilleure approximation RVB rendue */
XColor                 coul2;
XColor                 exact_c; /* couleur RVB exacte correspondant au nom */
Font                   fonte;

/*Declaration des variabes propres a l'arbre */
int Xo,Yo,X,Y;
static bool InitContexte( int LFen, int HFen );
static void CloseFen( void );
static bool Touche( KeySym tch );
static void TraceABic( _gdsl_bintree_t A, _gdsl_bintree_t sent, int x, int y, char * (*key_display)(gdsl_element_t) );
#endif /* HAVE_GFX_DEBUG */

extern void 
gdsl_bstree_gfx_display (const gdsl_bstree_t A, const char * (*key_display) (gdsl_element_t))
{
#ifdef HAVE_GFX_DEBUG
  int encore = TRUE;
  int Lfen, Hfen;      
  KeySym touche;
  _gdsl_bintree_t t = ROOT(A);
  _gdsl_bintree_t sl = LEFT( SENT( A ) );
  _gdsl_bintree_t sr = RIGHT( SENT( A ) );

  _gdsl_bintree_set_left_subtree( (_gdsl_bintree_t)SENT( A ), NULL );
  _gdsl_bintree_set_right_subtree( (_gdsl_bintree_t)SENT( A ), NULL );

  if( t != SENT(A) ) 
    {
      Lfen = ( A->card - 1 ) * DX + 2 * BORD;  /* largeur de la Fen Globale     */
      Hfen = bstree_height( t, SENT( A ) ) * DY + 2 * BORD;     /* Hauteur de la Fen GLOBALE     */
      Xo = bstree_size( LEFT( t ), SENT( A ) ) * DX + BORD;   /* abscisse de la racine de l'arbre  */
      X = Xo; Yo = BORD;Y = Yo;
      if ( InitContexte( Lfen, Hfen ) )
	{   
	  /* BOUCLE PRINCIPALE */
	  while( encore )
	    {
	      /*lecture de la file d'attente des evenements */
	      XNextEvent( display, &evenement );
	      switch( evenement.type )
		{
		case Expose: 
		  /* apres le dernier evenement du type expose(affichage complet de la Fen) */
		  if( evenement.xexpose.count == 0 )
		    TraceABic( t, SENT( A ), X, Y, key_display );
		  break;

		case KeyPress:
		  touche = XLookupKeysym( (XKeyEvent *)&evenement, 0 );
		  if( ( encore = Touche( touche ) ) )
		    {
		      XClearArea(display, Fen, 0, 0, 0, 0, TRUE );
		      TraceABic( t, SENT( A ), X, Y, key_display );
		    }
		}
	    }
	}
      CloseFen(); 
    }
  _gdsl_bintree_set_left_subtree( (_gdsl_bintree_t)SENT( A ), sl );
  _gdsl_bintree_set_right_subtree( (_gdsl_bintree_t)SENT( A ), sr );
#endif /* HAVE_GFX_DEBUG */
}

#ifdef HAVE_GFX_DEBUG
static bool 
InitContexte( int LFen, int HFen )
{
  /* demande dans la base de donnee          */
  /* connexion au serveur X                  */
  display = XOpenDisplay( NULL );

  if ( display == NULL ) 
    {
      printf( "Impossible d'ouvrir le display.\n" );
      return( FALSE );
    }
  
  /* consultation des caracteristiques du serveur X      */
  ecran = DefaultScreen( display );
  blanc = WhitePixel( display, ecran );
  noir = BlackPixel( display, ecran );
  
  /* definition de deux couleurs                         */
  Cmap = DefaultColormap( display, DefaultScreen( display ) );
  if( XAllocNamedColor( display, Cmap, COLOR1, &coul1, &exact_c ) == 0 )
    coul1.pixel = noir;
  
  if( XAllocNamedColor( display, Cmap, COLOR2, &coul2, &exact_c ) == 0 )
    coul2.pixel=blanc;
  
  /* Creation de la Fen top-niveau du nom de "GLOBALE" */
  indic.x = X_Win_pos;
  indic.y = Y_Win_pos;
  indic.width = LFen;
  indic.height = HFen;
  indic.flags = PPosition | PSize;
  
  Fen = XCreateSimpleWindow( display, DefaultRootWindow( display ),
			     indic.x, indic.y, indic.width, indic.height,
			     4, coul2.pixel, coul1.pixel );
  
  /* definition des proprietes standard de la Fen Globale pour gestionnaire X */
  XSetStandardProperties( display, Fen, "Arbre bicolore", "Arbre bicolore", None, NULL, 0, &indic );
  
  /* Choix des evenements qui nous interressent en ce qui concerne la Fen */
  XSelectInput( display, Fen, KeyPressMask | ExposureMask ); 
  
  /* projection de toutes la Fen pour la rendre visible            */
  /* l'action "visible" est dans le tampon, mais non encore effectuee  */
  /* les actions s'effectuent avec la fonction XNextEvent, par exemple */
  XMapRaised( display, Fen );

  /* Creation du contexte graphique necessaire pour les requetes d'affichage  */
  ContGraph = XCreateGC( display, Fen, 0, NULL );
  XSetBackground( display, ContGraph, coul1.pixel) ;
  XSetForeground( display, ContGraph, coul2.pixel );

  /* Selection de la police */
  fonte = XLoadFont( display, "6x10" );
  XSetFont( display, ContGraph, fonte );

  return TRUE;
}  

/*----------------------------*/

static void 
CloseFen( void )
{
  /* liberation des ressources de la Fen */
  XFreeGC( display, ContGraph );
  XDestroyWindow( display, Fen );
  XCloseDisplay( display );
}

/*----------------------------*/

static bool 
Touche( KeySym tch )
{
  switch( tch )
    {
    case XK_Up: 
      Y += 100;   
      break;
      
    case XK_Right:
      X -= 100;   
      break;

    case XK_Down:
      Y -= 100;   
      break;

    case XK_Left:
      X += 100;   
      break;
      
    case XK_Return:
      X = Xo;
      Y = Yo;
      break;

    case XK_space:
      return FALSE;
      break; 
    }

  return TRUE;
}

/*------------------------------*/

static void 
TraceABic( _gdsl_bintree_t A, _gdsl_bintree_t sent, int x, int y, char * (*key_display)(gdsl_element_t) )
{
  int Xg, Xd, Ygd;
  char chr[35];
  
  if( x >= 0 && y >=0 )
    { 
      /* condition lie a des problemes d'affichage hors fenetre... */
      XSetForeground(display, ContGraph, coul2.pixel); 
      snprintf( chr, 34, "%s", key_display( CONTENT( A ) ) );
      XDrawString(display, Fen, ContGraph,x-RAYON ,y+RAYON ,chr ,strlen(chr));
    }
  
  Ygd = y + DY;
  
  if( LEFT(A) != sent )
    {
      Xg = x - ( 1 + bstree_size( RIGHT( LEFT( A ) ), sent ) ) * DX;
      XSetLineAttributes(display, ContGraph, 1, LineSolid, CapNotLast, JoinMiter);
      XSetForeground(display, ContGraph, noir);
      XDrawLine(display, Fen, ContGraph, x-RAYON, y+RAYON, Xg+RAYON, Ygd-RAYON );
      TraceABic( LEFT( A ), sent, Xg, Ygd, key_display );
    }

  if ( RIGHT( A ) != sent )
    {
      Xd = x + ( 1 + bstree_size( LEFT( RIGHT( A ) ), sent ) ) * DX;
      XSetLineAttributes(display, ContGraph, 1, LineSolid, CapNotLast, JoinMiter);
      XSetForeground(display, ContGraph, noir);
      XDrawLine(display, Fen, ContGraph, x+RAYON, y+RAYON, Xd-RAYON, Ygd-RAYON);
      TraceABic( RIGHT( A ), sent, Xd, Ygd, key_display );
    }
}
#endif /* HAVE_GFX_DEBUG */

