/*
 * This file is part of Generic Data Structures Library (GDSL).
 * Copyright (C) 1998-2001 Nicolas Darnis <ndarnis@free.fr>
 *
 * The GDSL library is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * The GDSL library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the GDSL library; see the file COPYING.
 * If not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 * File: gdsl_hash.c
 * $Revision: 1.3 $
 * $Date: 2002/02/21 09:55:51 $
 * $Author: ndarnis $
 * $Id: gdsl_hash.c,v 1.3 2002/02/21 09:55:51 ndarnis Exp $
 */

#include <config.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "_gdsl_mem.h"
#include "gdsl_hash.h"
#include "gdsl_list.h"

struct hash_table 
{
  char *            ht_name;
  ulong             ht_number_element;
  gdsl_list_t *     ht_table;
  gdsl_key_func_t   ht_key_func;
  gdsl_hash_func_t  ht_hash_func;
  gdsl_alloc_func_t ht_alloc_func;
  gdsl_free_func_t  ht_free_func;
};

typedef struct hash_element 
{
  ulong key;
  gdsl_element_t content;
} *hash_element;

typedef void * (*hash_fct_ptr_t) (void *, void *, void *);

struct infos 
{
  hash_fct_ptr_t f;
  void *ud;
  void *d;
};

static gdsl_element_t 
default_alloc (void *e);

static void 
default_free (gdsl_element_t e);

static ulong
default_key (gdsl_element_t e);

static ulong
default_hash (ulong k);

static int 
search_element_by_key (gdsl_element_t e, void *key);

static int 
local_map_f (gdsl_element_t e, void *user_data);

static void
local_dump_f (gdsl_element_t e, FILE *file, void *user_data);

static int 
destroy_element (gdsl_element_t e, void *user_infos);

/******************************************************************************/
/* Management functions                                                       */
/******************************************************************************/

extern gdsl_hash_t
gdsl_hash_create (const char *name, gdsl_key_func_t key_func, 
		  gdsl_alloc_func_t alloc_func, gdsl_free_func_t free_func,
		  gdsl_hash_func_t hash_func, ulong size)
{
  gdsl_hash_t h;

  h = (gdsl_hash_t) _gdsl_mem_alloc (sizeof (struct hash_table));
  if (h == NULL)
    return NULL;

  h->ht_name = NULL;

  if (name != NULL)
    {
      h->ht_name = (char *) _gdsl_mem_alloc ((1 + strlen (name)) * sizeof (char));
      if (h->ht_name == NULL)
	{
	  _gdsl_mem_free (h);
	  return NULL;
	}
      strcpy (h->ht_name, name);
    }

  h->ht_table = (gdsl_list_t *) _gdsl_mem_alloc (size * sizeof (gdsl_list_t));
  if (h->ht_table == NULL)
    {
      _gdsl_mem_free (h->ht_name);
      _gdsl_mem_free (h);
      return NULL;
    }

  {
    ulong i;

    for (i = 0; i < size; i++)
      {
	h->ht_table[i] = gdsl_list_create (NULL, NULL, NULL);
	if (h->ht_table[i] == NULL)
	  {
	    ulong j;

	    for (j = 0; j < i; j++)
	      gdsl_list_destroy (h->ht_table[j]);

	    if (h->ht_name != NULL)
	      _gdsl_mem_free (h->ht_name);

	    _gdsl_mem_free (h->ht_table);
	    _gdsl_mem_free (h);
	    return NULL;
	  }
      }
  }

  h->ht_number_element = size;
  h->ht_key_func       = key_func   ? key_func   : default_key;
  h->ht_hash_func      = hash_func  ? hash_func  : default_hash;
  h->ht_alloc_func     = alloc_func ? alloc_func : default_alloc;
  h->ht_free_func      = free_func  ? free_func  : default_free;

  return h;
}

extern void
gdsl_hash_destroy (gdsl_hash_t h)
{
  ulong i;
  
  for (i = 0; i < h->ht_number_element; i++)
    {
      gdsl_list_map_forward (h->ht_table[i], destroy_element, (void *) h);
      gdsl_list_destroy (h->ht_table[i]);
    }

  if (h->ht_name != NULL)
    _gdsl_mem_free (h->ht_name);

  _gdsl_mem_free (h->ht_table);
  _gdsl_mem_free (h);
}

extern void
gdsl_hash_flush (gdsl_hash_t h)
{
  ulong i;
  
  for (i = 0; i < h->ht_number_element; i++)
    {
      gdsl_list_map_forward (h->ht_table[i], destroy_element, (void *) h);
      gdsl_list_destroy (h->ht_table[i]);
      h->ht_table[i] = gdsl_list_create (NULL, NULL, NULL);
    }
}

extern gdsl_hash_t
gdsl_hash_set_name( gdsl_hash_t h, const char *name )
{
  if (h->ht_name != NULL)
    {
      _gdsl_mem_free (h->ht_name);
      h->ht_name = NULL;
    }

  if (name != NULL)
    {
      h->ht_name = (char *) _gdsl_mem_alloc ((1 + strlen (name)) * sizeof (char));

      if (h->ht_name == NULL)
	return NULL;

      strcpy (h->ht_name, name);
    }

  return h;
}

/******************************************************************************/
/* Information functions                                                      */
/******************************************************************************/

extern const char *
gdsl_hash_get_name (gdsl_hash_t h)
{
  return h->ht_name;
}

extern double
gdsl_hash_fill_factor (gdsl_hash_t h)
{
  ulong i;
  ulong n;

  n = 0;
  for (i = 0; i < h->ht_number_element; i++)
    n += gdsl_list_get_card (h->ht_table[i]);

  return (double) (n / h->ht_number_element);
}

/******************************************************************************/
/* Insert functions                                                           */
/******************************************************************************/

extern gdsl_element_t
gdsl_hash_insert (gdsl_hash_t h, void *e)
{
  gdsl_list_t   l;
  hash_element he;

  he = (hash_element) _gdsl_mem_alloc (sizeof (struct hash_element));
  if (he == NULL)
    return NULL;

  he->content = (h->ht_alloc_func) (e);
  if (he->content == NULL)
    {
      _gdsl_mem_free (he);
      return NULL;
    }

  he->key = (h->ht_key_func) (e);
  l = h->ht_table[(h->ht_hash_func) (he->key)];

  {
    gdsl_element_t tmp;

    tmp = gdsl_list_insert_head (l, he);
    if (tmp == NULL)
      {
	h->ht_free_func (he->content);
	_gdsl_mem_free (he);
	return NULL;
      }
    }

  return he->content;
}

/******************************************************************************/
/* Remove functions                                                           */
/******************************************************************************/

extern gdsl_element_t
gdsl_hash_remove (gdsl_hash_t h, void *key)
{
  ulong         k;
  hash_element  he;
  gdsl_element_t e;
  gdsl_list_t    l;

  k = (h->ht_key_func) (key);
  l = h->ht_table[(h->ht_hash_func) (k)];

  he = (hash_element) gdsl_list_remove_by_value (l, search_element_by_key, (void *) &k);
  if (he == NULL)
    return NULL;

  e = he->content;
  _gdsl_mem_free (he);

  return e;
}

/******************************************************************************/
/* Search functions                                                           */
/******************************************************************************/

extern gdsl_element_t
gdsl_hash_search_by_key (gdsl_hash_t h, void *key)
{
  ulong k;
  gdsl_list_t l;

  k = (h->ht_key_func) (key);
  l = h->ht_table[(h->ht_hash_func) (k)];

  return gdsl_list_search_by_value (l, search_element_by_key, (void *) &k);
}

/******************************************************************************/
/* Parse functions                                                            */
/******************************************************************************/

extern void
gdsl_hash_map (gdsl_hash_t h, gdsl_map_func_t f, void *user_data)
{
  ulong        i;
  struct infos infos;

  infos.f = (hash_fct_ptr_t) f;
  infos.d = user_data;

  for (i = 0; i < h->ht_number_element; i++)
    if (gdsl_list_get_card (h->ht_table[i]) != 0)
      if (gdsl_list_map_forward (h->ht_table[i], local_map_f, (void *) &infos) != NULL)
	return;
}

/******************************************************************************/
/* Dumping functions                                                          */
/******************************************************************************/

extern void
gdsl_hash_write (gdsl_hash_t h, gdsl_write_func_t f, FILE *file, void *user_data)
{
  ulong        i;
  struct infos infos;

  infos.f = (hash_fct_ptr_t) f;
  infos.d = (void *) file;
  infos.ud = user_data;

  for (i = 0; i < h->ht_number_element; i++)
    if (gdsl_list_get_card (h->ht_table[i]) != 0)
      gdsl_list_write (h->ht_table[i], local_dump_f, file, (void *) &infos);
}

extern void
gdsl_hash_write_xml (gdsl_hash_t h, gdsl_write_func_t f, FILE *file, void *user_data)
{
  ulong        i;
  struct infos infos;

  infos.f = (hash_fct_ptr_t) f;
  infos.d = (void *) file;
  infos.ud = user_data;

  fprintf (file, "<GDSL_HASH REF=\"%p\" NAME=\"%s\">\n",
	   (void *) h, h->ht_name ? h->ht_name : "");

  for (i = 0; i < h->ht_number_element; i++)
    if (gdsl_list_get_card (h->ht_table[i]) != 0)
      gdsl_list_write_xml (h->ht_table[i], local_dump_f, file, (void *) &infos);

  fprintf (file, "</GDSL_HASH>\n");
}

extern void
gdsl_hash_dump (gdsl_hash_t h, gdsl_write_func_t f, FILE *file, void *user_data)
{
  ulong i;
  struct infos infos;

  infos.f = (hash_fct_ptr_t) f;
  infos.d = file;
  infos.ud = user_data;

  fprintf (file, "<GDSL_HASH REF=\"%p\" NAME=\"%s\">\n",
	   (void *) h, h->ht_name ? h->ht_name : "");

  for (i = 0; i < h->ht_number_element; i++)
    if (gdsl_list_get_card( h->ht_table[i]) != 0)
      gdsl_list_dump (h->ht_table[i], local_dump_f, file, (void *) &infos);
  
  fprintf (file, "</GDSL_HASH>\n");
}

/******************************************************************************/
/* Private functions                                                          */
/******************************************************************************/

static gdsl_element_t 
default_alloc (void *e)
{
  return e;
}

static void 
default_free (gdsl_element_t e)
{
  ;
}

static ulong
default_key (gdsl_element_t e)
{
  int i, j;
  ulong n = 0;
  char *chaine = (char *) e;

  for (j = 0, i = strlen (chaine) - 1; i >= 0; i--, j++)
    {
      char c = chaine[i];
      n += c * pow (128, j);
    }

  return n;
}

static ulong
default_hash (ulong k)
{
  return k % 701;
}

static int 
search_element_by_key (gdsl_element_t e, void *key)
{
  hash_element el = (hash_element) e;
  ulong *k = (ulong *) key;
  return el->key - *k;
}

static int 
local_map_f (gdsl_element_t e, void *user_data)
{
  hash_element he = (hash_element) e;
  struct infos *infos = (struct infos *) user_data;
  gdsl_map_func_t map = (gdsl_map_func_t) (infos->f);
  map (he->content, infos->d);
  return 1;
}

static void
local_dump_f (gdsl_element_t e, FILE *file, void *user_data)
{
  hash_element he = (hash_element) e;
  struct infos *infos = (struct infos *) user_data;
  (infos->f) (he->content, file, infos->ud);
}

static int
destroy_element (gdsl_element_t e, void *user_infos)
{
  gdsl_hash_t h = (gdsl_hash_t) user_infos;
  hash_element he = (hash_element) e;
  (h->ht_free_func) (he->content);
  _gdsl_mem_free (he);
  return 1;
}

