/*
 * This file is part of Generic Data Structures Library (GDSL).
 * Copyright (C) 1998-2001 Nicolas Darnis <ndarnis@free.fr>
 *
 * The GDSL library is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * The GDSL library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the GDSL library; see the file COPYING.
 * If not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 * File: gdsl_list.c
 * $Revision: 1.6 $
 * $Date: 2002/02/21 09:55:51 $
 * $Author: ndarnis $
 * $Id: gdsl_list.c,v 1.6 2002/02/21 09:55:51 ndarnis Exp $
 */

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "_gdsl_list_node.h"
#include "_gdsl_list.h"
#include "gdsl_types.h"
#include "gdsl_list.h"
#include "_gdsl_mem.h"

struct _gdsl_list
{
  _gdsl_list_node_t d;       /* Start of the list (sentinel)     */
  _gdsl_list_node_t z;       /* End of the list (sentinel)       */
  _gdsl_list_node_t c;       /* Cursor                           */
  char *           name;     /* Name of the list                 */
  ulong            card;     /* Cardinality of the list          */

  /* Functions pointers */
  gdsl_alloc_func_t alloc_f; /* Alloc element function pointer   */
  gdsl_free_func_t  free_f;  /* Dealloc element function pointer */
};

static gdsl_element_t
default_alloc (void *e);

static void 
default_free (gdsl_element_t e);

static _gdsl_list_node_t
search_by_function (gdsl_list_t l, gdsl_compare_func_t f, void *v);

static _gdsl_list_node_t
search_by_position (gdsl_list_t l, ulong i);

static gdsl_element_t
update_cursor (gdsl_list_t l, _gdsl_list_node_t n);

static _gdsl_list_node_t
sort (_gdsl_list_node_t u, gdsl_compare_func_t f, _gdsl_list_node_t z);

static _gdsl_list_node_t
merge (_gdsl_list_node_t s, _gdsl_list_node_t t, gdsl_compare_func_t f, 
       _gdsl_list_node_t z);

/******************************************************************************/
/* Management functions                                                       */
/******************************************************************************/

extern gdsl_list_t
gdsl_list_create (const char *name, gdsl_alloc_func_t alloc_f, 
		 gdsl_free_func_t free_f)
{
  gdsl_list_t l;
  
  l = (gdsl_list_t) _gdsl_mem_alloc (sizeof (struct _gdsl_list));
  if (l == NULL)
    {
      return NULL;
    }

  l->d = _gdsl_list_node_alloc ();
  if (l->d == NULL)
    {
      _gdsl_mem_free (l);
      return NULL;
    }

  l->z = _gdsl_list_node_alloc ();
  if (l->z == NULL)
    {
      _gdsl_list_node_free (l->d);
      _gdsl_mem_free (l);
      return NULL;
    }
  
  l->name = NULL;

  if (name != NULL)
    {
      l->name = (char *) _gdsl_mem_alloc ((1 + strlen (name)) * sizeof (char));

      if (l->name == NULL)
	{
	  _gdsl_list_node_free (l->z);
	  _gdsl_list_node_free (l->d);
	  _gdsl_mem_free (l);
	  return NULL;
	}

      strcpy (l->name, name);
    }

  _gdsl_list_link (l->d, l->z);
  _gdsl_list_node_set_succ (l->z, l->z);
  _gdsl_list_node_set_pred (l->d, l->d);

  l->c = l->d;
  l->card = 0UL;
  l->alloc_f = alloc_f ? alloc_f : default_alloc;
  l->free_f  = free_f  ? free_f  : default_free;

  return l;
}

extern void 
gdsl_list_destroy (gdsl_list_t l)
{
  if (!gdsl_list_is_empty (l))
    {
      gdsl_list_flush (l);
    }

  _gdsl_list_node_free (l->d);
  _gdsl_list_node_free (l->z);

  if (l->name != NULL)
    {
      _gdsl_mem_free (l->name);
    }

  _gdsl_mem_free (l);
}

extern void
gdsl_list_flush (gdsl_list_t l)
{
  _gdsl_list_node_t save;
  _gdsl_list_node_t tmp;

  tmp = _gdsl_list_node_get_succ (l->d);
  while (tmp != l->z)
    {
      save = _gdsl_list_node_get_succ (tmp);
      l->free_f (_gdsl_list_node_get_content (tmp));
      _gdsl_list_node_free (tmp);
      tmp = save;
    }

  _gdsl_list_link (l->d, l->z);
  _gdsl_list_node_set_succ (l->z, l->z);
  _gdsl_list_node_set_pred (l->d, l->d);

  l->c = l->d;
  l->card = 0UL;
}

extern gdsl_list_t
gdsl_list_set_name (gdsl_list_t l, const char *name)
{
  if (l->name != NULL)
    {
      _gdsl_mem_free (l->name);
      l->name = NULL;
    }

  if (name != NULL)
    {
      l->name = (char *) _gdsl_mem_alloc ((1 + strlen (name)) * sizeof (char));
      if (l->name == NULL)
	{
	  return NULL;
	}
      strcpy (l->name, name);
    }

  return l;
}

/******************************************************************************/
/* Information functions                                                      */
/******************************************************************************/

extern char *
gdsl_list_get_name (gdsl_list_t l)
{
  return l->name;
}

extern ulong
gdsl_list_get_card (gdsl_list_t l)
{
  return l->card;
}

extern bool
gdsl_list_is_empty (gdsl_list_t l)
{
  return (bool) (_gdsl_list_node_get_succ (l->d) == l->z);
}

extern gdsl_element_t
gdsl_list_get_head (gdsl_list_t l)
{
  return _gdsl_list_node_get_content (_gdsl_list_node_get_succ (l->d));
}

extern gdsl_element_t
gdsl_list_get_tail (gdsl_list_t l)
{
  return _gdsl_list_node_get_content (_gdsl_list_node_get_pred (l->z));
}

/******************************************************************************/
/* Insert functions                                                           */
/******************************************************************************/

extern gdsl_element_t 
gdsl_list_insert_head (gdsl_list_t l, void *v)
{
  gdsl_element_t    e;
  _gdsl_list_node_t head;

  head = _gdsl_list_node_alloc ();
  if (head == NULL)
    {
      return NULL;
    }

  e = (l->alloc_f) (v);
  if (e == NULL)
    {
      _gdsl_list_node_free (head);
      return NULL;
    }

  l->card++;

  _gdsl_list_node_set_content (head, e);  
  _gdsl_list_link (head, _gdsl_list_node_get_succ (l->d));
  _gdsl_list_link (l->d, head);

  return e;
}

extern gdsl_element_t
gdsl_list_insert_tail (gdsl_list_t l, void *v)
{
  gdsl_element_t    e;
  _gdsl_list_node_t tail;

  tail = _gdsl_list_node_alloc ();
  if (tail == NULL)
    {
      return NULL;
    }

  e = (l->alloc_f) (v);
  if (e == NULL)
    {
      _gdsl_list_node_free (tail);
      return NULL;
    }

  l->card++;

  _gdsl_list_node_set_content (tail, e);
  _gdsl_list_link (_gdsl_list_node_get_pred (l->z), tail);
  _gdsl_list_link (tail, l->z);

  return e;
}

/******************************************************************************/
/* Remove functions                                                           */
/******************************************************************************/

extern gdsl_element_t
gdsl_list_remove_head (gdsl_list_t l)
{
  if (!gdsl_list_is_empty (l))
  {
    _gdsl_list_node_t head = _gdsl_list_node_get_succ (l->d);
    gdsl_element_t    e    = _gdsl_list_node_get_content (head);

    _gdsl_list_remove (head);
    _gdsl_list_node_free (head);

    l->card--;
    return e;
  }

  return NULL;
}

extern gdsl_element_t
gdsl_list_remove_tail (gdsl_list_t l)
{
  if (!gdsl_list_is_empty (l))
    {
      _gdsl_list_node_t tail = _gdsl_list_node_get_pred (l->z);
      gdsl_element_t e = _gdsl_list_node_get_content (tail);

      _gdsl_list_remove (tail);
      _gdsl_list_node_free (tail);

      l->card--;
      return e;
    }

  return NULL;
}

extern gdsl_element_t
gdsl_list_remove_by_value (gdsl_list_t l, gdsl_compare_func_t f, void *v)
{
  gdsl_element_t    e;
  _gdsl_list_node_t n;

  n = search_by_function (l, f, v);
  if (n == NULL)
    {
      return NULL;
    }

  e = _gdsl_list_node_get_content (n);

  _gdsl_list_remove (n);
  _gdsl_list_node_free (n);

  l->card--;

  return e;
}

/******************************************************************************/
/* Destroy functions                                                          */
/******************************************************************************/

extern gdsl_list_t
gdsl_list_destroy_head (gdsl_list_t l)
{
  gdsl_element_t e;

  e = gdsl_list_remove_head (l);
  if (e == NULL)
    return NULL;

  l->free_f (e);

  return l;
}

extern gdsl_list_t
gdsl_list_destroy_tail (gdsl_list_t l)
{
  gdsl_element_t e;

  e = gdsl_list_remove_tail (l);
  if (e == NULL)
    return NULL;

  l->free_f (e);

  return l;
}

extern gdsl_list_t
gdsl_list_destroy_by_value (gdsl_list_t l, gdsl_compare_func_t f, void *v)
{
  gdsl_element_t e;

  e = gdsl_list_remove_by_value (l, f, v);
  if (e == NULL)
    return NULL;

  l->free_f (e);

  return l;
}

/******************************************************************************/
/* Search functions                                                           */
/******************************************************************************/

extern gdsl_element_t
gdsl_list_search_by_value (gdsl_list_t l, gdsl_compare_func_t f, void *v)
{
  _gdsl_list_node_t n;

  n = search_by_function (l, f, v);

  return n ? _gdsl_list_node_get_content (n) : NULL;
}

extern gdsl_element_t
gdsl_list_search_by_position (gdsl_list_t l, ulong p)
{
  _gdsl_list_node_t n;

  n = search_by_position (l, p);

  return n ? _gdsl_list_node_get_content (n) : NULL;
}

extern gdsl_element_t
gdsl_list_search_max (gdsl_list_t l, gdsl_compare_func_t f)
{
  _gdsl_list_node_t tmp;
  gdsl_element_t    max;

  tmp = _gdsl_list_node_get_succ (l->d);
  max = _gdsl_list_node_get_content (tmp);

  while (tmp != l->z)
    {
      gdsl_element_t e = _gdsl_list_node_get_content (tmp);
      if (f (e, max) > 0)
	max = e;
      tmp = _gdsl_list_node_get_succ (tmp);
    }

  return max;
}

extern gdsl_element_t
gdsl_list_search_min (gdsl_list_t l, gdsl_compare_func_t f)
{
  _gdsl_list_node_t tmp;
  gdsl_element_t    min;

  tmp = _gdsl_list_node_get_succ (l->d);
  min = _gdsl_list_node_get_content (tmp);

  while (tmp != l->z)
    {
      gdsl_element_t e = _gdsl_list_node_get_content (tmp);
      if (f (e, min) < 0)
	min = e;
      tmp = _gdsl_list_node_get_succ (tmp);
    }

  return min;
}

/******************************************************************************/
/* Sort functions                                                             */
/******************************************************************************/

extern gdsl_list_t
gdsl_list_merge_sort (gdsl_list_t l, gdsl_compare_func_t f, gdsl_element_t max)
{

  if (gdsl_list_get_card (l) <= 1)
    {
      return l;
    }

  _gdsl_list_node_set_content (l->z, max);
  _gdsl_list_node_set_succ (l->d, sort (_gdsl_list_node_get_succ (l->d ), f, l->z));
  _gdsl_list_node_set_content (l->z, NULL);

  return l;
}

/******************************************************************************/
/* Parse functions                                                            */
/******************************************************************************/

extern gdsl_element_t 
gdsl_list_map_forward (gdsl_list_t l, gdsl_map_func_t f, void *user_data)
{
  gdsl_element_t    e;
  _gdsl_list_node_t tmp;

  tmp = _gdsl_list_node_get_succ (l->d);
  while (tmp != l->z)
    {
      e = _gdsl_list_node_get_content (tmp);

      if (f (e, user_data) == 0)
	return e;

      tmp = _gdsl_list_node_get_succ (tmp);
    }

  return NULL;
}

extern gdsl_element_t 
gdsl_list_map_backward (gdsl_list_t l, gdsl_map_func_t f, void *user_data)
{
  gdsl_element_t    e;
  _gdsl_list_node_t tmp;

  tmp = _gdsl_list_node_get_pred (l->z);
  while (tmp != l->d)
    {
      e = _gdsl_list_node_get_content (tmp);

      if (f (e, user_data) == 0)
	return e;

      tmp = _gdsl_list_node_get_pred (tmp);
    }

  return NULL;
}

/******************************************************************************/
/* Input/output functions                                                     */
/******************************************************************************/

extern void
gdsl_list_write (const gdsl_list_t l, const gdsl_write_func_t f, 
		 FILE *file, void *user_data)
{
  _gdsl_list_node_t tmp;

  tmp = _gdsl_list_node_get_succ (l->d);
  while (tmp != l->z)
    {
      _gdsl_list_node_write (tmp, f, file, user_data);
      tmp = _gdsl_list_node_get_succ (tmp);
    }
}

extern void
gdsl_list_write_xml (const gdsl_list_t l, const gdsl_write_func_t f, 
		     FILE *file, void *user_data)
{
  _gdsl_list_node_t tmp;

  tmp = _gdsl_list_node_get_succ (l->d);

  fprintf (file, "<GDSL_LIST REF=\"%p\" NAME=\"%s\" CARD=\"%ld\" HEAD=\"%p\" TAIL=\"%p\"", 
	   (void *) l, l->name ? l->name : "", l->card, (void *) tmp, 
	   (void *) _gdsl_list_node_get_pred (l->z));

  if (l->c != l->d)
    fprintf (file, " CURSOR=\"%p\">\n", (void *) l->c);
  else
    fprintf (file, " CURSOR=\"\">\n");

  while (tmp != l->z)
    {
      if (tmp == _gdsl_list_node_get_succ (l->d))
	{
	  fprintf (file, "<GDSL_LIST_NODE REF=\"%p\" CONTENT=\"%p\" SUCC=\"%p\" PRED=\"\">", 
		   (void *) tmp, (void *) _gdsl_list_node_get_content (tmp), 
		   (void *) _gdsl_list_node_get_succ (tmp));
	}
      else if (tmp == _gdsl_list_node_get_pred (l->z))
	{
	  fprintf (file, "<GDSL_LIST_NODE REF=\"%p\" CONTENT=\"%p\" SUCC=\"\" PRED=\"%p\">", 
		   (void *) tmp, (void *) _gdsl_list_node_get_content (tmp), 
		   (void *) _gdsl_list_node_get_pred (tmp));
	}
      else
	{
	  fprintf (file, "<GDSL_LIST_NODE REF=\"%p\" CONTENT=\"%p\" SUCC=\"%p\" PRED=\"%p\">", 
		   (void *) tmp, (void *) _gdsl_list_node_get_content (tmp), 
		   (void *) _gdsl_list_node_get_succ (tmp), (void *) _gdsl_list_node_get_pred (tmp));
	}

      if (f != NULL && _gdsl_list_node_get_content (tmp))
	_gdsl_list_node_write (tmp, f, file, user_data);

      fprintf (file, "</GDSL_LIST_NODE>\n");

      tmp = _gdsl_list_node_get_succ (tmp);
    }

  fprintf (file, "</GDSL_LIST>\n");
}

extern void
gdsl_list_dump (gdsl_list_t l, gdsl_write_func_t f, FILE *file, void *user_data)
{
  _gdsl_list_node_t tmp;

  tmp = _gdsl_list_node_get_succ (l->d);

  fprintf (file, "<GDSL_LIST REF=\"%p\" NAME=\"%s\" CARD=\"%ld\" HEAD=\"%p\" TAIL=\"%p\"", 
	   (void *) l, l->name ? l->name : "", l->card, (void *) l->d, (void *) l->z);

  if (l->c != l->d)
    fprintf (file, " CURSOR=\"%p\">\n", (void *) l->c);
  else
    fprintf (file, " CURSOR=\"\">\n");


  if (_gdsl_list_node_get_content (l->d))
    fprintf (file, "<GDSL_LIST_HEAD REF=\"%p\" CONTENT=\"%p\" SUCC=\"%p\" PRED=\"%p\">\n", 
	     (void *) l->d, (void *) _gdsl_list_node_get_content (l->d), 
	     (void *) _gdsl_list_node_get_succ (l->d), (void *) _gdsl_list_node_get_pred (l->d));
  else
    fprintf (file, "<GDSL_LIST_HEAD REF=\"%p\" CONTENT=\"\" SUCC=\"%p\" PRED=\"%p\">\n", 
	     (void *) l->d, (void *) _gdsl_list_node_get_succ (l->d), (void *) _gdsl_list_node_get_pred (l->d));

  while (tmp != l->z)
    {
      if (_gdsl_list_node_get_content (tmp))
	{
	  fprintf (file, "<GDSL_LIST_NODE REF=\"%p\" CONTENT=\"%p\" SUCC=\"%p\" PRED=\"%p\">", 
		   (void *) tmp, (void *) _gdsl_list_node_get_content (tmp),
		   (void *) _gdsl_list_node_get_succ (tmp), (void *) _gdsl_list_node_get_pred (tmp));
	}
      else
	{
	  fprintf (file, "<GDSL_LIST_NODE REF=\"%p\" CONTENT=\"\" SUCC=\"%p\" PRED=\"%p\">", 
		   (void *) tmp, (void *) _gdsl_list_node_get_succ (tmp), 
		   (void *) _gdsl_list_node_get_pred (tmp));
	}
      
      if (f != NULL && _gdsl_list_node_get_content (tmp))
	_gdsl_list_node_write (tmp, f, file, user_data);
      
      fprintf (file, "</GDSL_LIST_NODE>\n");

      tmp = _gdsl_list_node_get_succ (tmp);
    }

  if (_gdsl_list_node_get_content (l->z))
    fprintf (file, "<GDSL_LIST_TAIL REF=\"%p\" CONTENT=\"%p\" PRED=\"%p\" SUCC=\"%p\">\n</GDSL_LIST>\n", 
	     (void *) l->z, (void *) _gdsl_list_node_get_content (l->z), 
	     (void *) _gdsl_list_node_get_pred (l->z), (void *) _gdsl_list_node_get_succ (l->z));
  else
    fprintf (file, "<GDSL_LIST_TAIL REF=\"%p\" CONTENT=\"\" PRED=\"%p\" SUCC=\"%p\">\n</GDSL_LIST>\n", 
	     (void *) l->z, (void *) _gdsl_list_node_get_pred(l->z), 
	     (void *) _gdsl_list_node_get_succ (l->z));
}

/******************************************************************************/
/* Cursor specific functions                                                  */
/* WARNING: be carefull to not remove or destroy an element on wich the       */
/*          cursor is positioned!                                             */
/******************************************************************************/

extern void
gdsl_list_cursor_move_to_head (gdsl_list_t l)
{
  l->c = _gdsl_list_node_get_succ (l->d);
}

extern void
gdsl_list_cursor_move_to_tail (gdsl_list_t l)
{
  l->c = _gdsl_list_node_get_pred (l->z);
}

extern void
gdsl_list_cursor_move_forward (gdsl_list_t l)
{
  l->c = _gdsl_list_node_get_succ (l->c);
}

extern void
gdsl_list_cursor_move_backward (gdsl_list_t l)
{
  l->c = _gdsl_list_node_get_pred (l->c);
}

extern bool
gdsl_list_cursor_is_on_head (gdsl_list_t l)
{
  /* Returns FALSE if L's cursor is on l->s node */
  if (gdsl_list_is_empty (l))
    return FALSE;

  return (bool) (l->c == _gdsl_list_node_get_succ (l->d));
}

extern bool
gdsl_list_cursor_is_on_tail (gdsl_list_t l)
{
  /* Returns FALSE if L's cursor is on l->s node */
  if (gdsl_list_is_empty (l))
    return FALSE;

  return (bool) (l->c == _gdsl_list_node_get_pred (l->z));
}

extern bool 
gdsl_list_cursor_has_succ (gdsl_list_t l)
{
  return (bool) (_gdsl_list_node_get_succ (l->c) != l->z);
}

extern bool 
gdsl_list_cursor_has_pred (gdsl_list_t l)
{
  return (bool) (_gdsl_list_node_get_pred (l->c) != l->d);
}

extern void
gdsl_list_cursor_set (gdsl_list_t l, gdsl_element_t e)
{
  if (e != NULL)
    _gdsl_list_node_set_content (l->c, e);
}

extern gdsl_element_t
gdsl_list_cursor_get (gdsl_list_t l)
{
  return _gdsl_list_node_get_content (l->c);
}

extern gdsl_element_t
gdsl_list_cursor_search_by_value (gdsl_list_t l, gdsl_compare_func_t f, void *v)
{
  return update_cursor (l, search_by_function (l, f, v));
}

extern gdsl_element_t
gdsl_list_cursor_search_by_position (gdsl_list_t l, ulong i)
{
  return update_cursor (l, search_by_position (l, i));
}

extern gdsl_element_t
gdsl_list_cursor_insert_after (gdsl_list_t l, void *v)
{
  gdsl_element_t    e;
  _gdsl_list_node_t n;

  if (l->c == l->d)
    return NULL;

  if (l->c == l->z)
    return NULL;

  n = _gdsl_list_node_alloc ();
  if (n == NULL)
    return NULL;

  e = (l->alloc_f) (v);
  if (e == NULL)
    {
      _gdsl_list_node_free (n);
      return NULL;
    }

  _gdsl_list_node_set_content (n, e);
  _gdsl_list_insert_after (n, l->c);

  l->card++;

  return e;
}

extern gdsl_element_t
gdsl_list_cursor_insert_before (gdsl_list_t l, void *v)
{
  gdsl_element_t    e;
  _gdsl_list_node_t n;

  if (l->c == l->d)
    return NULL;

  if (l->c == l->z)
    return NULL;

  n = _gdsl_list_node_alloc ();
  if (n == NULL)
    return NULL;

  e = (l->alloc_f) (v);
  if (e == NULL)
    {
      _gdsl_list_node_free (n);
      return NULL;
    }

  _gdsl_list_node_set_content (n, e);
  _gdsl_list_insert_before (n, l->c);

  l->card++;

  return e;
}

extern gdsl_element_t
gdsl_list_cursor_remove_after (gdsl_list_t l)
{
  gdsl_element_t    e;
  _gdsl_list_node_t tmp;

  if (l->c == l->d)
    return NULL;

  tmp = _gdsl_list_node_get_succ (l->c);
  if (tmp == l->z)
    return NULL;

  _gdsl_list_remove (tmp);
  e = _gdsl_list_node_get_content (tmp);
  _gdsl_list_node_free (tmp);

  l->card--;

  return e;
}

extern gdsl_element_t
gdsl_list_cursor_remove_before (gdsl_list_t l)
{
  gdsl_element_t    e;
  _gdsl_list_node_t tmp;

  if (l->c == l->z)
    return NULL;

  tmp = _gdsl_list_node_get_pred (l->c);
  if (tmp == l->d)
    return NULL;

  _gdsl_list_remove (tmp);
  e = _gdsl_list_node_get_content (tmp);
  _gdsl_list_node_free (tmp);

  l->card--;

  return e;
}

extern gdsl_list_t
gdsl_list_cursor_destroy_after (gdsl_list_t l)
{
  gdsl_element_t e;

  e = gdsl_list_cursor_remove_after (l);
  if (e != NULL)
    {
      l->free_f (e);
      return l;
    }

  return NULL;
}

extern gdsl_list_t
gdsl_list_cursor_destroy_before (gdsl_list_t l)
{
  gdsl_element_t e;

  e = gdsl_list_cursor_remove_before (l);
  if (e != NULL)
    {
      l->free_f (e);
      return l;
    }

  return NULL;
}

/******************************************************************************/
/* Private functions                                                          */
/******************************************************************************/

static gdsl_element_t
default_alloc (void *e)
{
  return e;
}

static void 
default_free (gdsl_element_t e)
{
  ;
}

static _gdsl_list_node_t
search_by_function (gdsl_list_t l, gdsl_compare_func_t f, void *v)
{
  _gdsl_list_node_t left;
  _gdsl_list_node_t right;

  for (left = _gdsl_list_node_get_succ (l->d), right = _gdsl_list_node_get_pred (l->z);
       left != _gdsl_list_node_get_succ (right);
       left = _gdsl_list_node_get_succ (left), right = _gdsl_list_node_get_pred (right))
    {
      if (f (_gdsl_list_node_get_content (left), v) == 0)
	return left;
      
      if (f (_gdsl_list_node_get_content (right), v) == 0)
	return right;
      
      if (left == right)
	return NULL;
    }

  return NULL;
}

static _gdsl_list_node_t
search_by_position (gdsl_list_t l, ulong i)
{
  ulong            m;
  _gdsl_list_node_t tmp;

  if (i <= 0 || i > l->card)
    return NULL;

  m = (l->card / 2) + 1;
  if (i < m)
    {
      tmp = _gdsl_list_node_get_succ (l->d);

      while (i > 1)
	{
	  tmp = _gdsl_list_node_get_succ (tmp);
	  i--;
	}
    }
  else
    {
      i = l->card - i;
      tmp = _gdsl_list_node_get_pred (l->z);

      while (i > 0)
	{
	  tmp = _gdsl_list_node_get_pred (tmp);
	  i--;
	}
    }

  return tmp;
}

static gdsl_element_t
update_cursor (gdsl_list_t l, _gdsl_list_node_t n)
{
  if (n == NULL) 
    return NULL;

  l->c = n;

  return _gdsl_list_node_get_content (n);
}

static _gdsl_list_node_t
sort (_gdsl_list_node_t u, gdsl_compare_func_t f, _gdsl_list_node_t z)
{
  _gdsl_list_node_t s;
  _gdsl_list_node_t t;

  if (_gdsl_list_node_get_succ (u) == z) 
    return u;

  s = u;
  t = _gdsl_list_node_get_succ (_gdsl_list_node_get_succ (_gdsl_list_node_get_succ (u)));
  while (t != z)
    {
      u = _gdsl_list_node_get_succ (u);
      t = _gdsl_list_node_get_succ (_gdsl_list_node_get_succ (t));
    }

  t = _gdsl_list_node_get_succ (u);
  _gdsl_list_node_set_succ (u, z);

  return merge (sort (s, f, z), sort (t, f, z), f, z);
}

static _gdsl_list_node_t
merge (_gdsl_list_node_t s, _gdsl_list_node_t t, gdsl_compare_func_t f, _gdsl_list_node_t z)
{
  _gdsl_list_node_t u = z;

  do 
    {
      if (f (_gdsl_list_node_get_content (s), _gdsl_list_node_get_content (t)) <= 0)
	{
	  _gdsl_list_node_set_succ (u, s);
	  u = s;
	  s = _gdsl_list_node_get_succ (s);
	}
      else
	{
	  _gdsl_list_node_set_succ (u, t);
	  u = t;
	  t = _gdsl_list_node_get_succ (t);
	}
    } while (u != z);
  
  u = _gdsl_list_node_get_succ (z);
  _gdsl_list_node_set_succ (z, z);
  
  return u;
}
