/*
 * This file is part of Generic Data Structures Library (GDSL).
 * Copyright (C) 1998-2001 Nicolas Darnis <ndarnis@free.fr>
 *
 * The GDSL library is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * The GDSL library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the GDSL library; see the file COPYING.
 * If not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 * File: gdsl_perm.c
 * $Revision: 1.7 $
 * $Date: 2002/02/21 09:55:51 $
 * $Author: ndarnis $
 * $Id: gdsl_perm.c,v 1.7 2002/02/21 09:55:51 ndarnis Exp $
 */

#include <config.h>

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <assert.h>

#include "gdsl_list.h"
#include "gdsl_perm.h"
#include "_gdsl_mem.h"

struct gdsl_perm
{
  char *            name;
  ulong             n;
  ulong *           e;
  gdsl_alloc_func_t alloc_f;
  gdsl_free_func_t  free_f;
  gdsl_element_t *  elements;
};

struct cycle
{
  ulong  lg;
  ulong *cycle;
};

static gdsl_element_t 
default_alloc (void *e);

static void 
default_free (gdsl_element_t e);

static gdsl_element_t
cycle_alloc (void *e);

static void
cycle_free (gdsl_element_t e);

static int
cycle_comp (gdsl_element_t e, void *v);

/******************************************************************************/
/* Management functions                                                       */
/******************************************************************************/

extern gdsl_perm_t
gdsl_perm_create (const char *name, const ulong n, const ulong *values, 
		  void **elements, const gdsl_alloc_func_t alloc_f, 
		  const gdsl_free_func_t free_f)
{
  ulong       i;
  gdsl_perm_t p;

  assert (n > 0);
  
  p = (gdsl_perm_t) _gdsl_mem_alloc (sizeof (struct gdsl_perm));
  if (p == NULL)
    {
      return NULL;
    }

  p->n = n;

  p->e = (ulong *) _gdsl_mem_alloc (n * sizeof (ulong));
  if (p->e == NULL)
    {
      _gdsl_mem_free (p);
      return NULL;
    }

  p->elements = (gdsl_element_t *) _gdsl_mem_alloc (n * sizeof (gdsl_element_t));
  if (p->elements == NULL)
    {
      _gdsl_mem_free (p->e);
      _gdsl_mem_free (p);
      return NULL;
    }

  p->free_f  = free_f  ? free_f  : default_free;
  p->alloc_f = alloc_f ? alloc_f : default_alloc;

  p->name = NULL;

  if (name != NULL)
    {
      p->name = (char *) _gdsl_mem_alloc ((1 + strlen (name)) * sizeof (char));
      if (p->name == NULL)
	{
	  _gdsl_mem_free (p->elements);
	  _gdsl_mem_free (p->e);
	  _gdsl_mem_free (p);
	  return NULL;
	}
      strcpy (p->name, name);
    }

  if (values == NULL)
    {
      for (i = 0; i < n; i++)
	{
	  p->e[i] = i;
	}
    }
  else
    {
      for (i = 0; i < n; i++)
	{
	  p->e[i] = values[i];
	}
    }

  if (elements == NULL)
    {
      for (i = 0; i < n; i++)
	{
	  p->elements[i] == NULL;
	}
    }
  else
    {
      for (i = 0; i < n; i++)
	{
	  p->elements[i] = p->alloc_f (elements[i]);
	  if (p->elements[i] == NULL)
	    {
	      ulong j;
	      for (j = 0; j < i; j++)
		{
		  p->free_f (p->elements[j]);
		}
	      _gdsl_mem_free (p->elements);
	      _gdsl_mem_free (p->e);
	      _gdsl_mem_free (p);
	      return NULL;
	    }
	}
    }

  return p;
}

extern void
gdsl_perm_destroy (gdsl_perm_t p)
{
  ulong i;

  assert (p != NULL);

  _gdsl_mem_free (p->e);

  for (i = 0; i < p->n; i++)
    {
      p->free_f (p->elements[i]);
    }

  if (p->name != NULL)
    {
      _gdsl_mem_free (p->name);
    }

  _gdsl_mem_free (p->elements);
  _gdsl_mem_free (p);
}

/******************************************************************************/
/* Information functions                                                      */
/******************************************************************************/

extern const char *
gdsl_perm_get_name (const gdsl_perm_t p)
{
  assert (p != NULL);

  return p->name;
}

extern ulong
gdsl_perm_get (const gdsl_perm_t p, const ulong n)
{
  assert (p != NULL);
  assert (n <= p->n);

  return p->e[n];
}

extern ulong
gdsl_perm_get_card (const gdsl_perm_t p)
{
  assert (p != NULL);

  return p->n;
}

extern gdsl_element_t
gdsl_perm_get_content (const gdsl_perm_t p, const ulong n)
{
  assert (p != NULL);
  assert (n <= p->n);

  return p->elements[n];
}

extern ulong
gdsl_perm_inversions_count (const gdsl_perm_t p)
{
  ulong i;
  ulong j;
  ulong count = 0;

  assert (p != NULL);

  for (i = 0; i < p->n - 1; i++)
    {
      for (j = i + 1; j < p->n; j++)
	{
	  if (p->e[i] > p->e[j])
	    {
	      count++;
	    }
	}
    }

  return count;
}

extern ulong
gdsl_perm_cycles_count (const gdsl_perm_t p)
{
  ulong i;
  ulong j;
  ulong count = 0;
  char *atteint;

  assert (p != NULL);

  atteint = (char *) _gdsl_mem_alloc (p->n * sizeof (char));
  if (atteint == NULL)
    {
      return 0;
    }

  for (i = 0; i < p->n; i++)
    {
      atteint[i] = 0;
    }

  for (i = 0; i < p->n; i++)
    {
      if (atteint[i] == 0)
	{
	  atteint[i] = 1;

	  for (j = p->e[i]; atteint[j] == 0; j = p->e[j])
	    {
	      atteint[j] = 1;
	    }

	  count++;
	}
    }

  _gdsl_mem_free (atteint);

  return count;
}

extern ulong
gdsl_perm_canonical_cycles_count (const gdsl_perm_t p)
{
  ulong i;
  ulong count = 0;
  ulong max = 0;

  assert (p != NULL);

  for (i = 0; i < p->n; i++)
    {
      if (p->e[i] >= max)
	{
	  max = p->e[i];
	  count++;
	}
    }

  return count;
}

extern gdsl_perm_t
gdsl_perm_get_next (gdsl_perm_t p)
{
  /* Replaces p with the next permutation (in the standard lexicographical
   * ordering).  Returns NULL if there is no next permutation.
   */
  long i, j, k;

  assert (p != NULL);

  if (p->n < 2)
    {
      return NULL;
    }

  i = p->n - 2;

  while ((p->e[i] > p->e[i+1]) && (i != 0))
    {
      i--;
    }

  if ((i == 0) && (p->e[0] > p->e[1]))
    {
     return NULL;
    }

  k = i + 1;

  for (j = i + 2; j < p->n; j++ )
    {
      if ((p->e[j] > p->e[i]) && (p->e[j] < p->e[k]))
        {
          k = j;
        }
    }

  /* swap i and k */

  {
    ulong tmp = p->e[i];
    gdsl_element_t tmp_e = p->elements[i];
    p->e[i] = p->e[k];
    p->e[k] = tmp;
    p->elements[i] = p->elements[k];
    p->elements[k] = tmp_e;
  }

  for (j = i + 1; j <= ((p->n + i) / 2); j++)
    {
      ulong tmp = p->e[j];
      gdsl_element_t tmp_e = p->elements[j];
      p->e[j] = p->e[p->n + i - j];
      p->e[p->n + i - j] = tmp;
      p->elements[j] = p->elements[p->n + i -j];
      p->elements[p->n + i - j] = tmp_e;
    }

  return p;
}

extern gdsl_perm_t
gdsl_perm_get_prev (gdsl_perm_t p)
{
  long i, j, k;

  assert (p != NULL);

  if (p->n < 2)
    {
      return NULL;
    }

  i = p->n - 2;

  while ((p->e[i] < p->e[i+1]) && (i != 0))
    {
      i--;
    }

  if ((i == 0) && (p->e[0] < p->e[1]))
    {
      return NULL;
    }

  k = i + 1;

  for (j = i + 2; j < p->n; j++ )
    {
      if ((p->e[j] < p->e[i]) && (p->e[j] > p->e[k]))
        {
          k = j;
        }
    }

  /* swap i and k */

  {
    ulong tmp = p->e[i];
    gdsl_element_t tmp_e = p->elements[i];
    p->e[i] = p->e[k];
    p->e[k] = tmp;
    p->elements[i] = p->elements[k];
    p->elements[k] = tmp_e;
  }

  for (j = i + 1; j <= ((p->n + i) / 2); j++)
    {
      ulong tmp = p->e[j];
      gdsl_element_t tmp_e = p->elements[j];
      p->e[j] = p->e[p->n + i - j];
      p->e[p->n + i - j] = tmp;
      p->elements[j] = p->elements[p->n + i - j];
      p->elements[p->n + i - j] = tmp_e;
    }

  return p;
}

/******************************************************************************/
/* Modification functions                                                     */
/******************************************************************************/

extern gdsl_perm_t
gdsl_perm_set_name (gdsl_perm_t p, const char *name)
{
  assert (p != NULL);

  if (p->name != NULL)
    {
      _gdsl_mem_free (p->name);
      p->name = NULL;
    }

  if (name != NULL)
    {
      p->name = (char *) _gdsl_mem_alloc ((1 + strlen (name)) * sizeof (char));

      if (p->name == NULL)
	{
	  return NULL;
	}

      strcpy (p->name, name);
    }
  
  return p;
}

extern void
gdsl_perm_set (gdsl_perm_t p, const ulong n, const ulong e)
{
  assert (p != NULL);
  assert (n <= p->n);

  p->e[n] = e;
}


extern gdsl_element_t
gdsl_perm_set_content (gdsl_perm_t p, const ulong n, void *e)
{
  assert (p != NULL);
  assert (n <= p->n);

  p->elements[n] = p->alloc_f (e);

  return p->elements[n];
}

/******************************************************************************/
/* Operations functions                                                       */
/******************************************************************************/

extern gdsl_perm_t
gdsl_perm_mul (gdsl_perm_t result, const gdsl_perm_t a, const gdsl_perm_t b)
{
  ulong       i;
  gdsl_perm_t tmp;

  assert (result != NULL);
  assert (a != NULL);
  assert (b != NULL);
  assert (result->n == a->n);
  assert (a->n == b->n);

  tmp = gdsl_perm_create (NULL, a->n, NULL, NULL, NULL, NULL);
  if (tmp == NULL)
    {
      return NULL;
    }

  /* Sauvegarde des elements de a dans tmp */
  for (i = 0; i < a->n; i++)
    {
      tmp->elements[a->e[i]] = a->elements[i];
    }

  for (i = 0; i < result->n; i++)
    {
      result->e[i] = b->e[a->e[i]];
    }

  /* Restauration des elements de tmp dans e */
  for (i = 0; i < result->n; i++)
    {
      result->elements[i] = tmp->elements[result->e[i]];
    }

  return result;
}

extern gdsl_perm_t
gdsl_perm_to_canonical (gdsl_perm_t p)
{
  ulong       i;
  ulong       k;
  char *      atteint;
  gdsl_list_t l;
  gdsl_perm_t tmp;

  assert (p != NULL);

  l = gdsl_list_create (NULL, cycle_alloc, cycle_free);
  if (l == NULL)
    {
      return NULL;
    }

  tmp = gdsl_perm_create (NULL, p->n, NULL, NULL, NULL, NULL);
  if (tmp == NULL)
    {
      gdsl_list_destroy (l);
      return NULL;
    }

  atteint = (char *) _gdsl_mem_alloc (p->n * sizeof (char));
  if (atteint == NULL)
    {
      gdsl_perm_destroy (tmp);
      gdsl_list_destroy (l);
      return NULL;
    }

  for (i = 0; i < p->n; i++)
    {
      atteint[i] = 0;
      tmp->e[i] = i;
      tmp->elements[p->e[i]] = p->elements[i];
    }

  for (k = 0, i = 0; i < p->n; i++)
    {
      if (atteint[i] == 0)
	{
	  ulong j;
	  ulong begin = k;
	  ulong end = k;
	  ulong maxj = 0;
	  struct cycle cycle;

	  atteint[i] = 1;

	  tmp->e[k++] = i;
	  for (j = p->e[i]; atteint[j] == 0; j = p->e[j], end++)
	    {
	      atteint[j] = 1;
	      tmp->e[k++] = j;
	      maxj = j > maxj ? j : maxj;
	    }

	  if (begin == end)
	    {
	      cycle.lg = 1;
	      cycle.cycle = &tmp->e[begin];

	      if (gdsl_list_insert_tail (l, &cycle) == NULL)
		{
		  gdsl_list_destroy (l);
		  _gdsl_mem_free (atteint);
		  gdsl_perm_destroy (tmp);
		  return NULL;
		}
	      
	      continue;
	    }

	  cycle.lg = end - begin + 1;
	  cycle.cycle = &tmp->e[begin];

	  if (gdsl_list_insert_tail (l, &cycle) == NULL)
	    {
	      gdsl_list_destroy (l);
	      _gdsl_mem_free (atteint);
	      gdsl_perm_destroy (tmp);
	      return NULL;
	    }

	  /* On reorganise le cycle */
	  while (tmp->e[begin] != maxj)
	    {
	      ulong t = tmp->e[begin];
	      
	      for (j = begin; j < end; j++)
		{
		  tmp->e[j] = tmp->e[j+1];
		}

	      tmp->e[j] = t;
	    }
	}
    }

  {
    struct cycle *cycle;
    ulong maxv[1];
    struct cycle max;

    maxv[0] = p->n + 1;

    max.lg = 1;
    max.cycle = maxv;

    gdsl_list_merge_sort (l, cycle_comp, (gdsl_element_t) &max);

    for (k = 0, gdsl_list_cursor_move_to_head (l); 
	 cycle = (struct cycle *) gdsl_list_cursor_get (l); 
	 gdsl_list_cursor_move_forward (l))
      {
	for (i = 0; i < cycle->lg; i++, k++)
	  {
	    p->e[k] = cycle->cycle[i];
	  }
      }
  }

  /* Restauration des elements de tmp dans e */
  for (i = 0; i < p->n; i++)
    {
      p->elements[i] = tmp->elements[p->e[i]];
    }

  gdsl_list_destroy (l);
  _gdsl_mem_free (atteint);
  gdsl_perm_destroy (tmp);

  return p;
}

extern gdsl_perm_t
gdsl_perm_canonical_to_perm (gdsl_perm_t p)
{
  ulong  i;
  ulong  first;
  gdsl_perm_t tmp;

  assert (p != NULL);

  tmp = gdsl_perm_create (NULL, p->n, NULL, NULL, NULL, NULL);
  if (tmp == NULL)
    {
      return NULL;
    }
  
  /* Sauvegarde des elements de p dans tmp */
  for (i = 0; i < p->n; i++)
    {
      tmp->e[i] = p->e[i];
      tmp->elements[p->e[i]] = p->elements[i];
    }

  for (first = tmp->e[0], i = 1; i < p->n; i++)
    {
      if (tmp->e[i] > first)
	{
	  p->e[tmp->e[i-1]] = first;
	  first = tmp->e[i];
	}
      else
	{
	  p->e[tmp->e[i-1]] = tmp->e[i];
	}
    }

  p->e[tmp->e[i-1]] = first;

  /* Restauration des elements de tmp dans e */
  for (i = 0; i < p->n; i++)
    {
      p->elements[i] = tmp->elements[p->e[i]];
    }

  gdsl_perm_destroy (tmp);

  return p;
}

extern gdsl_perm_t
gdsl_perm_inverse (gdsl_perm_t p)
{
  ulong       i;
  gdsl_perm_t tmp;

  assert (p != NULL);

  tmp = gdsl_perm_create (NULL, p->n, NULL, NULL, NULL, NULL);
  if (tmp == NULL)
    {
      return NULL;
    }

  /* Sauvegarde des elements de p dans tmp */
  for (i = 0; i < p->n; i++)
    {
      tmp->e[i] = p->e[i];
      tmp->elements[p->e[i]] = p->elements[i];
    }

  for (i = 0; i < p->n; i++)
    {
      p->e[tmp->e[i]] = i;
    }

  /* Restauration des elements de tmp dans e */
  for (i = 0; i < p->n; i++)
    {
      p->elements[i] = tmp->elements[p->e[i]];
    }

  gdsl_perm_destroy (tmp);

  return p;
}

extern void
gdsl_perm_reverse (gdsl_perm_t p)
{
  ulong i;
  
  assert (p != NULL);

  for (i = 0; i < (p->n / 2); i++) 
    {
      ulong j = p->n - i - 1;
      ulong tmp = p->e[i];
      gdsl_element_t tmp_e = p->elements[i];
      p->e[i] = p->e[j];
      p->elements[i] = p->elements[j];
      p->e[j] = tmp;
      p->elements[j] = tmp_e;
    }
}

extern gdsl_perm_t
gdsl_perm_randomize (gdsl_perm_t p)
{
  ulong          i;
  long           j;
  long           k;
  gdsl_perm_t    tmp;
  struct timeval tv;

  assert (p != NULL);

  tmp = gdsl_perm_create (NULL, p->n, NULL, NULL, NULL, NULL);
  if (tmp == NULL)
    {
      return NULL;
    }

  /* Sauvegarde des elements de p dans tmp */
  for (i = 0; i < p->n; i++)
    {
      tmp->elements[p->e[i]] = p->elements[i];
    }

  /* Construction alatoire d'un tableau d'inversions t[] */
  gettimeofday (&tv, NULL);
  srand (tv.tv_usec);

  for (i = 0; i < p->n - 1; i++)
    {
      tmp->e[i] = rand () % (p->n - i); /* qq soit i E [0,N-1], 0 <= t[i] <= N-i-1 */
    }
  tmp->e[i] = 0;

  /* Conversion de tmp->e[] en permutation */
  for (k = p->n - 1; k >= 0; k--)
    {
      for (j = p->n - k - 1; j > tmp->e[k]; j--)
	{
	  p->e[j] = p->e[j-1];
	}

      p->e[j] = k;
    }

  /* Restauration des elements de tmp dans e */
  for (i = 0; i < p->n; i++)
    {
      p->elements[i] = tmp->elements[p->e[i]];
    }

  gdsl_perm_destroy (tmp);
  return p;
}

/******************************************************************************/
/* Parse functions                                                            */
/******************************************************************************/

extern gdsl_element_t
gdsl_perm_map_forward (const gdsl_perm_t p, const gdsl_map_func_t map_f, 
		       void *user_data)
{
  ulong          i;
  gdsl_element_t e;

  assert (p != NULL);
  assert (map_f != NULL);

  for (i = 0; i < p->n; i++)
    {
      e = p->elements[i];

      if (map_f (e, user_data) == 0)
	{
	  return e;
	}
    }

  return NULL;
}

extern gdsl_element_t
gdsl_perm_map_backward (const gdsl_perm_t p, const gdsl_map_func_t map_f, 
			void *user_data)
{
  long           i;
  gdsl_element_t e;

  assert (p != NULL);
  assert (map_f != NULL);
  
  for (i = p->n; i >= 0; i--)
    {
      e = p->elements[i];

      if (map_f (e, user_data) == 0)
	{
	  return e;
	}
    }

  return NULL;
}

/******************************************************************************/
/* Input/output functions                                                     */
/******************************************************************************/

extern void
gdsl_perm_write (const gdsl_perm_t p, const gdsl_write_func_t write_f, 
		 FILE *file, void *user_data)
{
  ulong i;
  
  assert (p != NULL);

  fprintf (file, "( ");
  for (i = 0; i < p->n - 1; i++)
    {
      fprintf (file, "%ld ", p->e[i]);
      if (write_f != NULL)
	{
	  write_f (p->elements[i], file, user_data);
	}
    }
  fprintf (file, "%ld ", p->e[i]);
  if (write_f != NULL)
    {
      write_f (p->elements[i], file, user_data);
    }
  fprintf (file, ")");
}

extern void
gdsl_perm_write_xml (const gdsl_perm_t p, const gdsl_write_func_t write_f, 
		     FILE *file, void *user_data)
{
  ulong i;

  assert (p != NULL);

  fprintf (file, "<GDSL_PERM REF=\"%p\" NAME=\"%s\" CARD=\"%ld\">\n", 
	   (void *) p, p->name == NULL ? "" : p->name, p->n); 

  for (i = 0; i < p->n; i++)
    {
      gdsl_element_t e = p->elements[i];

      if (e == NULL)
	{
	  fprintf (file, "<GDSL_PERM_ELEMENT INDIX=\"%ld\" VALUE=\"%ld\" CONTENT=\"\">", 
		   i, p->e[i]);
	}
      else
	{
	  fprintf (file, "<GDSL_PERM_ELEMENT INDIX=\"%ld\" VALUE=\"%ld\" CONTENT=\"%p\">", 
		   i, p->e[i], e);
	}
 
      if (write_f != NULL)
      {
	fprintf (file, "\n");
	write_f (e, file, user_data);
	fprintf (file, "\n");
      }

      fprintf (file, "</GDSL_PERM_ELEMENT>\n");
    }

  fprintf (file, "</GDSL_PERM>\n");
}

extern void
gdsl_perm_dump (const gdsl_perm_t p, const gdsl_write_func_t write_f, 
		FILE *file, void *user_data)
{
  assert (p != NULL);

  gdsl_perm_write_xml (p, write_f, file, user_data);
}

/******************************************************************************/
/* Private functions                                                          */
/******************************************************************************/

static gdsl_element_t 
default_alloc (void *e)
{
  return e;
}

static void 
default_free (gdsl_element_t e)
{
  ;
}

static gdsl_element_t
cycle_alloc (void *e)
{
  struct cycle *tmp = (struct cycle *) e;
  struct cycle *cycle;

  cycle = (struct cycle *) _gdsl_mem_alloc (sizeof (struct cycle));
  if (cycle == NULL)
    {
      return NULL;
    }
  
  cycle->lg = tmp->lg;
  cycle->cycle = tmp->cycle;

  return cycle;
}

static void
cycle_free (gdsl_element_t e)
{
  _gdsl_mem_free (e);
}

static int
cycle_comp (gdsl_element_t e, void *v)
{
  return ((struct cycle *) e)->cycle[0] - ((struct cycle *) v)->cycle[0];
}
