/*
 * This file is part of Generic Data Structures Library (GDSL).
 * Copyright (C) 1998-2001 Nicolas Darnis <ndarnis@free.fr>
 *
 * The GDSL library is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * The GDSL library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the GDSL library; see the file COPYING.
 * If not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 * File: gdsl_queue.c
 * $Revision: 1.5 $
 * $Date: 2002/02/21 09:55:51 $
 * $Author: ndarnis $
 * $Id: gdsl_queue.c,v 1.5 2002/02/21 09:55:51 ndarnis Exp $
 */

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "_gdsl_list.h"
#include "_gdsl_list_node.h"
#include "gdsl_types.h"
#include "gdsl_queue.h"
#include "_gdsl_mem.h"

struct _gdsl_queue
{
  _gdsl_list_node_t d;    /* Start of the queue (sentinel) */
  _gdsl_list_node_t z;    /* End of the queue (sentinel) */
  char *            name; /* Name of the queue */
  ulong             card; /* Cardinality of the queue */

  /* Functions pointers */
  gdsl_alloc_func_t alloc_f;
  gdsl_free_func_t  free_f;
};

static gdsl_element_t 
default_alloc (void *e);

static void 
default_free (gdsl_element_t e);

static _gdsl_list_node_t
search_by_key (gdsl_queue_t q, gdsl_compare_func_t f, void *key);

/******************************************************************************/
/* Management functions                                                       */
/******************************************************************************/

extern gdsl_queue_t
gdsl_queue_create (const char *name, gdsl_alloc_func_t alloc_f, 
		   gdsl_free_func_t free_f)
{
  gdsl_queue_t q = NULL;

  q = (gdsl_queue_t) _gdsl_mem_alloc (sizeof (struct _gdsl_queue));
  if (q == NULL)
    return NULL;

  q->d = _gdsl_list_node_alloc ();
  if (q->d == NULL)
    {
      _gdsl_mem_free (q);
      return NULL;
    }

  q->z = _gdsl_list_node_alloc ();
  if (q->z == NULL)
    {
      _gdsl_list_node_free (q->d);
      _gdsl_mem_free (q);
      return NULL;
    }

  q->name = NULL;

  if (name != NULL)
    {
      q->name = (char *) _gdsl_mem_alloc ((1 + strlen (name)) * sizeof (char));
      if (q->name == NULL)
	{
	  _gdsl_list_node_free (q->z);
	  _gdsl_list_node_free (q->d);
	  _gdsl_mem_free (q);
	  return NULL;
	}
      strcpy (q->name, name);
    }

  q->card = 0UL;

  _gdsl_list_link (q->d, q->z);
  _gdsl_list_node_set_succ (q->z, q->z);
  _gdsl_list_node_set_pred (q->d, q->d);

  q->alloc_f = alloc_f ? alloc_f : default_alloc;
  q->free_f  = free_f  ? free_f  : default_free;

  return q;
}

extern void 
gdsl_queue_destroy (gdsl_queue_t q)
{
  if (!gdsl_queue_is_empty (q))
    gdsl_queue_flush (q);

  _gdsl_list_node_free (q->d);
  _gdsl_list_node_free (q->z);

  if (q->name != NULL)
    _gdsl_mem_free (q->name);

  _gdsl_mem_free (q);
}

extern void
gdsl_queue_flush (gdsl_queue_t q)
{
  _gdsl_list_node_t save;
  _gdsl_list_node_t tmp = _gdsl_list_node_get_succ (q->d);

  while (tmp != q->z)
    {
      save = _gdsl_list_node_get_succ (tmp);
      q->free_f (_gdsl_list_node_get_content (tmp));
      _gdsl_list_node_free (tmp);
      tmp = save;
    }

  q->card = 0UL;

  _gdsl_list_link (q->d, q->z);
  _gdsl_list_node_set_succ (q->z, q->z);
  _gdsl_list_node_set_pred (q->d, q->d);
}

extern gdsl_queue_t
gdsl_queue_set_name (gdsl_queue_t q, const char *name)
{
  if (q->name != NULL)
    {
      _gdsl_mem_free (q->name);
      q->name = NULL;
    }
  
  if (name != NULL)
    {
      q->name = (char *) _gdsl_mem_alloc ((1 + strlen (name)) * sizeof (char));

      if (q->name == NULL)
	return NULL;
    
      strcpy (q->name, name);
    }

  return q;
}

/******************************************************************************/
/* Information functions                                                      */
/******************************************************************************/

extern const char *
gdsl_queue_get_name (gdsl_queue_t q)
{
  return q->name;
}

extern ulong
gdsl_queue_get_card (gdsl_queue_t q)
{
  return q->card;
}

extern bool 
gdsl_queue_is_empty (gdsl_queue_t q)
{
  return (bool) (_gdsl_list_node_get_succ (q->d) == q->z);
}

extern gdsl_element_t
gdsl_queue_get_head (gdsl_queue_t q)
{
  return _gdsl_list_node_get_content (_gdsl_list_node_get_pred (q->z));
}

extern gdsl_element_t
gdsl_queue_get_tail (gdsl_queue_t q)
{
  return _gdsl_list_node_get_content (_gdsl_list_node_get_succ (q->d));
}

/******************************************************************************/
/* Insert functions                                                           */
/******************************************************************************/

extern gdsl_element_t
gdsl_queue_put (gdsl_queue_t q, void *key)
{
  gdsl_element_t    e;
  _gdsl_list_node_t head = NULL;

  head = _gdsl_list_node_alloc ();
  if (head == NULL)
    return NULL;

  e = (q->alloc_f) (key);
  if (e == NULL)
    {
      _gdsl_list_node_free (head);
      return NULL;
    }

  q->card++;
  _gdsl_list_node_set_content (head, e);
  
  {
    _gdsl_list_node_t tmp = _gdsl_list_node_get_succ (q->d);
    _gdsl_list_link (q->d, head);
    _gdsl_list_link (head, tmp);
  }
  
  return e;
}

/******************************************************************************/
/* Remove functions                                                           */
/******************************************************************************/

extern gdsl_element_t
gdsl_queue_get (gdsl_queue_t q)
{
  if (!gdsl_queue_is_empty (q))
    {
      _gdsl_list_node_t tail = _gdsl_list_node_get_pred (q->z);
      gdsl_element_t e = _gdsl_list_node_get_content (tail);

      _gdsl_list_remove (tail);
      _gdsl_list_node_free (tail);

      q->card--;
      return e;
    }

  return NULL;
}

/******************************************************************************/
/* Search functions                                                           */
/******************************************************************************/

extern gdsl_element_t
gdsl_queue_search (gdsl_queue_t q, gdsl_compare_func_t comp_f, void *key)
{
  _gdsl_list_node_t n = search_by_key (q, comp_f, key);
  return n == NULL ? NULL : _gdsl_list_node_get_content (n);
}

/******************************************************************************/
/* Parse functions                                                            */
/******************************************************************************/

extern gdsl_element_t 
gdsl_queue_map (gdsl_queue_t q, gdsl_map_func_t f, void *user_data)
{
  gdsl_element_t    e;
  _gdsl_list_node_t tmp = _gdsl_list_node_get_succ (q->d);

  while (tmp != q->z)
    {
      e = _gdsl_list_node_get_content (tmp);

      if (f (e, user_data) == 0)
	return e;

      tmp = _gdsl_list_node_get_succ (tmp);
    }

  return NULL;
}

/******************************************************************************/
/* Output functions                                                           */
/******************************************************************************/

extern void
gdsl_queue_write (gdsl_queue_t q, gdsl_write_func_t write_f, FILE *file, 
		  void *user_data)
{
  _gdsl_list_node_t tmp = _gdsl_list_node_get_succ (q->d);

  while (tmp != q->z)
    {
      _gdsl_list_node_write (tmp, write_f, file, user_data);
      tmp = _gdsl_list_node_get_succ (tmp);
    }
}

extern void
gdsl_queue_write_xml (gdsl_queue_t q, gdsl_write_func_t write_f, FILE *file, 
		      void *user_data)
{
  _gdsl_list_node_t tmp = _gdsl_list_node_get_succ (q->d);

  fprintf (file, "<GDSL_QUEUE REF=\"%p\" NAME=\"%s\" CARD=\"%ld\" HEAD=\"%p\" TAIL=\"%p\">\n", 
	   (void *) q, q->name, q->card, (void *) tmp, 
	   (void *) _gdsl_list_node_get_pred (q->z));

  while (tmp != q->z)
    {
      if (tmp == _gdsl_list_node_get_succ (q->d))
	{
	  fprintf (file, "<GDSL_QUEUE_NODE REF=\"%p\" CONTENT=\"%p\" SUCC=\"%p\" PRED=\"\">", 
		   (void *) tmp, (void *) _gdsl_list_node_get_content (tmp), 
		   (void *) _gdsl_list_node_get_succ (tmp));
	}
      else if (tmp == _gdsl_list_node_get_pred (q->z))
	{
	  fprintf (file, "<GDSL_QUEUE_NODE REF=\"%p\" CONTENT=\"%p\" SUCC=\"\" PRED=\"%p\">", 
		   (void *) tmp, (void *) _gdsl_list_node_get_content (tmp), 
		   (void *) _gdsl_list_node_get_pred (tmp));
	}
      else
	{
	  fprintf (file, "<GDSL_QUEUE_NODE REF=\"%p\" CONTENT=\"%p\" SUCC=\"%p\" PRED=\"%p\">", 
		   (void *) tmp, (void *) _gdsl_list_node_get_content (tmp), 
		   (void *) _gdsl_list_node_get_succ (tmp), 
		   (void *) _gdsl_list_node_get_pred (tmp));
	}

      if (write_f != NULL && _gdsl_list_node_get_content (tmp) != NULL)
	_gdsl_list_node_write (tmp, write_f, file, user_data);

      fprintf (file, "</GDSL_QUEUE_NODE>\n");

      tmp = _gdsl_list_node_get_succ (tmp);
    }

  fprintf (file, "</GDSL_QUEUE>\n");
}

extern void
gdsl_queue_dump (gdsl_queue_t q, gdsl_write_func_t write_f, FILE *file, 
		 void *user_data)
{
  _gdsl_list_node_t tmp = _gdsl_list_node_get_succ (q->d);

  fprintf (file, "<GDSL_QUEUE REF=\"%p\" NAME=\"%s\" CARD=\"%ld\" HEAD=\"%p\" TAIL=\"%p\">\n", 
	   (void *) q, q->name, q->card, (void *) q->d, (void *) q->z);

  fprintf (file, "<GDSL_QUEUE_HEAD REF=\"%p\" SUCC=\"%p\">\n", 
	   (void *) q->d, (void *) _gdsl_list_node_get_succ (q->d));

  while (tmp != q->z)
    {
      if (_gdsl_list_node_get_content (tmp) != NULL)
	{
	  fprintf (file, "<GDSL_QUEUE_NODE REF=\"%p\" CONTENT=\"%p\" SUCC=\"%p\" PRED=\"%p\">", 
		   (void *) tmp, (void *) _gdsl_list_node_get_content (tmp), 
		   (void *) _gdsl_list_node_get_succ (tmp), 
		   (void *) _gdsl_list_node_get_pred (tmp));
	}
      else
	{
	  fprintf (file, "<GDSL_QUEUE_NODE REF=\"%p\" CONTENT=\"\" SUCC=\"%p\" PRED=\"%p\">", 
		   (void *) tmp, (void *) _gdsl_list_node_get_succ (tmp), 
		   (void *) _gdsl_list_node_get_pred (tmp));
	}
      
      if (write_f != NULL && _gdsl_list_node_get_content (tmp) != NULL)
	{
	  fprintf (file, "\n");
	  _gdsl_list_node_write (tmp, write_f, file, user_data);
	  fprintf (file, "\n");
	}
      
      fprintf (file, "</GDSL_QUEUE_NODE>\n");

      tmp = _gdsl_list_node_get_succ (tmp);
    }

  fprintf (file, "<GDSL_QUEUE_TAIL REF=\"%p\" PRED=\"%p\">\n</GDSL_QUEUE>\n", 
	   (void *) q->z, (void *) _gdsl_list_node_get_pred (q->z));
}

/******************************************************************************/
/* Private functions                                                          */
/******************************************************************************/

static gdsl_element_t
default_alloc (void *e)
{
  return e;
}

static void 
default_free (gdsl_element_t e)
{
  ;
}

static _gdsl_list_node_t
search_by_key (gdsl_queue_t q, gdsl_compare_func_t comp_f, void *key)
{
  _gdsl_list_node_t tmp = _gdsl_list_node_get_succ (q->d);

  while (tmp != q->z)
    {
      if (comp_f( _gdsl_list_node_get_content (tmp), key) == 0)
	return tmp;

      tmp = _gdsl_list_node_get_succ (tmp);
    }

  return NULL;
}
