/*
 * This file is part of Generic Data Structures Library (GDSL).
 * Copyright (C) 1998-2001 Nicolas Darnis <ndarnis@free.fr>
 *
 * The GDSL library is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * The GDSL library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the GDSL library; see the file COPYING.
 * If not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 * File: gdsl_rbtree.c
 * $Revision: 1.3 $
 * $Date: 2002/02/21 09:55:51 $
 * $Author: ndarnis $
 * $Id: gdsl_rbtree.c,v 1.3 2002/02/21 09:55:51 ndarnis Exp $
 */

#include <config.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "_gdsl_bintree.h"
#include "_gdsl_bstree.h"
#include "gdsl_rbtree.h"
#include "gdsl_types.h"
#include "_gdsl_mem.h"

typedef enum { RED, BLACK } gdsl_rbtree_node_color_t;

struct gdsl_rbtree_node
{
  struct gdsl_rbtree_node *left;    /* Node's left sub-tree */
  struct gdsl_rbtree_node *right;   /* Node's right sub-tree */
  struct gdsl_rbtree_node *parent;  /* Node's parent */
  gdsl_element_t           content; /* Node's content */
  gdsl_rbtree_node_color_t color;   /* Node's color */
};
typedef struct gdsl_rbtree_node *gdsl_rbtree_node_t;

#define LEFT(n)           ( (n)->left )
#define RIGHT(n)          ( (n)->right )
#define PARENT(n)         ( (n)->parent )
#define CONTENT(n)        ( (n)->content )
#define COLOR(n)          ( (n)->color )

struct gdsl_rbtree
{
  char *                  name;     /* Tree's name */
  ulong                   card;     /* Tree's cardinality */
  struct gdsl_rbtree_node sent;     /* Tree's sentinel */
  gdsl_alloc_func_t       alloc_f;  /* */
  gdsl_free_func_t        free_f;   /* */
  gdsl_compare_func_t     comp_f;   /* */
};

#define SENT(t)           ( &(t)->sent )
#define ROOT(t)           ( (t)->sent.right )

static gdsl_element_t 
default_alloc (void *v);

static void 
default_free (gdsl_element_t e);

static int 
default_compare (gdsl_element_t e, void *v);

static gdsl_rbtree_node_t
rbtree_node_alloc (gdsl_rbtree_t t, gdsl_element_t e);

static void
rbtree_node_free (gdsl_rbtree_node_t n);

static void 
rbtree_destroy (gdsl_rbtree_node_t n, gdsl_rbtree_node_t sent, 
		gdsl_free_func_t f);

static ulong
rbtree_size (gdsl_rbtree_node_t n, gdsl_rbtree_node_t sent);

static ulong
rbtree_height (gdsl_rbtree_node_t n, gdsl_rbtree_node_t sent);

static gdsl_rbtree_node_t
rbtree_left_rot (gdsl_rbtree_node_t n);

static gdsl_rbtree_node_t
rbtree_right_rot (gdsl_rbtree_node_t n);

static gdsl_rbtree_node_t
rbtree_search (gdsl_rbtree_node_t n, gdsl_rbtree_node_t sent, 
	       gdsl_compare_func_t f, void *v);

static gdsl_rbtree_node_t
rbtree_next (gdsl_rbtree_t t, gdsl_rbtree_node_t n);

static gdsl_element_t
rbtree_prefix_parse (gdsl_rbtree_node_t root, gdsl_rbtree_node_t sent, gdsl_map_func_t f, void *d);

static gdsl_element_t
rbtree_infix_parse (gdsl_rbtree_node_t root, gdsl_rbtree_node_t sent, gdsl_map_func_t f, void *d);

static gdsl_element_t
rbtree_postfix_parse (gdsl_rbtree_node_t root, gdsl_rbtree_node_t sent, gdsl_map_func_t f, void *d);

static void
rbtree_write (gdsl_rbtree_node_t n, gdsl_rbtree_node_t sent, gdsl_write_func_t f, 
	      FILE *file, void *d);

static void
rbtree_write_xml (gdsl_rbtree_node_t n, gdsl_rbtree_node_t sent, 
		  gdsl_write_func_t f, FILE *file, void *d);

static void
rbtree_dump (gdsl_rbtree_node_t n, gdsl_rbtree_node_t sent, gdsl_write_func_t f, 
	     FILE *file, void *d);

/******************************************************************************/
/* Allocate and deallocate functions                                          */
/******************************************************************************/

extern gdsl_rbtree_t
gdsl_rbtree_create (const char *name, gdsl_alloc_func_t alloc_f, 
		    gdsl_free_func_t free_f, gdsl_compare_func_t comp_f)
{
  gdsl_rbtree_t t;

  t = (gdsl_rbtree_t) _gdsl_mem_alloc (sizeof (struct gdsl_rbtree));
  if (t == NULL)
    return NULL;

  t->name = NULL;

  if (name != NULL)
    {
      t->name = (char *) _gdsl_mem_alloc ((1 + strlen (name)) * sizeof (char));
      if (t->name == NULL)
	{
	  _gdsl_mem_free (t);
	  return NULL;
	}
      strcpy (t->name, name);
    }

  t->alloc_f = alloc_f ? alloc_f : default_alloc;
  t->free_f  = free_f  ? free_f  : default_free;
  t->comp_f  = comp_f  ? comp_f  : default_compare;

  t->card = 0UL;

  t->sent.left   = SENT (t);
  t->sent.right  = SENT (t);
  t->sent.parent = SENT (t);
  t->sent.color  = BLACK;

  return t;
}

extern void 
gdsl_rbtree_destroy (gdsl_rbtree_t t)
{
  rbtree_destroy (ROOT (t), SENT (t), t->free_f);
  
  if (t->name != NULL)
    _gdsl_mem_free (t->name);

  _gdsl_mem_free (t);
}

extern void
gdsl_rbtree_flush (gdsl_rbtree_t t)
{
  rbtree_destroy (ROOT (t), SENT (t), t->free_f);

  t->card = 0UL;

  t->sent.left = SENT (t);
  t->sent.right = SENT (t);
  t->sent.parent = SENT (t);
  t->sent.color  = BLACK;
}

#ifdef _FOR_FUTURE_USE_
extern gdsl_bstree_t
gdsl_rbtree_copy (gdsl_rbtree_t t)
{
  /* !! TO DO... !! */
  return NULL;
}
#endif

extern gdsl_rbtree_t
gdsl_rbtree_set_name (gdsl_rbtree_t t, const char *name)
{
  if (t->name != NULL)
    {
      _gdsl_mem_free (t->name);
      t->name = NULL;
    }

  if (name != NULL)
    {
      t->name = (char *) _gdsl_mem_alloc ((1 + strlen (name)) * sizeof (char));
      if (t->name != NULL)
	return NULL;
      strcpy (t->name, name);
    }

  return t;
}

/******************************************************************************/
/* Information functions                                                      */
/******************************************************************************/

extern char *
gdsl_rbtree_get_name (gdsl_rbtree_t t)
{
  return t->name;
}

extern bool
gdsl_rbtree_is_empty (gdsl_rbtree_t t)
{
  return (bool) (t->card == 0); /* alt. ROOT( t ) == SENT( t ) */
}

extern gdsl_element_t
gdsl_rbtree_get_root (gdsl_rbtree_t t)
{
  return CONTENT (ROOT (t));
}

extern ulong
gdsl_rbtree_get_size (gdsl_rbtree_t t)
{
  return t->card;
}

extern ulong
gdsl_rbtree_height (gdsl_rbtree_t t)
{
  return rbtree_height (ROOT (t), SENT (t));
}

/******************************************************************************/
/* Inserting/removing functions                                               */
/******************************************************************************/

extern gdsl_element_t
gdsl_rbtree_insert (gdsl_rbtree_t t, void *v)
{
  int comp = -1;
  gdsl_element_t e;
  gdsl_rbtree_node_t root;
  gdsl_rbtree_node_t parent;
  gdsl_rbtree_node_t n;

  /* First, we do a classic binary search tree insertion: */
  root = ROOT (t);
  parent = SENT (t);
  while (root != SENT (t))
    {
      parent = root;
      comp = t->comp_f (CONTENT (root), v);

      /* Duplicates keys are not allowed (must be changed) */
      if (comp == 0)
	return NULL;

      root = (comp > 0) ? LEFT (root) : RIGHT (root); 
    }

  /* Then, we create the new node n and we insert it into t: */
  e = (t->alloc_f) (v);
  if (e == NULL)
    return NULL;

  n = rbtree_node_alloc (t, e);
  if (n == NULL)
    {
      t->free_f (e);
      return NULL;
    }

  /* Insertion of n into t: */
  n->parent = parent;

  if (comp > 0)
    parent->left = n;
  else
    parent->right = n;

  t->card++;

  /* Finaly, we do red-black specific adjustments: */
  {
    gdsl_rbtree_node_t uncle = NULL;
    gdsl_rbtree_node_t gparent = NULL;

    while (COLOR (parent) == RED)
      {
	gparent = PARENT (parent);
	if (parent == LEFT (gparent))
	  {
	    uncle = RIGHT (gparent);
	    if (COLOR (uncle) == RED)
	      {
		parent->color = BLACK;
		uncle->color = BLACK;
		gparent->color = RED;
		n = gparent;
		parent = PARENT (gparent);
	      }
	    else
	      {
		if (n == RIGHT (parent))
		  {
		    rbtree_left_rot (parent);
		    parent = n;
		  }
		parent->color = BLACK;
		gparent->color = RED;
		rbtree_right_rot (gparent);
		break;
	      }
	  }
	else
	  {
	    uncle = LEFT (gparent);
	    if (COLOR (uncle) == RED)
	      {
		parent->color = BLACK;
		uncle->color = BLACK;
		gparent->color = RED;
		n = gparent;
		parent = PARENT (gparent);
	      }
	    else
	      {
		if (n == LEFT (parent))
		  {
		    rbtree_right_rot (parent);
		    parent = n;
		  }
		parent->color = BLACK;
		gparent->color = RED;
		rbtree_left_rot (gparent);
	      }
	  }
      }
    }

  t->sent.right->color = BLACK;
  return e;
}

extern gdsl_element_t
gdsl_rbtree_remove_by_value (gdsl_rbtree_t t, void *v)
{
  gdsl_element_t     e;
  gdsl_rbtree_node_t n;
  gdsl_rbtree_node_t child;

  /* First, we search into t for a node n containing v: */
  n = rbtree_search (ROOT (t), SENT (t), t->comp_f, v);
  if (n == NULL)
    return NULL;

  /* Then, we do a classical removing of a node n from a binary search tree t: */
  if (LEFT (n) != SENT (t) && RIGHT (n) != SENT (t))
    {
      gdsl_rbtree_node_t next = rbtree_next (t, n);
      gdsl_rbtree_node_t nextparent = PARENT (next);
      gdsl_rbtree_node_color_t nextcolor = COLOR (next);

      child = RIGHT (next);
      child->parent = nextparent;

      if (LEFT (nextparent) == next)
	nextparent->left = child;
      else
	nextparent->right = child;

      next->parent = PARENT (n);
      next->left = LEFT (n);
      next->right = RIGHT (n);
      next->left->parent = next;
      next->right->parent = next;
      next->color = COLOR (n);
      n->color = nextcolor;
      
      if (LEFT (PARENT (n)) == n)
	n->parent->left = next;
      else
	n->parent->right = next;
    }
  else
    {
      child = LEFT (n) != SENT (t) ? LEFT (n) : RIGHT (n);

      child->parent = PARENT (n);

      if (n == LEFT (PARENT (n)))
	n->parent->left = child;
      else
	n->parent->right = child;
    }

  t->card--;

  /* Finaly, we do red-black specific adjustments: */
  if (COLOR (n) == BLACK)
    {
      gdsl_rbtree_node_t parent, sister;

      t->sent.right->color = RED;

      while (COLOR (child) == BLACK)
	{
	  parent = PARENT (child);
	  if (child == LEFT (parent))
	    {
	      sister = RIGHT (parent);
	      if (COLOR (sister) == RED)
		{
		  sister->color = BLACK;
		  parent->color = RED;
		  rbtree_left_rot (parent);
		  sister = RIGHT (parent);
		}
	      if (COLOR (LEFT (sister)) == BLACK && COLOR (RIGHT (sister)) == BLACK)
		{
		  sister->color = RED;
		  child = parent;
		}
	      else
		{
		  if (COLOR (RIGHT (sister)) == BLACK)
		    {
		      sister->left->color = BLACK;
		      sister->color = RED;
		      rbtree_right_rot (sister);
		      sister = RIGHT (parent);
		    }
		  sister->color = COLOR (parent);
		  sister->right->color = BLACK;
		  parent->color = BLACK;
		  rbtree_left_rot (parent);
		  break;
		}
	    }
	  else
	    {
	      sister = LEFT (parent);
	      if (COLOR (sister) == RED)
		{
		  sister->color = BLACK;
		  parent->color = RED;
		  rbtree_right_rot (parent);
		  sister = LEFT (parent);
		}
	      if (COLOR (RIGHT (sister)) == BLACK && COLOR (LEFT (sister)) == BLACK)
		{
		  sister->color = RED;
		  child = parent;
		}
	      else
		{
		  if (COLOR (LEFT (sister)) == BLACK)
		    {
		      sister->right->color = BLACK;
		      sister->color = RED;
		      rbtree_left_rot (sister);
		      sister = LEFT (parent);
		    }
		  sister->color = COLOR (parent);
		  sister->left->color = BLACK;
		  parent->color = BLACK;
		  rbtree_right_rot (parent);
		  break;
		}
	    }
	}

      child->color = BLACK;
      t->sent.right->color = BLACK;
    }

  e = CONTENT (n);
  rbtree_node_free (n);

  return e;
}

extern gdsl_rbtree_t
gdsl_rbtree_destroy_by_value (gdsl_rbtree_t t, void *v)
{
  gdsl_element_t e;

  e = gdsl_rbtree_remove_by_value (t, v);
  if (e == NULL)
    return NULL;

  t->free_f (e);

  return t;
}

/******************************************************************************/
/* Research functions                                                         */
/******************************************************************************/

extern gdsl_element_t
gdsl_rbtree_search (gdsl_rbtree_t t, gdsl_compare_func_t f, void *v)
{
  gdsl_rbtree_node_t n;

  n = rbtree_search (ROOT (t), SENT (t), f ? f : t->comp_f, v);

  return n == NULL ? NULL : CONTENT (n);
}

/******************************************************************************/
/* Parse functions                                                            */
/******************************************************************************/

extern gdsl_element_t
gdsl_rbtree_prefix_parse (gdsl_rbtree_t t, gdsl_map_func_t f, void *d)
{
  return rbtree_prefix_parse (ROOT (t), SENT (t), f, d);
}

extern gdsl_element_t
gdsl_rbtree_infix_parse (gdsl_rbtree_t t, gdsl_map_func_t f, void *d)
{
  return rbtree_infix_parse (ROOT (t), SENT (t), f, d);
}

extern gdsl_element_t
gdsl_rbtree_postfix_parse (gdsl_rbtree_t t, gdsl_map_func_t f, void *d)
{
  return rbtree_postfix_parse (ROOT (t), SENT (t), f, d);
}

/******************************************************************************/
/* Input/output functions                                                     */
/******************************************************************************/

extern void
gdsl_rbtree_write (gdsl_rbtree_t t, gdsl_write_func_t f, FILE *file, void *d)
{
  rbtree_write (ROOT (t), SENT (t), f, file, d);
}

extern void
gdsl_rbtree_write_xml (gdsl_rbtree_t t, gdsl_write_func_t f, FILE *file, 
		       void *d)
{
  fprintf (file, "<GDSL_RBTREE REF=\"%p\" NAME=\"%s\" CARD=\"%ld\">\n", 
	   (void *) t, t->name, t->card);

  rbtree_write_xml (ROOT (t), SENT (t), f, file, d);

  fprintf (file, "</GDSL_RBTREE>\n");
}

extern void
gdsl_rbtree_dump (gdsl_rbtree_t t, gdsl_write_func_t f, FILE *file, void *d)
{
  fprintf (file, "<GDSL_RBTREE REF=\"%p\" NAME=\"%s\" CARD=\"%ld\">\n", 
	   (void *) t, t->name, t->card);
  fprintf (file, "<GDSL_RBTREE_SENT REF=\"%p\" LEFT=\"%p\" RIGHT=\"%p\" PARENT=\"%p\"", 
	   (void *)SENT( t ), (void *)t->sent.left, (void *)t->sent.right, (void *)PARENT( SENT( t ) ) );
  fprintf (file, " COLOR=\"%s\"/>\n", COLOR (SENT (t)) == RED ? "RED" : "BLACK");

  rbtree_dump (ROOT (t), SENT (t), f, file, d);

  fprintf (file, "</GDSL_RBTREE>\n");
}

/******************************************************************************/
/* Private functions                                                          */
/******************************************************************************/

static gdsl_element_t 
default_alloc (void *v)
{
  return v;
}

static void 
default_free (gdsl_element_t e)
{
  ;
}

static int 
default_compare (gdsl_element_t e, void *v)
{
  return 0;
}

static gdsl_rbtree_node_t
rbtree_node_alloc (gdsl_rbtree_t t, gdsl_element_t e)
{
  gdsl_rbtree_node_t n;

  n = (gdsl_rbtree_node_t) _gdsl_mem_alloc (sizeof (struct gdsl_rbtree_node));
  if (n == NULL)
    return NULL;
  
  n->left    = SENT (t);
  n->right   = SENT (t);
  n->parent  = SENT (t);
  n->content = e;
  n->color   = RED;

  return n;
}

static void
rbtree_node_free (gdsl_rbtree_node_t n)
{
  _gdsl_mem_free (n);
}

static void 
rbtree_destroy (gdsl_rbtree_node_t n, gdsl_rbtree_node_t sent, gdsl_free_func_t f)
{
  if (n != sent)
    {
      rbtree_destroy (LEFT (n), sent, f);
      rbtree_destroy (RIGHT (n), sent, f);
      f (CONTENT (n));
      _gdsl_mem_free (n);
    }
}

static ulong
rbtree_size (gdsl_rbtree_node_t n, gdsl_rbtree_node_t sent)
{
  if (n == sent)
    return 0UL;

  return (ulong) (1UL 
		  + rbtree_size (LEFT (n), sent)
		  + rbtree_size (RIGHT (n), sent));
}

static ulong
rbtree_height (gdsl_rbtree_node_t n, gdsl_rbtree_node_t sent)
{
  if (n == sent)
    return 0UL;

  if (LEFT (n) == sent && RIGHT (n) == sent)
    return 0UL;

  return (ulong) (1UL + 
		  MAX (rbtree_height (LEFT (n), sent),
		       rbtree_height (RIGHT (n), sent)));
}

static gdsl_rbtree_node_t
rbtree_left_rot (gdsl_rbtree_node_t n)
{
  gdsl_rbtree_node_t rn;

  rn = RIGHT (n);
  n->right = LEFT (rn);

  /* LEFT( rn ) is always != NULL */
  rn->left->parent = n;

  rn->parent = PARENT (n);
  
  if (n == LEFT (PARENT (n)))
    n->parent->left = rn;
  else
    n->parent->right = rn;

  rn->left = n;
  n->parent = rn;

  return rn;
}

static gdsl_rbtree_node_t
rbtree_right_rot (gdsl_rbtree_node_t n)
{
  gdsl_rbtree_node_t ln;

  ln = LEFT (n);
  n->left = RIGHT (ln);

  /* RIGHT( ln ) is always != NULL */
  ln->right->parent = n;

  ln->parent = PARENT (n);

  if (n == RIGHT (PARENT (n)))
    n->parent->right = ln;
  else
    n->parent->left = ln;

  ln->right = n;
  n->parent = ln;

  return ln;
}

static gdsl_rbtree_node_t
rbtree_search (gdsl_rbtree_node_t root, gdsl_rbtree_node_t sent, 
	       gdsl_compare_func_t f, void *v)
{
  int comp;

#if defined (_GDSL_OPTIMIZE_TIME_)
  while (root != sent)
    {
      comp = f (CONTENT (root), v);

      if (comp == 0)
	return root;

      if (comp > 0)
	root = LEFT (root);
      else
	root = RIGHT (root);
    }

  return NULL;
#else
  if (root == sent)
    return NULL;

  comp = f (CONTENT (root), v);

  if (comp == 0)
    return root;

  if (comp > 0)
    return rbtree_search (LEFT (root), sent, f, v);
  
  return rbtree_search (RIGHT (root), sent, f, v);
#endif
}

static gdsl_rbtree_node_t
rbtree_next (gdsl_rbtree_t t, gdsl_rbtree_node_t n)
{
  n = RIGHT (n);
  
  while (LEFT (n) != SENT (t))
    n = LEFT (n);

  return n;
}

static gdsl_element_t
rbtree_prefix_parse (gdsl_rbtree_node_t root, gdsl_rbtree_node_t sent, 
		     gdsl_map_func_t f, void *d)
{
  if (root != sent)
    {
      gdsl_element_t e = CONTENT (root);
      if (f (e, d) == 0)
	return e;

      rbtree_prefix_parse (LEFT (root), sent, f, d);
      rbtree_prefix_parse (RIGHT (root), sent, f, d);
    }

  return NULL;
}

static gdsl_element_t
rbtree_infix_parse (gdsl_rbtree_node_t root, gdsl_rbtree_node_t sent, 
		    gdsl_map_func_t f, void *d)
{
  if (root != sent)
    {
      gdsl_element_t e;

      rbtree_infix_parse (LEFT (root), sent, f, d);

      e = CONTENT (root);
      if (f (e, d) == 0)
	return e;

      rbtree_infix_parse (RIGHT (root), sent, f, d);
    }

  return NULL;
}

static gdsl_element_t
rbtree_postfix_parse (gdsl_rbtree_node_t root, gdsl_rbtree_node_t sent, 
		      gdsl_map_func_t f, void *d)
{
  if (root != sent)
    {
      gdsl_element_t e;

      rbtree_postfix_parse (LEFT (root), sent, f, d);
      rbtree_postfix_parse (RIGHT (root), sent, f, d);

      e = CONTENT (root);
      if (f (e, d) == 0)
	return e;
    }

  return NULL;
}

static void
rbtree_write (gdsl_rbtree_node_t n, gdsl_rbtree_node_t sent, 
	      gdsl_write_func_t f, FILE *file, void *d)
{
  if (n != sent)
    {
      f (CONTENT (n), file, d);
      rbtree_write (LEFT (n), sent, f, file, d);
      rbtree_write (RIGHT (n), sent, f, file, d);
    }
}

static void
rbtree_write_xml (gdsl_rbtree_node_t n, gdsl_rbtree_node_t sent, 
		  gdsl_write_func_t f, FILE *file, void *d)
{
  if (n != sent)
    {
      if (LEFT (n) == sent && RIGHT (n) == sent)
	fprintf (file, "<GDSL_RBTREE_LEAF REF=\"%p\"", (void *) n);
      else
	fprintf (file, "<GDSL_RBTREE_NODE REF=\"%p\"", (void *) n);

      if (LEFT (n) != sent || RIGHT (n) != sent)
	{
	  if (LEFT (n) != sent)
	    fprintf (file, " LEFT=\"%p\"", (void *) LEFT (n));
	  else
	    fprintf (file, " LEFT=\"\"");
	  
	  if (RIGHT (n) != sent)
	    fprintf (file, " RIGHT=\"%p\"", (void *) RIGHT (n));
	  else
	    fprintf (file, " RIGHT=\"\"");
	}

      if (PARENT (n) != sent)
	fprintf (file, " PARENT=\"%p\"", (void *) PARENT (n));
      else
	fprintf (file, " PARENT=\"\"");

      fprintf (file, " COLOR=\"%s\">", COLOR (n) == RED ? "RED" : "BLACK");

      if (f != NULL && CONTENT (n) != NULL)
	f (CONTENT (n), file, d);

      if (LEFT (n) == sent && RIGHT (n) == sent)
	fprintf (file, "</GDSL_RBTREE_LEAF>\n");
      else
	fprintf (file, "</GDSL_RBTREE_NODE>\n");

      rbtree_write_xml (LEFT (n), sent, f, file, d);
      rbtree_write_xml (RIGHT (n), sent, f, file, d);
    }
}

static void
rbtree_dump (gdsl_rbtree_node_t n, gdsl_rbtree_node_t sent, gdsl_write_func_t f, 
	     FILE *file, void *d)
{
  if (n != sent)
    {
      fprintf (file, "<GDSL_RBTREE_NODE REF=\"%p\"", (void *) n);

      if (CONTENT (n))
	fprintf (file, " CONTENT=\"%p\"", (void *) CONTENT (n));
      else
	fprintf (file, " CONTENT=\"\"");

      fprintf (file, " LEFT=\"%p\" RIGHT=\"%p\"", (void *) LEFT (n), 
	       (void *) RIGHT (n));

      if (PARENT (n))
	fprintf (file, " PARENT=\"%p\"", (void *) PARENT (n));
      else
	fprintf (file, " PARENT=\"\"");

      fprintf (file, " COLOR=\"%s\">", COLOR (n) == RED ? "RED" : "BLACK");

      if (f != NULL && CONTENT (n) != NULL)
	f (CONTENT (n), file, d);

      fprintf (file, "</GDSL_RBTREE_NODE>\n");

      rbtree_dump (LEFT (n), sent, f, file, d);
      rbtree_dump (RIGHT (n), sent, f, file, d);
    }
}

#ifdef HAVE_GFX_DEBUG
#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#define X_Win_pos 0
#define Y_Win_pos 0

#define RAYON 6
#define BORD  3 * RAYON
#define DX  3 * RAYON
#define DY 6 * RAYON

#define COLOR1   "grey"
#define COLOR2   "blue"
#define COLOR3   "red"

/* declaration des identificateurs de ressources X */
Window Fen;                      /* la fenetre Globale    */
GC     ContGraph;                /* contexte graphique    */

/* declaration des variables utilisees par les fonctions X */
Display                *display;
XEvent                 evenement;            /* file des evenements */
XSizeHints             indic;
int                    ecran;
unsigned long          noir,blanc;
Colormap               Cmap;
XColor                 coul1;   /* meilleure approximation RVB rendue */
XColor                 coul2;
XColor                 coul3;
XColor                 exact_c; /* couleur RVB exacte correspondant au nom */
Font                   fonte;

/*Declaration des variabes propres a l'arbre */
int Xo,Yo,X,Y;
static bool InitContexte( int LFen, int HFen );
static void CloseFen( void );
static bool Touche( KeySym tch );
static void TraceABic( gdsl_rbtree_node_t A, gdsl_rbtree_node_t sent, int x, int y, char * (*key_display)(gdsl_element_t) );
#endif /* HAVE_GFX_DEBUG */
extern void 
gdsl_rbtree_gfx_display (gdsl_rbtree_t A, char * (*key_display) (gdsl_element_t))
{
#ifdef HAVE_GFX_DEBUG
  int encore = TRUE;
  int Lfen, Hfen;      
  KeySym touche;
  gdsl_rbtree_node_t t = ROOT(A);
  gdsl_rbtree_node_t sl = A->sent.left;
  gdsl_rbtree_node_t sr = A->sent.right;

  A->sent.left = A->sent.right = NULL;

  if( t != SENT(A) ) 
    {
      Lfen = ( A->card - 1 ) * DX + 2 * BORD;  /* largeur de la Fen Globale     */
      Hfen = rbtree_height( t, SENT( A ) ) * DY + 2 * BORD;     /* Hauteur de la Fen GLOBALE     */
      Xo = rbtree_size( LEFT( t ), SENT( A ) ) * DX + BORD;   /* abscisse de la racine de l'arbre  */
      X = Xo; Yo = BORD;Y = Yo;
      if ( InitContexte( Lfen, Hfen ) )
	{   
	  /* BOUCLE PRINCIPALE */
	  while( encore )
	    {
	      /*lecture de la file d'attente des evenements */
	      XNextEvent( display, &evenement );
	      switch( evenement.type )
		{
		case Expose: 
		  /* apres le dernier evenement du type expose(affichage complet de la Fen) */
		  if( evenement.xexpose.count == 0 )
		    TraceABic( t, SENT( A ), X, Y, key_display );
		  break;

		case KeyPress:
		  touche = XLookupKeysym( (XKeyEvent *)&evenement, 0 );
		  if( ( encore = Touche( touche ) ) )
		    {
		      XClearArea(display, Fen, 0, 0, 0, 0, TRUE );
		      TraceABic( t, SENT( A ), X, Y, key_display );
		    }
		}
	    }
	}
      CloseFen(); 
    }

  A->sent.left = sl;
  A->sent.right = sr;
#endif /* HAVE_GFX_DEBUG */
}

#ifdef HAVE_GFX_DEBUG
static bool 
InitContexte( int LFen, int HFen )
{
  /* demande dans la base de donnee          */
  /* connexion au serveur X                  */
  display = XOpenDisplay( NULL );

  if ( display == NULL ) 
    {
      printf( "Impossible d'ouvrir le display.\n" );
      return( FALSE );
    }
  
  /* consultation des caracteristiques du serveur X      */
  ecran = DefaultScreen( display );
  blanc = WhitePixel( display, ecran );
  noir = BlackPixel( display, ecran );
  
  /* definition de deux couleurs                         */
  Cmap = DefaultColormap( display, DefaultScreen( display ) );
  if( XAllocNamedColor( display, Cmap, COLOR1, &coul1, &exact_c ) == 0 )
    coul1.pixel = noir;
  
  if( XAllocNamedColor( display, Cmap, COLOR2, &coul2, &exact_c ) == 0 )
    coul2.pixel = blanc;

  if( XAllocNamedColor( display, Cmap, COLOR3, &coul3, &exact_c ) == 0 )
    coul3.pixel = noir;
  
  /* Creation de la Fen top-niveau du nom de "GLOBALE" */
  indic.x = X_Win_pos;
  indic.y = Y_Win_pos;
  indic.width = LFen;
  indic.height = HFen;
  indic.flags = PPosition | PSize;
  
  Fen = XCreateSimpleWindow( display, DefaultRootWindow( display ),
			     indic.x, indic.y, indic.width, indic.height,
			     4, coul2.pixel, coul1.pixel );
  
  /* definition des proprietes standard de la Fen Globale pour gestionnaire X */
  XSetStandardProperties( display, Fen, "Arbre bicolore", "Arbre bicolore", None, NULL, 0, &indic );
  
  /* Choix des evenements qui nous interressent en ce qui concerne la Fen */
  XSelectInput( display, Fen, KeyPressMask | ExposureMask ); 
  
  /* projection de toutes la Fen pour la rendre visible            */
  /* l'action "visible" est dans le tampon, mais non encore effectuee  */
  /* les actions s'effectuent avec la fonction XNextEvent, par exemple */
  XMapRaised( display, Fen );

  /* Creation du contexte graphique necessaire pour les requetes d'affichage  */
  ContGraph = XCreateGC( display, Fen, 0, NULL );
  XSetBackground( display, ContGraph, coul1.pixel) ;
  XSetForeground( display, ContGraph, coul2.pixel );

  /* Selection de la police */
  fonte = XLoadFont( display, "6x10" );
  XSetFont( display, ContGraph, fonte );

  return TRUE;
}  

/*----------------------------*/

static void 
CloseFen( void )
{
  /* liberation des ressources de la Fen */
  XFreeGC( display, ContGraph );
  XDestroyWindow( display, Fen );
  XCloseDisplay( display );
}

/*----------------------------*/

static bool 
Touche( KeySym tch )
{
  switch( tch )
    {
    case XK_Up: 
      Y += 100;   
      break;
      
    case XK_Right:
      X -= 100;   
      break;

    case XK_Down:
      Y -= 100;   
      break;

    case XK_Left:
      X += 100;   
      break;
      
    case XK_Return:
      X = Xo;
      Y = Yo;
      break;

    case XK_space:
      return FALSE;
      break; 
    }

  return TRUE;
}

/*------------------------------*/

static void 
TraceABic( gdsl_rbtree_node_t A, gdsl_rbtree_node_t sent, int x, int y, char * (*key_display)(gdsl_element_t) )
{
  int Xg, Xd, Ygd;
  char chr[35];
  
  if( x >= 0 && y >=0 )
    { 
      /* condition lie a des problemes d'affichage hors fenetre... */
      XSetForeground(display, ContGraph, coul2.pixel); 
      snprintf( chr, 34, "%s", key_display( CONTENT( A ) ) );
      XDrawString(display, Fen, ContGraph,x-RAYON ,y+RAYON ,chr ,strlen(chr));
    }
  
  Ygd = y + DY;
  
  if( LEFT(A) != sent )
    {
      Xg = x - ( 1 + rbtree_size( RIGHT( LEFT( A ) ), sent ) ) * DX;
      if( COLOR( LEFT( A ) ) == RED )
	{
	  XSetLineAttributes(display, ContGraph, 3, LineSolid, CapNotLast, JoinMiter);
	  XSetForeground(display, ContGraph, coul3.pixel);
	}
      else
	{
	  XSetLineAttributes(display, ContGraph, 1, LineSolid, CapNotLast, JoinMiter);
	  XSetForeground(display, ContGraph, noir);
	}
      XDrawLine(display, Fen, ContGraph, x-RAYON, y+RAYON, Xg+RAYON, Ygd-RAYON );
      TraceABic( LEFT( A ), sent, Xg, Ygd, key_display );
    }

  if ( RIGHT( A ) != sent )
    {
      Xd = x + ( 1 + rbtree_size( LEFT( RIGHT( A ) ), sent ) ) * DX;
      if( COLOR( RIGHT( A ) ) == RED )
	{
	  XSetLineAttributes(display, ContGraph, 3, LineSolid, CapNotLast, JoinMiter);
	  XSetForeground(display, ContGraph, coul3.pixel);
	}
      else
	{
	  XSetLineAttributes(display, ContGraph, 1, LineSolid, CapNotLast, JoinMiter);
	  XSetForeground(display, ContGraph, noir);
	}
      XDrawLine(display, Fen, ContGraph, x+RAYON, y+RAYON, Xd-RAYON, Ygd-RAYON);
      TraceABic( RIGHT( A ), sent, Xd, Ygd, key_display );
    }
}
#endif /* HAVE_GFX_DEBUG */

