/*
 * This file is part of Generic Data Structures Library (GDSL).
 * Copyright (C) 1998-2001 Nicolas Darnis <ndarnis@free.fr>
 *
 * The GDSL library is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * The GDSL library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the GDSL library; see the file COPYING.
 * If not, write to the Free Software Foundation, Inc., 
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
 *
 * File: gdsl_stack.c
 * $Revision: 1.3 $
 * $Date: 2002/02/21 09:55:51 $
 * $Author: ndarnis $
 * $Id: gdsl_stack.c,v 1.3 2002/02/21 09:55:51 ndarnis Exp $
 */

#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "_gdsl_list.h"
#include "_gdsl_list_node.h"
#include "gdsl_types.h"
#include "gdsl_stack.h"
#include "_gdsl_mem.h"

#define _GDSL_STACK_GROWING_FACTOR 30

struct _gdsl_stack
{
  char *            name;  /* Name of the stack */
  ulong             card;  /* Cardinality of the stack */
  ulong             size;  /* Size of the stack */
  gdsl_element_t *  nodes; /* Elements of the stack */

  /* Functions pointers */
  gdsl_alloc_func_t alloc_f;
  gdsl_free_func_t  free_f;
};

static gdsl_element_t default_alloc (void *e);
static void default_free (gdsl_element_t e);

/******************************************************************************/
/* Management functions                                                       */
/******************************************************************************/

extern gdsl_stack_t
gdsl_stack_create (const char *name, gdsl_alloc_func_t alloc_f, 
		   gdsl_free_func_t free_f)
{
  register int i;
  gdsl_stack_t  s = NULL;

  s = (gdsl_stack_t) _gdsl_mem_alloc (sizeof (struct _gdsl_stack));
  if (s == NULL)
    return NULL;

  s->nodes = (gdsl_element_t *) _gdsl_mem_alloc ((1 + _GDSL_STACK_GROWING_FACTOR) 
						 * sizeof (gdsl_element_t));
  if (s->nodes == NULL)
    {
      _gdsl_mem_free (s);
      return NULL;
    }

  for (i = 0; i <= _GDSL_STACK_GROWING_FACTOR; i++)
    s->nodes[i] = NULL;

  s->card = 0UL;
  s->size = _GDSL_STACK_GROWING_FACTOR;
  s->name = NULL;

  if (name != NULL)
    {
      s->name = (char *) _gdsl_mem_alloc ((1 + strlen (name)) * sizeof (char));
      if (s->name == NULL)
	{
	  _gdsl_mem_free (s->nodes);
	  _gdsl_mem_free (s);
	  return NULL;
	}
      strcpy (s->name, name);
    }

  s->alloc_f = alloc_f ? alloc_f : default_alloc;
  s->free_f = free_f  ? free_f  : default_free;

  return s;
}

extern void 
gdsl_stack_destroy (gdsl_stack_t s)
{
  if (!gdsl_stack_is_empty (s))
    gdsl_stack_flush (s);

  _gdsl_mem_free (s->nodes);
  _gdsl_mem_free (s->name);
  _gdsl_mem_free (s);
}

extern gdsl_stack_t
gdsl_stack_flush (gdsl_stack_t s)
{
  register int i;

  for (i = 1; i <= s->card; i++)
    s->free_f (s->nodes[i]);

  _gdsl_mem_free (s->nodes);

  s->nodes = (gdsl_element_t *) _gdsl_mem_alloc ((1 + _GDSL_STACK_GROWING_FACTOR) 
						 * sizeof (gdsl_element_t));
  if (s->nodes == NULL)
    return NULL;

  for (i = 0; i <= _GDSL_STACK_GROWING_FACTOR; i++)
    s->nodes[i] = NULL;

  s->card = 0UL;
  s->size = _GDSL_STACK_GROWING_FACTOR;

  return s;
}

extern gdsl_stack_t
gdsl_stack_set_name (gdsl_stack_t s, const char *name)
{
  if (s->name != NULL)
    {
      _gdsl_mem_free (s->name);
      s->name = NULL;
    }

  if (name != NULL)
    {
      s->name = (char *) _gdsl_mem_alloc ((1 + strlen (name)) * sizeof (char));

      if (s->name == NULL)
	return NULL;

      strcpy (s->name, name);
    }

  return s;
}

/******************************************************************************/
/* Information functions                                                      */
/******************************************************************************/

extern const char *
gdsl_stack_get_name (gdsl_stack_t s)
{
  return s->name;
}

extern bool 
gdsl_stack_is_empty (gdsl_stack_t s)
{
  return (bool) (s->card == 0 ? TRUE : FALSE);
}

extern gdsl_element_t
gdsl_stack_get_top (gdsl_stack_t s)
{
  return s->nodes[s->card];
}

extern gdsl_element_t
gdsl_stack_get_bottom (gdsl_stack_t s)
{
  return s->nodes[1];
}

/******************************************************************************/
/* Insert functions                                                           */
/******************************************************************************/

extern gdsl_element_t
gdsl_stack_push (gdsl_stack_t s, void *key)
{
  gdsl_element_t e;

  e = (s->alloc_f) (key);
  if (e == NULL)
    return NULL;

  if (s->card == _GDSL_STACK_GROWING_FACTOR)
    {
      s->nodes = _gdsl_mem_realloc (s->nodes, (1 + s->size + _GDSL_STACK_GROWING_FACTOR) 
				    * sizeof (gdsl_element_t));
      if (s->nodes == NULL)
	{
	  s->free_f (e);
	  return NULL;
	}
      s->size += _GDSL_STACK_GROWING_FACTOR;
    }

  s->card++;
  s->nodes[s->card] = e;
  return e;
}

/******************************************************************************/
/* Remove functions                                                           */
/******************************************************************************/

extern gdsl_element_t
gdsl_stack_pop (gdsl_stack_t s)
{
  gdsl_element_t e;

  if (s->card == 0)
    return NULL;

  e = s->nodes[s->card];
  s->nodes[s->card] = NULL;
  s->card--;
  return e;
}

/******************************************************************************/
/* Search functions                                                           */
/******************************************************************************/

extern gdsl_element_t
gdsl_stack_search (gdsl_stack_t s, gdsl_compare_func_t comp_f, void *key)
{
  register int i;

  for (i = 1; i <= s->card; i++)
    if (comp_f (s->nodes[i], key) == 0)
      return s->nodes[i];

  return NULL;
}

/******************************************************************************/
/* Parse functions                                                            */
/******************************************************************************/

extern void
gdsl_stack_map (gdsl_stack_t s, gdsl_map_func_t map_f, void *user_data)
{
  register int i;

  for (i = 1; i <= s->card; i++)
    map_f (s->nodes[i], user_data);
}

/******************************************************************************/
/* Output functions                                                           */
/******************************************************************************/

extern void
gdsl_stack_write (gdsl_stack_t s, gdsl_write_func_t write_f, FILE *file, 
		  void *user_data)
{
  register int i;

  for (i = 1; i <= s->card; i++)
    write_f (s->nodes[i], file, user_data);
}

extern void
gdsl_stack_write_xml (gdsl_stack_t s, gdsl_write_func_t write_f, FILE *file, 
		      void *user_data)
{
  register int i;

  fprintf (file, "<GDSL_STACK REF=\"%p\" NAME=\"%s\" CARD=\"%ld\" TOP=\"%p\" BOTTOM=\"%p\">\n", 
	   (void *) s, s->name, s->card, (void *) &(s->nodes[s->card]), (void *) &(s->nodes[1]));

  for (i = 1; i <= s->card; i++)
    {
      fprintf (file, "<GDSL_STACK_NODE REF=\"%p\" CONTENT=\"%p\">", 
	       (void *) &(s->nodes[i]), (void *) s->nodes[i]);
      
      if (write_f && s->nodes[i])
	write_f (s->nodes[i], file, user_data);

      fprintf (file, "</GDSL_STACK_NODE>\n");
    }

  fprintf (file, "</GDSL_STACK>\n");
}

extern void
gdsl_stack_dump (gdsl_stack_t s, gdsl_write_func_t write_f, FILE *file, void *user_data)
{
  register int i;

  fprintf (file, "<GDSL_STACK REF=\"%p\" NAME=\"%s\" CARD=\"%ld\" TOP=\"%p\" BOTTOM=\"%p\" SIZE=\"%ld\">\n", 
	   (void *) s, s->name, s->card, (void *) &(s->nodes[s->card]), (void *) &(s->nodes[1]), s->size);

  for (i = 0; i <= s->size; i++)
    {
      fprintf (file, "<GDSL_STACK_NODE REF=\"%p\" CONTENT=\"%p\">", 
	       (void *) &(s->nodes[i]), (void *) s->nodes[i]);
      
      if (write_f && s->nodes[i])
	write_f (s->nodes[i], file, user_data);

      fprintf (file, "</GDSL_STACK_NODE>\n");
    }
}

/******************************************************************************/
/* Private functions                                                          */
/******************************************************************************/

static gdsl_element_t default_alloc (void *e)
{
  return e;
}

static void default_free (gdsl_element_t e)
{
  ;
}
